import pymzml
from pylab import *
from scipy import stats
from scipy.interpolate import spline
from scipy.signal import argrelextrema
from statsmodels.nonparametric.smoothers_lowess import lowess
import glob
import numpy
import os
import sys
import csv
import timeit
import re
from pyteomics import mass
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
import matplotlib.gridspec as gs
import matplotlib.colorbar as cb
from mpl_toolkits.mplot3d import axes3d, Axes3D
from matplotlib.font_manager import FontProperties
import mpl_toolkits.axisartist as axisartist
from mpl_toolkits.axisartist.grid_finder import MaxNLocator
import warnings
warnings.filterwarnings("ignore")

# http://epydoc.sourceforge.net/manual-fields.html#module-metadata-variables
__author__ = "Lang Ho Lee and Brett Pieper"
__copyright__ = "This software script is protected by the Copyright Act of 1976, 17 U.S.C. §§ 101-810, as amended. Rights reserved.  Please contact The Brigham and Women’s Hospital, Inc. for further information."
__credits__ = ["Lang Ho Lee", "Brett Pieper", "Sasha Singh"]
__license__ = "GPL"
__version__ = "1.0 (2016-05-31)"
__maintainer__ = "Lang Ho Lee"
__email__ = "llee27@partners.org"
__status__ = "Development"  # "Prototype", "Development", or "Production"

# Print out the script info
pgm_info = """eXtraction of PRM ion intensities (XPI) ver.{0}
Credits          : {1}
Maintainer       : {2} ({3})
Copyright        : {4}
License          : {5}""".format(__version__, ", ".join(__credits__), __maintainer__,
                                 __email__, __copyright__, __license__)


# You can adjust the figure size by setting flag_figure_size = True
# Default:  flag_figure_size = False
flag_figure_size = False
# When flag_figure_size = True, the following variables will be the size (inch)
fixed_height = 2
fixed_width = 3

# These were used for XPI paper
# check_hdl_ylim_ion_filtering and check_hdl_ylim
is_hdl = False
regression_flag = False

# Default ratio_max_threshold to ignore it
default_ratio_max_threshold = 0.0

# XPI output file names
file_all = "XPI_output_all.txt"
file_out = "XPI_output.txt"
file_check = "XPI_output_4check.txt"
file_check_prev_peaks = "XPI_output_4check_before_peaks.txt"
file_time = "XPI_output_time.txt"
file_transition = "XPI_transition.txt"
file_viz_config = "XPIViz_config.txt"
file_peak_config = "XPIPeak_config.txt"
file_peak_info = "XPI_output_peaks.txt"
file_refined = "XPI_output_refined.csv"
file_filtered = "XPI_selected.csv"

# XPIViz directories
dir_2d_profile = "2D_Profile"
dir_3d_profile = "3D_Profile"
dir_scatter_protein = "Scatter_Plots_Protein"
dir_scatter_peptide = "Scatter_Plots_Peptide"
dir_scatter_filter = "Filtering"
dir_std_curve = "Standard_Curves"
dir_peak_picking = "Peak_Picking"
dir_ion_lines = "Ion_Lines"

# XPIQuant config and XPI output columns
colname_prm_run = "PRM_Run_Name"
colname_transit_id = "Transition_ID"
colname_config_ratio = 'Ratio'
colname_modification = 'Modification'
colname_data_directory = 'Data directory'
colname_inclusion_list = 'Inclusion list'
colname_pick_method = 'Pick Method'
colname_max_peak = 'Max peak number'
colname_ttr = "Enrichment"
colname_normalize = "TIC Normalize"
colname_skip_inclusion = "Skip Inclusion"
colname_ms2_window = 'MS2 window'
colname_max_length_fragment = 'Max fragment'
colname_background = 'Background subtraction'
colname_max_l_sites = 'Max L sites'
colname_ax_ion = 'A Ion'

colname_sum_intensity = 'sum_intensity'
colname_sum_ratio = 'sum_ratio'
colname_sum_ratio_max = 'sum_ratio_max'
colname_ylim = 'ylim_viz'

# For XPIQuant config file
PRM_fragment_min_length = 2
default_max_num_peaks = 1
default_picking_method = 'SUM'
list_intensity_picking_method = ['MAX', 'TOP3', 'SUM', 'AVG', 'MED', 'QMAX', 'QSUM', 'QTOP3', 'QAVG']
list_requirements = [colname_data_directory, colname_inclusion_list, 'Labeling', colname_ms2_window,
                     colname_max_length_fragment, colname_modification, colname_config_ratio, colname_ttr,
                     colname_normalize, colname_skip_inclusion, colname_background, colname_max_l_sites, colname_ax_ion]

# Transition file column names
colname_theoretical_mz = 'Theoretical_MZ'
colname_fragment_mz = 'Fragment_MZ'
colname_pep_seq = 'Peptide Sequence'
colname_ion_series = 'ion-series'
colname_precursor_charge = 'precursor_charge'
colname_rt = 'RT'
colname_prot_id = 'Protein Name'
colname_filter_string = 'filter string'
transition_file_header = [colname_theoretical_mz, colname_rt, colname_pep_seq, colname_prot_id,
                          colname_precursor_charge, colname_filter_string, colname_fragment_mz,
                          colname_ion_series, 'Fragment_Sequence']

# Decimal place for showing numbers in the plots
decimal_place = 2

# for all the ticks
font_size_label = 30
font_size_ticks = 20
font_size_label_2d = 10
font_size_ticks_2d = 7
circle_size = 100
circle_size_small = 100

# for all the transparent level
alpha = 0.5

# delimiters
delimiter_tab = '\t'
delimiter_label = ','
delimiter_residue = ':'
delimiter_config = '+'
delimiter_ratio = '/'
delimiter_dash = '-'
delimeter_replicates = '_'

# Inclusion list file
colname_inclusion_mass = "Mass [m/z]"
colname_inclusion_charge = "CS [z]"
colname_inclusion_rt_start = "Start [min]"
colname_inclusion_rt_end = "End [min]"
colname_inclusion_seq = "Sequence"
colname_inclusion_prot = "Protein"
list_requirements_inclusion = [colname_inclusion_mass, colname_inclusion_charge, colname_inclusion_rt_start,
                               colname_inclusion_rt_end, colname_inclusion_seq, colname_inclusion_prot]
# For XPI_output.txt
list_label_header = ["Average_m/z", "Intensity", "#_PRMs","Average_RT", "Max_deltaMZ"]
# xpi_output_header = "PRM_Run_Name\tTransition_ID\tProtein_Name\tPrecursor_m/z\tPrecursor_Charge\tRetention_Time\t" \
# "Peptide_Sequence\tIon-Series"
xpi_output_header = "PRM_Run_Name\tTransition_ID\tProtein_Name\tMS1_Isolation_m/z\tPrecursor_Charge\tRetention_Time\t" \
                    "Fragment_Sequence\tIon-Series"
# For XPIViz
mode_for_2d_profile = '2dp'
mode_for_3d_profile = '3dp'
mode_for_scatter_protein = 'prot'
mode_for_scatter_peptide = 'pep'
mode_for_standard_curve = 'stdc'
mode_for_output = 'report'
mode_for_filter = 'fil'
mode_for_linegraph_ions = 'ions'

# For fonts
font = FontProperties()
font.set_family('sans-serif')
alignment = {'horizontalalignment': 'center', 'verticalalignment': 'baseline'}
plt.rcParams['font.sans-serif'] = 'Arial'

# Usages
xpi_viz_config = "XPIViz configurations:\t\tSee the manual section 9"


# Ancestor class
class File(object):
    def __init__(self, filename):
        self.filename = filename

    def get_filename(self):
        return self.filename

    def get_file(self, delim='\t'):
        fp = open(self.filename, "rU")
        fp_data = csv.DictReader(fp, delimiter=delim)
        return fp_data

    def check_file(self):
        if not os.path.exists(self.filename):
            print("Not Found: {0}".format(self.filename))
            exit()
        else:
            pass


class Message:
    @staticmethod
    def print_message(message, usage=''):
        if usage != '':
            print(usage)
        print(message)
        exit()


class Timer:
    def __init__(self):
        self.messages = []
        self.times = []

        self.start = ''
        self.end = ''
        self.elapsed = ''

    def start_timer(self, message, is_print=True):
        self.start = timeit.default_timer()
        self.messages.append(message)
        if is_print:
            print(message, end=" ... ")
        else:
            if message != "":
                print(message)

    def stop_timer(self):
        """
        Stop timer and calculate elapsed time
        """
        self.end = timeit.default_timer()
        self.elapsed = self.end - self.start
        self.times.append(self.elapsed)

    def print_elapsed(self):
        """
        Print out the elapsed time
        """
        print("Done in {0:.4f} seconds".format(self.elapsed))

    def get_times(self):
        total = 0.0
        return_string = ""
        for i in range(len(self.messages)):
            return_string += "{0}\t{1:.4f} seconds\n".format(self.messages[i], self.times[i])
            total += self.times[i]
        return_string += "Total elapsed\t{0:.4f} seconds\n".format(total)
        return return_string


class Progress:
    def __init__(self, total, desc=''):
        self.total = total
        self.curruent = 0
        self.desc = desc

    def print_progress(self, current):
        """
        Print out progress
        """
        if current < self.total:
            sys.stdout.write('{2}... {0}/{1}\r'.format(current, self.total, self.desc))
        else:
            print('{2}... {0}/{1}\r'.format(current, self.total, self.desc))


class XPIVizConfig:
    def __init__(self, file_path):
        self.file_path = file_path
        self.intensity_picking_method = 'SUM'
        self.flag_log = True
        self.flag_same_z = True
        self.flag_trend_line = True
        self.percent_ratio = True
        self.ratio_name = ''
        self.standard_label = ''
        self.is_colorful = ''
        self.dpi = 100
        self.data = {}
        self.filters = {}

        self.parse_file()

    @staticmethod
    def read_flag(flag_input):
        """
        Check false or true from user input
        :param flag_input: user input
        :return: Boolean type
        """
        flg = flag_input.lower()
        if flg == 'true':
            return True
        elif flg == 't':
            return True
        elif flg == 'false':
            return False
        elif flg == 'f':
            return False
        else:
            return False

    def parse_file(self):
        """
        Parse a config file for visualization of XPI data
        """
        fp = open(self.file_path, "rU")
        for data_line in fp:
            if data_line.strip() == "":
                continue
            if re.match("#", data_line):
                continue
            list_data = data_line.strip().split('\t')
            item = list_data[0]
            if item == 'Data':
                file_name = list_data[1]
                concentration = list_data[2]
                unit = list_data[3]
                labeling_name = list_data[4]
                if len(list_data) == 6:
                    rep_name = list_data[5]
                else:
                    rep_name = ''
                if file_name not in self.data:
                    self.data[file_name] = {}
                self.data[file_name]['MR'] = concentration
                self.data[file_name]['RatioName'] = labeling_name
                self.ratio_name = labeling_name
                self.data[file_name]['Unit'] = unit
                self.data[file_name]['Replicate'] = rep_name
            # For Standard curve
            elif item == 'Log10':
                self.flag_log = self.read_flag(list_data[1])
            # For Standard curve and Time-series plots (Scatter plot)
            elif item == 'Pick Method':
                if list_data[1].upper() not in list_intensity_picking_method:
                    Message.print_message("Wrong intensity-picking method: {0}".format(list_data[1]))
                self.intensity_picking_method = list_data[1]
            # Reference label that is usually more intense than other labels
            elif item == 'Standard label':
                self.standard_label = list_data[1]
            # For Time-series (Scatter plot)
            elif item == 'Median trend line':
                self.flag_trend_line = self.read_flag(list_data[1])
            # For Time-series (Scatter plot)
            elif item == 'Percent Ratio':
                self.percent_ratio = self.read_flag(list_data[1])
            # For Mass profile
            elif item == 'Make Zs same':
                self.flag_same_z = self.read_flag(list_data[1])
            elif item == 'Protein Color':
                self.is_colorful = self.read_flag(list_data[1])
            elif item == 'Filtering':
                prot_name = list_data[1]
                try:
                    threshold_ratio = float(list_data[2])
                except ValueError:
                    print("{0} is not a number.".format(list_data[2]))
                    print("Ratio sum threshold for {0} is set as 0.0 as default".format(prot_name))
                    threshold_ratio = 0.0
                try:
                    threshold_ratio_max = float(list_data[3])
                except ValueError:
                    print("{0} is not a number.".format(list_data[3]))
                    print("Ratio sum threshold for {0} is set as 1.0 as default".format(prot_name))
                    threshold_ratio_max = default_ratio_max_threshold
                if threshold_ratio_max == 0.0:
                    print("Max_Ratio_Sum is 0, so ignore Max_Ratio_Sum threshold in {0}".format(prot_name))
                try:
                    threshold_inten = float(list_data[4])
                except ValueError:
                    print("{0} is not a number.".format(list_data[4]))
                    print("Ratio sum threshold for {0} is set as 0.0 as default".format(prot_name))
                    threshold_inten = 0.0
                try:
                    ylim_viz = float(list_data[5])
                except IndexError:
                    print("You did not select y limit for protein and peptide plots")
                    ylim_viz = -1
                except ValueError:
                    print("{0} is not a number.".format(list_data[5]))
                    print("Y limit for {0} will not be used".format(prot_name))
                    ylim_viz = -1
                if prot_name in self.filters:
                    Message.print_message("Redundant proteins in 'Filtering' of XPIViz_config.txt: {0}".format(
                        prot_name))
                self.filters[prot_name] = {}
                self.filters[prot_name][colname_sum_ratio] = threshold_ratio
                self.filters[prot_name][colname_sum_intensity] = threshold_inten
                self.filters[prot_name][colname_sum_ratio_max] = threshold_ratio_max
                self.filters[prot_name][colname_ylim] = ylim_viz
                # self.filters[prot_name] = {'sum_ratio': threshold_ratio, 'sum_intensity': threshold_inten, 'sum_ratio_max': threshold_ratio_max}
            # For Mass profile
            elif item == 'DPI':
                try:
                    self.dpi = int(list_data[1])
                except ValueError:
                    print("DPI should be integer number, such as 100 or 300.  DPI is 100 as default")
                    self.dpi = 100
            else:
                continue


class Scatterplot:
    def __init__(self, dict_observed_ratio, dict_run2info, data_directory, dpi, is_log=False, poly_level=1):
        """
        Draw scatter plots and calculate Pearson Correlation Coefficients (PCC)
        :param dict_observed_ratio: dict_observed_ratio should have "PRM_Run_Name"
                                    and "Transition_ID" as well as Ratio data
        :param dict_run2info: dict_run2info should have PRM Run information in 'MR' for mixing ratio
                              and 'RatioName' for labeling
        :param data_directory: data_directory is a destination of output files and images
        :param dpi: DPI for image quality
        :param is_log: if it is true, get log10
        :param poly_level: 1=linear, >2=smooth fit
        :return: none
        """
        self.observed_data = dict_observed_ratio
        self.dict_run2info = dict_run2info
        self.data_directory = data_directory
        self.dpi = dpi
        self.is_log = is_log
        self.poly_level = poly_level
        self.dict_transit2pcc = {}
        self.list_time_points = []
        self.prot_list = []

        self.ratio_name = get_ratio_name(dict_run2info)
        self.dict_extracted_ratio = self.get_ratio()
        self.dict_transit2data = self.make_plotting_data()

    def get_ratio(self):
        """
        Extract Ratio information for each transitions and each PRM experiments
        :return: a dictionary {Transition ID:{PRM Run Name:Ratio}}
        """
        return_dictionary = {}
        for transit_id in self.observed_data:
            for prm_run in self.observed_data[transit_id]:
                if transit_id not in return_dictionary:
                    return_dictionary[transit_id] = {}
                if prm_run in return_dictionary[transit_id]:
                    print("Skip redundant data: {0} {1}".format(transit_id, prm_run))
                ratio = get_value_by_col_name(self.ratio_name, self.observed_data[transit_id][prm_run])
                return_dictionary[transit_id][prm_run] = ratio
        return return_dictionary

    @staticmethod
    def get_log10(number):
        if number < 0:
            Message.print_message("Ratio is negative value")
        elif number == 0:
            return numpy.inf
        else:
            return -1 * numpy.log10(number)

    def get_time_points(self):
        """
        Get every time points from XPIViz_config
        :return: none
        """
        for prm_run in self.dict_run2info:
            time = float(self.dict_run2info[prm_run]['MR'])
            if time not in self.list_time_points:
                time_point = time
                if self.is_log:
                    time_point = self.get_log10(time)
                if time_point not in self.list_time_points:
                    self.list_time_points.append(time_point)
        self.list_time_points.sort()

    def make_plotting_data(self):
        """
        Make a dictionary for standard curve
        :return: a dictionary {Transition ID:{Ratio Name:{'observed': [values], 'intended': [values]}}}
        """
        return_dictionary = {}
        for tr_id in self.dict_extracted_ratio:
            for prm_run in self.dict_extracted_ratio[tr_id]:
                # m0 intensity and m3 intensity
                m0_intensity = self.observed_data[tr_id][prm_run]['m0_intensity']
                m3_intensity = self.observed_data[tr_id][prm_run]['m3_intensity']
                replica_name = self.dict_run2info[prm_run]['Replicate']
                # Get mixing ratio info from config
                intended_ratio = get_value_by_col_name(col_name=prm_run, dict_arg=self.dict_run2info)
                if intended_ratio == "NA":
                    Message.print_message("'{0}' was not found in your XPIViz_config.txt.".format(prm_run))
                else:
                    intended_ratio = float(get_value_by_col_name(col_name='MR', dict_arg=intended_ratio))
                    # intended_ratio = float(intended_ratio['MR'])
                # Ratio:Heavy/Light
                # ratio_name = get_value_by_col_name(col_name=prm_run, dict_arg=self.dict_run2info)['RatioName']
                dict_tmp = get_value_by_col_name(col_name=prm_run, dict_arg=self.dict_run2info)
                ratio_name = get_value_by_col_name(col_name='RatioName', dict_arg=dict_tmp)
                if not re.search(":", ratio_name):
                    ratio_name = "{1}:{0}".format(ratio_name.strip(), colname_config_ratio)
                # get observed_ratio
                observed_ratio = get_value_by_col_name(prm_run, self.dict_extracted_ratio[tr_id])
                if observed_ratio == 'NA':
                    # print("Unavailable ratio found: {0}\t{1}\t{2}".format(observed_ratio, tr_id, prm_run))
                    continue
                # If observed ratio == 'Infinity', skip it
                elif observed_ratio == 'Infinity' or observed_ratio == 'inf' or observed_ratio == 'infinity' \
                        or observed_ratio == 'Inf':
                    # print("Unavailable ratio found: {0}\t{1}\t{2}".format(observed_ratio, tr_id, prm_run))
                    continue
                observed_ratio = float(observed_ratio)
                # Initialize data dictionary
                if tr_id not in return_dictionary:
                    return_dictionary[tr_id] = {}
                if ratio_name not in return_dictionary[tr_id]:
                    return_dictionary[tr_id][ratio_name] = {'observed': [], 'intended': [],
                                                            'm0_intensity': [], 'm3_intensity': [], 'Replicate': []}
                # If user chose log10
                unit = get_value_by_col_name('Unit', get_value_by_col_name(col_name=prm_run,
                                                                           dict_arg=self.dict_run2info))
                if self.is_log:
                    return_dictionary[tr_id][ratio_name]['intended'].append(self.get_log10(intended_ratio))
                    return_dictionary[tr_id][ratio_name]['observed'].append(self.get_log10(observed_ratio))
                    return_dictionary[tr_id][ratio_name]['Unit'] = "Log10({0})".format(unit)
                    return_dictionary[tr_id][ratio_name]['Label_ratio'] = "Log10({0})".format(ratio_name)
                else:
                    return_dictionary[tr_id][ratio_name]['intended'].append(intended_ratio)
                    return_dictionary[tr_id][ratio_name]['observed'].append(observed_ratio)
                    return_dictionary[tr_id][ratio_name]['Unit'] = unit
                    # get_value_by_col_name(col_name=prm_run, dict_arg=self.dict_run2info)['Unit']
                    return_dictionary[tr_id][ratio_name]['Label_ratio'] = ratio_name
                # TO DO: Take a log10 here for logged ion filtering plot (draw_intensity_ratio_plot)
                return_dictionary[tr_id][ratio_name]['m0_intensity'].append(m0_intensity)
                return_dictionary[tr_id][ratio_name]['m3_intensity'].append(m3_intensity)
                return_dictionary[tr_id][ratio_name]['Replicate'].append(replica_name)
                # Protein
                tmp = get_value_by_col_name(col_name=prm_run, dict_arg=self.observed_data[tr_id])
                return_dictionary[tr_id][ratio_name]['Protein'] = get_value_by_col_name(col_name='Protein_Name',
                                                                                        dict_arg=tmp)
        return return_dictionary

    @staticmethod
    def filter_unvalid_logratio(observed_arg, intended_arg):
        return_dict = {'intended': [], 'observed': []}
        zip_array = zip(observed_arg, intended_arg)
        for observed_local, intended_local in zip_array:
            if observed_local == numpy.inf:
                continue
            return_dict['intended'].append(intended_local)
            return_dict['observed'].append(observed_local)
        return return_dict

    def process_ratio_data(self, dict_tr2sum_ri, mode='standard_curve'):
        # Draw scatter plots and a regression line
        # for progress numbers
        # Data collection for drawing standard curves
        total_pg = 0
        for transition_id in self.observed_data:
            if transition_id not in self.dict_transit2data:
                continue
            total_pg += self.filter_fragment_by_sum_ratio(dict_tr2sum_ri[transition_id])
        pg_num = 1
        pg = Progress(total_pg, "Processing results...")
        # for error and warning report
        fig_num = 0
        report = ""
        # for recording results
        self.dict_transit2pcc = {}
        # A dictionary for plotting standard curve for each proteins
        dict_prot2data = {}
        for transition_id in self.observed_data:
            # check if it has zero observation
            if transition_id not in self.dict_transit2data:
                report = append_string(template=report,
                                       short_string="Skip {0} since it has less than 0 valid observations".
                                       format(transition_id),
                                       glue='\n')
                continue
            # Data collection for drawing standard curves
            if mode != 'pcc':
                if not self.filter_fragment_by_sum_ratio(dict_tr2sum_ri[transition_id]):
                    continue
                # print out progress while making images
                pg.print_progress(pg_num)
                pg_num += 1
            # Calculate PCC and draw scatter plots
            pcc = 0.0
            for ratio_name in self.dict_transit2data[transition_id]:
                dict_tmp = self.filter_unvalid_logratio(self.dict_transit2data[transition_id][ratio_name]['observed'],
                                                        self.dict_transit2data[transition_id][ratio_name]['intended'])
                ratio_mixing = dict_tmp['intended']
                ratio_observed = dict_tmp['observed']
                prot_id = self.dict_transit2data[transition_id][ratio_name]['Protein']
                # check if it has less than three observations due to R2 calculation
                num_obs = len(ratio_observed)
                if num_obs < 3:
                    report = append_string(template=report,
                                           short_string="Skip {0} since it has less than 3 valid observations: {1}".
                                           format(transition_id, num_obs),
                                           glue='\n')
                    continue
                # Initiate dict_prot2data
                if prot_id not in dict_prot2data:
                    dict_prot2data[prot_id] = {'intended': [], 'observed': []}
                # Add data to dict_prot2data
                dict_prot2data[prot_id]['intended'].extend(ratio_mixing)
                dict_prot2data[prot_id]['observed'].extend(ratio_observed)
                # Set basic information for making figures
                fig_title = "{0}_{1}".format(transition_id, ratio_name)
                x_axis_label = self.dict_transit2data[transition_id][ratio_name]['Unit']
                y_axis_label = self.dict_transit2data[transition_id][ratio_name]['Label_ratio']
                # labels for protein standard curves
                dict_prot2data[prot_id]['x_label'] = x_axis_label
                dict_prot2data[prot_id]['y_label'] = y_axis_label
                if self.is_log:
                    x_axis_label = "Log10({0})".format(x_axis_label)
                    y_axis_label = "Log10({0})".format(ratio_name)
                # Make images if the user want
                if mode == 'standard_curve':
                    pcc = draw_scatter_plot_reg_line(ratio_mixing, ratio_observed, fig_title, self.data_directory,
                                                     x_axis_label, y_axis_label, self.dpi, self.poly_level)
                # Calculate PCC only
                elif mode == 'pcc':
                    pcc = get_pcc(ratio_mixing, ratio_observed)
                else:
                    Message.print_message(
                        "Wrong mode selection. Choose one of these 'standard_curve', 'time_series' or 'pcc'")
                # deposit analysis results
                if transition_id not in self.dict_transit2pcc:
                    self.dict_transit2pcc[transition_id] = {}
                self.dict_transit2pcc[transition_id]['Ratio_Name'] = ratio_name
                self.dict_transit2pcc[transition_id]['PCC'] = pcc
                self.dict_transit2pcc[transition_id]['Number_OBS'] = num_obs
            fig_num += 1
        # Print out messages
        # print(report)
        if mode == 'standard_curve':
            print("Completed making {0} standard curve figures".format(fig_num))
            # Draw standard curve for each proteins
            for prot_id in dict_prot2data:
                pcc = draw_scatter_plot_reg_line(dict_prot2data[prot_id]['intended'],
                                                 dict_prot2data[prot_id]['observed'],
                                                 prot_id, self.data_directory,
                                                 dict_prot2data[prot_id]['x_label'],
                                                 dict_prot2data[prot_id]['y_label'], self.dpi, self.poly_level)
                print("Standard curve for protin {0} is done (R2={1:.4f})".format(prot_id, pcc))
        return self.dict_transit2pcc

    def merge_ions_by_peptide(self, dict_tr2sum_ri):
        """
        Organize xpi_output and ratio information by peptide sequence
        """
        dict_return = {}
        for transition_id in self.dict_transit2data:
            sequence = transition_id.split(delimiter_dash)[0]
            ion_series = transition_id.split(delimiter_dash)[1]
            for ratio_name in self.dict_transit2data[transition_id]:
                m0_intensity = self.dict_transit2data[transition_id][ratio_name]['m0_intensity']
                time_point = self.dict_transit2data[transition_id][ratio_name]['intended']
                enrichments = self.dict_transit2data[transition_id][ratio_name]['observed']
                unit = self.dict_transit2data[transition_id][ratio_name]['Unit']
                rep_names = self.dict_transit2data[transition_id][ratio_name]['Replicate']
                protein = self.dict_transit2data[transition_id][ratio_name]['Protein']
                if self.filter_fragment_by_sum_ratio(dict_tr2sum_ri[transition_id]):
                    if sequence not in dict_return:
                        dict_return[sequence] = {}
                    if ratio_name not in dict_return[sequence]:
                        dict_return[sequence][ratio_name] = {'observed': [], 'intended': [], 'ion_series': [],
                                                             'ion_series2': [], 'm0_intensity': []}
                    dict_return[sequence][ratio_name]['intended'].extend(time_point)
                    dict_return[sequence][ratio_name]['observed'].extend(enrichments)
                    dict_return[sequence][ratio_name]['m0_intensity'].extend(m0_intensity)
                    dict_return[sequence][ratio_name]['ion_series2'].extend(rep(ion_series, len(time_point)))
                    dict_return[sequence][ratio_name]['ion_series'].extend(self.make_uniq_replicate_id(ion_series, rep_names))
                    dict_return[sequence][ratio_name]['Unit'] = unit
                    dict_return[sequence][ratio_name]['Protein'] = protein
        return dict_return

    @staticmethod
    def filter_fragment_by_sum_ratio(dict_filtering, mode_arg='selected'):
        if mode_arg == 'all':
            return True
        elif dict_filtering[colname_sum_ratio]:
            return False
        elif dict_filtering[colname_sum_intensity]:
            return False
        elif dict_filtering[colname_sum_ratio_max]:
            return False
        else:
            return True

    def merge_ions_by_protein(self, dict_tr2sum_ri, mode_arg='selected'):
        """
        Organize xpi_output and ratio information by protein
        """
        dict_return = {}
        for transition_id in sorted(self.dict_transit2data):
            sequence = transition_id.split(delimiter_dash)[0]
            ion_series = transition_id.split(delimiter_dash)[1]
            pep_ion = "{0}-{1}".format(sequence, ion_series)
            for ratio_name in sorted(self.dict_transit2data[transition_id]):
                # Get the data from dict_transit2data
                m0_intensity = self.dict_transit2data[transition_id][ratio_name]['m0_intensity']
                m3_intensity = self.dict_transit2data[transition_id][ratio_name]['m3_intensity']
                ratio_mixing = self.dict_transit2data[transition_id][ratio_name]['intended']
                ratio_observed = self.dict_transit2data[transition_id][ratio_name]['observed']
                unit = self.dict_transit2data[transition_id][ratio_name]['Unit']
                prot = self.dict_transit2data[transition_id][ratio_name]['Protein']
                rep_names = self.dict_transit2data[transition_id][ratio_name]['Replicate']
                if prot not in self.prot_list:
                    self.prot_list.append(prot)
                # Filtering step of ratio sum and m0 intensity sum
                if self.filter_fragment_by_sum_ratio(dict_tr2sum_ri[transition_id], mode_arg=mode_arg):
                    # Initialize a dictionary
                    if prot not in dict_return:
                        dict_return[prot] = {}
                    if ratio_name not in dict_return[prot]:
                        dict_return[prot][ratio_name] = {'observed': [], 'intended': [], 'ion_series': [],
                                                         'ion_series2': [], 'm0_intensity': [], 'm3_intensity': []}
                    # Assign data to the dictionary
                    dict_return[prot][ratio_name]['intended'].extend(ratio_mixing)
                    dict_return[prot][ratio_name]['observed'].extend(ratio_observed)
                    dict_return[prot][ratio_name]['m0_intensity'].extend(m0_intensity)
                    dict_return[prot][ratio_name]['m3_intensity'].extend(m3_intensity)
                    dict_return[prot][ratio_name]['ion_series2'].extend(rep(pep_ion, len(ratio_mixing)))
                    dict_return[prot][ratio_name]['ion_series'].extend(self.make_uniq_replicate_id(pep_ion, rep_names))
                    dict_return[prot][ratio_name]['Unit'] = unit
        return dict_return

    @staticmethod
    def make_uniq_replicate_id(pep_ion_arg, rep_names_arg):
        return_list = []
        for rname in rep_names_arg:
            return_list.append(make_unique_replicate_transition_id(pep_ion_arg, rname))
        return return_list

    def print_pcc(self, out_file):
        fp_out = open(out_file, "w")
        fp_out.write("Transition_ID,Pearson_Correlation_Coefficient (R2),#_of_Observations,Ratio_Name\n")
        for transition_id in self.dict_transit2pcc:
            fp_out.write("{0},{1},{2},{3}\n".format(transition_id,
                                                    self.dict_transit2pcc[transition_id]['PCC'],
                                                    self.dict_transit2pcc[transition_id]['Number_OBS'],
                                                    self.dict_transit2pcc[transition_id]['Ratio_Name']))
        fp_out.close()
        print("Completed writing Standard_Curves_R2.csv")


class XPIPeak(File):
    def __init__(self, filename):
        File.__init__(self, filename)
        self.data = {}

        self.parse_file()

    def parse_file(self):
        file_contents = self.get_file()
        for row in file_contents:
            file_name = row['File_Name']
            labeling_group = row['Labeling_Group']
            tr_id = row['Transition']
            ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
            ion_series = row['Ion_series']
            # total_rt = row['Total_RT']
            selected_rt = row['Selected_RT']
            max_delta_mz = float(row['Max_Delta_Mass'])

            if peptide_sequence not in self.data:
                self.data[peptide_sequence] = {}
            if file_name not in self.data[peptide_sequence]:
                self.data[peptide_sequence][file_name] = {}
            if labeling_group not in self.data[peptide_sequence][file_name]:
                self.data[peptide_sequence][file_name][labeling_group] = {}
            if ion_series not in self.data[peptide_sequence][file_name][labeling_group]:
                self.data[peptide_sequence][file_name][labeling_group][ion_series] = \
                    {'peak_range': selected_rt, 'max_delta_mass': max_delta_mz}


class XPI4Check(File):
    def __init__(self, filename, ratio_name, standard_label, ttr=False):
        File.__init__(self, filename)
        self.ttr = ttr
        self.standard_label = standard_label
        self.ratio_name = ratio_name
        self.threshold = 10000.0
        self.list_ratio_components = self.check_ratio_name()
        self.numerator_name, self.denominator_name = self.parse_ratio_name()
        self.data = {}
        self.data_prot = {}

        self.parse_file()

    def parse_ratio_name(self):
        if self.ttr:
            try:
                denom = self.list_ratio_components[1].replace("(", "").replace(")", "").split(delimiter_config)[1]
                return self.list_ratio_components[0], denom
            except IndexError:
                Message.print_message("Ratio name is not fit to TTR in XPIViz_config.txt, Change TTR option to False")
                exit()
        else:
            return self.list_ratio_components[0], self.list_ratio_components[1]

    def check_ratio_name(self):
        if not re.match("{0}{1}".format(colname_config_ratio, delimiter_residue), self.ratio_name):
            Message.print_message("Wrong format of ratio name: {0}\n e.g. {1}:Heavy/Light".format(self.ratio_name,
                                                                                                  colname_config_ratio))
        list_ratio_components = self.ratio_name.split(delimiter_residue)[1].split(delimiter_ratio)
        if len(list_ratio_components) != 2:
            Message.print_message("Wrong format of ratio name: {0}\n e.g. {1}:Heavy/Light".format(self.ratio_name,
                                                                                                  colname_config_ratio))
        return list_ratio_components

    def get_intensity_from_data(self, label_name, method, data_row):
        selected_intensity = get_value_by_col_name(col_name="{0}{1}{2}".format(label_name, delimiter_residue, method), dict_arg=data_row, message_on=False)
        if selected_intensity != "NA":
            selected_intensity = float(selected_intensity)
        else:
            try:
                selected_intensity = float(label_name)
            except ValueError:
                Message.print_message("Wrong format of ratio name: {0}\n e.g. {1}:Heavy/Light".format(self.ratio_name, colname_config_ratio))
        return selected_intensity

    def parse_file(self):
        file_contents = self.get_file()
        for row in file_contents:
            transit_id = row['Transition_ID']
            prm_run = row['PRM_Run_Name']
            prot_id = row['Protein_Name']
            dict_method2ratio = {}
            dict_method2m0int = {}
            dict_method2hm3int = {}
            for method in list_intensity_picking_method:
                numerator_intensity = self.get_intensity_from_data(self.numerator_name, method, row)
                denominator_intensity = self.get_intensity_from_data(self.denominator_name, method, row)
                # get standard label intensity
                if self.numerator_name == self.standard_label:
                    dict_method2m0int[method] = numerator_intensity
                    dict_method2hm3int[method] = denominator_intensity
                elif self.denominator_name == self.standard_label:
                    dict_method2m0int[method] = denominator_intensity
                    dict_method2hm3int[method] = numerator_intensity
                else:
                    Message.print_message("Please check XPIViz config file. You have wrong label name")
                # Calculate ratio
                if float(dict_method2m0int[method]) >= self.threshold:
                    dict_method2ratio[method] = get_ratio(numerator_intensity, denominator_intensity, ttr=self.ttr)
                else:
                    dict_method2ratio[method] = 0
                    dict_method2m0int[method] = 0
                    dict_method2hm3int[method] = 0
            for method in dict_method2ratio:
                ratio = dict_method2ratio[method]
                m0int = dict_method2m0int[method]
                m3int = dict_method2hm3int[method]
                if method not in self.data:
                    self.data[method] = {}
                if transit_id not in self.data[method]:
                    self.data[method][transit_id] = {}
                if prm_run not in self.data[method][transit_id]:
                    self.data[method][transit_id][prm_run] = {self.ratio_name: ratio, 'Protein_Name': prot_id,
                                                              'm0_intensity': m0int, 'm3_intensity': m3int}
                # For validation box-plots
                if method not in self.data_prot:
                    self.data_prot[method] = {}
                if prot_id not in self.data_prot[method]:
                    self.data_prot[method][prot_id] = {}
                if transit_id not in self.data_prot[method][prot_id]:
                    self.data_prot[method][prot_id][transit_id] = {}
                if prm_run not in self.data_prot[method][prot_id][transit_id]:
                    self.data_prot[method][prot_id][transit_id][prm_run] = {self.ratio_name: ratio,
                                                                            'Protein_Name': prot_id,
                                                                            'm0_intensity': m0int,
                                                                            'm3_intensity': m3int}


class XPIOutput(File):
    def __init__(self, filename):
        File.__init__(self, filename)
        self.data = {}

        self.parse_file()

    def parse_file(self):
        file_contents = self.get_file()
        for row in file_contents:
            transit_id = row['Transition_ID']
            prm_run = row['PRM_Run_Name']
            if transit_id not in self.data:
                self.data[transit_id] = {}
            if prm_run not in self.data[transit_id]:
                self.data[transit_id][prm_run] = row
            else:
                print("Not found: {0}\t{1}\t{2}".format(prm_run, transit_id, row))


class XPIOutAll(File):
    def __init__(self, filename, config_viz):
        File.__init__(self, filename)
        self.data = {}
        self.tr2prot = {}
        self.config_viz = config_viz

        self.parse_file(filename)

    def parse_file(self, filename):
        fp = open(filename, "rU")
        total_rt = ''
        file_name = ''
        transition_id = ''
        labeling_group = ''
        flag_transition = False
        flag_scans = False
        for data_line in fp:
            list_data = data_line.strip().split('\t')
            first_item = list_data[0]
            if first_item == "":
                total_rt = ''
                flag_transition = False
                flag_scans = False
                continue
            if flag_transition:
                prot_name = list_data[0]
                self.tr2prot[transition_id] = prot_name
                total_rt = list_data[3]
                flag_transition = False
                continue
            if flag_scans:
                rt = float(list_data[3])
                # If file path of RAW MS data of XPIViz_config.txt is not matched to XPI_output_all...
                if transition_id == '':
                    Message.print_message("Check your XPIViz_config.txt file.  Data file names are not matched.")
                self.data[transition_id][file_name][labeling_group]['total_rt'].append(total_rt)
                self.data[transition_id][file_name][labeling_group]['rt'].append(rt)
                self.data[transition_id][file_name][labeling_group]['mz'].append(float(list_data[4]))
                intensity = float(list_data[5])
                self.data[transition_id][file_name][labeling_group]['intensity'].append(intensity)
                mz_diff = float(list_data[7])
                self.data[transition_id][file_name][labeling_group]['mz_diff'].append(mz_diff)
                continue
            if first_item in self.config_viz.data:
                file_name = first_item
                labeling_group = list_data[1]
                transition_id = list_data[2]
                if transition_id not in self.data:
                    self.data[transition_id] = {}
                if file_name not in self.data[transition_id]:
                    self.data[transition_id][file_name] = {}
                if labeling_group not in self.data[transition_id][file_name]:
                    self.data[transition_id][file_name][labeling_group] = {'intensity': [], 'mz_diff': [], 'rt': [],
                                                                           'total_rt': [], 'prot_name': '', 'mz': []}
                total_rt = ''
                flag_transition = False
                flag_scans = False
                continue
            elif first_item == "Protein_Name":
                flag_transition = True
                continue
            elif first_item == "Scan_ID":
                flag_scans = True
                continue

    def get_intensities(self, transition_id):
        return self.data[transition_id]

    def print_intensities(self, transition_id):
        if transition_id not in self.data:
            print("\"{0}\" was not found".format(transition_id))
            return ''
        return_string = "File_Name,Labeling,Xlabel,Intensity\n"
        for file_name in self.data[transition_id]:
            for labeling_group in self.data[transition_id][file_name]:
                for intensity in self.data[transition_id][file_name][labeling_group]:
                    return_string += "{0},{1},{2},{3}\n".format(
                        file_name, labeling_group, file_name + "-" + labeling_group, intensity
                    )
        return return_string


class XPIPeakInfo(File):
    def __init__(self, filename):
        File.__init__(self, filename)
        self.data = {}

        self.check_file()
        self.parse_file()

    def parse_file(self):
        fp_data = self.get_file()
        for row in fp_data:
            file_name = row["File_Name"]
            labeling_group = row["Labeling_Group"]
            tr_id = row["Transition"]
            peak_range = row["Selected_RT"]
            try:
                max_delta_mass = float(row["Max_Delta_Mass"])
            except ValueError:
                Message.print_message("{0} is not number".format(row["Max_Delta_Mass"]))
                max_delta_mass = 1.0
            ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
            if peptide_sequence not in self.data:
                self.data[peptide_sequence] = {}
            if file_name not in self.data[peptide_sequence]:
                self.data[peptide_sequence][file_name] = {}
            if labeling_group not in self.data[peptide_sequence][file_name]:
                self.data[peptide_sequence][file_name][labeling_group] = {}
            if ion_series not in self.data[peptide_sequence][file_name][labeling_group]:
                self.data[peptide_sequence][file_name][labeling_group][ion_series] = {}
            # dict_narrowedRT[peptide_sequence][file_name][labeling_group][ion_series] =
            #                                   {'peak_range': peak_range, 'max_delta_mass': max_delta_mass}
            self.data[peptide_sequence][file_name][labeling_group][ion_series] = {'peak_range': peak_range,
                                                                                  'max_delta_mass': max_delta_mass}


########################
# START Functions part #
########################
def sort_dict_keys(dict_arg):
    list_return = list(dict_arg.keys())
    list_return.sort()
    return list_return


def get_ratio(numerator, denominator, ttr=False):
    """
    Get ratio (numerator/denominator)
    :param numerator:    float
    :param denominator:  float
    :param ttr:    'False': simple ratio, 'True': TTR
    :return:    float
    """
    if numerator == 0.0 and denominator == 0.0:
        return 'NA'
    elif numerator == 0.0:
        return 0.0
    elif denominator == 0.0:
        return 'Infinity'
    else:
        if ttr:
            return numerator / (numerator + denominator)
        else:
            return numerator / denominator


def pick_intensity(list_intensities, method):
    """
    Calculate statistics of PRM intensities
    ['MAX','TOP3','SUM','AVG','MED','QMAX','QSUM','QTOP3','QAVG']

    :param list_intensities:    a list of float types
    :param method:              method for picking the intensity
    :return:
    """
    if len(list_intensities) == 0:
        return 0.0

    method = method.lower()
    if method == "avg":
        return numpy.mean(list_intensities)

    elif method == "max":
        return numpy.max(list_intensities)

    elif method == "sum":
        return numpy.sum(list_intensities)

    elif method == 'med':
        return numpy.median(list_intensities)

    elif method == 'top3':
        return pick_top3(list_intensities)

    elif method == 'min':
        return numpy.min(list_intensities)

    list_q = remove_top_bot_quartile(list_intensities)
    if len(list_q) == 0:
        return 0.0
    if method == 'qmax':
        return numpy.max(list_q)

    elif method == 'qsum':
        return numpy.sum(list_q)

    elif method == 'qtop3':
        return pick_top3(list_q)

    elif method == 'qavg':
        return numpy.mean(list_q)

    else:
        print("Chose wrong option, so return SUM by default")
        return numpy.sum(list_intensities)


def remove_top_bot_quartile(list_a):
    """
    Remove 1st and 4th quartile
    :param list_a:  a list
    :return:        float type
    """
    if len(list_a) <= 3:
        return list_a
    else:
        sec_quart = numpy.percentile(list_a, 25)
        thr_quart = numpy.percentile(list_a, 75)
        list_return = []
        for value in list_a:
            if value < sec_quart:
                continue
            if value > thr_quart:
                continue
            list_return.append(value)

        return list_return


def pick_top3(list_a):
    """
    Pick top three values from a list
    :param list_a:  a list
    :return:        float type
    """
    list_a.sort()
    if len(list_a) >= 3:
        return numpy.sum([list_a[-1], list_a[-2], list_a[-3]])
    else:
        return numpy.sum(list_a)


def get_statistics_intensity(list_intensities):
    """
    Calculate statistics of PRM intensities
    :param list_intensities:    a list of float types
    :return:
    """

    return_value = []
    for method in list_intensity_picking_method:
        if len(list_intensities) == 0:
            return_value.append(0.0)
        else:
            return_value.append(pick_intensity(list_intensities, method))
    return return_value


def get_header_intensity_picking_methods(prefix):
    """
    Get headers of intensity picking methods
    :param prefix: prefix of headers
    :return: a string type
    """
    header_intensity = ""
    for method in list_intensity_picking_method:
        tmp_method = "{0}{1}".format(prefix, method)
        header_intensity = append_string(short_string=tmp_method, template=header_intensity, glue='\t')
    return header_intensity


def append_string(short_string, template, glue):
    """
    :param short_string:       short string to append the template
    :param template:    the template to be prolonged
    :param glue:        characters to be located between short and template
    :return:            string type
    """
    if template == '':
        return short_string
    else:
        return template + glue + short_string


def check_file(file_path):
    """
    Check if the file exists
    :param file_path: a file path to be checked
    :return: none
    """
    if not os.path.exists(file_path):
        Message.print_message("Not Found: {0}".format(file_path))


def get_value_by_col_name(col_name, dict_arg, message_on=True):
    if col_name not in dict_arg:
        if message_on:
            print("Unmatched key: {0}".format(col_name))
        return 'NA'
    return dict_arg[col_name]


def get_pcc(x_arg, y_arg):
    """
    Calculate PCC
    :param x_arg: a numeric list
    :param y_arg: a numeric list
    :return: a float number
    """
    # calculate Pearson's correlation coefficient
    # LinregressResult(slope=0.34871776664904164,
    # intercept=0.47516294137304582,
    # rvalue=0.8898089923093081,
    # pvalue=0.0072893465841404638,
    # stderr=0.07997869526654415)
    if len(x_arg) < 3:
        return 1
    if len(x_arg) != len(y_arg):
        print("Can't calculate R2")
        return 0
    lr = stats.linregress(x_arg, y_arg)
    return lr[2]


def get_medians_replicates(x_arg, y_arg, tr_id_list='NA', method_arg="MED"):
    """
    Get median of replicates within same time points
    :param method_arg: Method for quantification
    :param x_arg: time points
    :param y_arg: expression values
    :param tr_id_list: transition ID list
    :return: a dictionary of time points and median expression values
    """
    if tr_id_list == 'NA':
        dict_time2level = {}
        for i in range(len(x_arg)):
            time = x_arg[i]
            level = y_arg[i]
            if level == numpy.inf:
                continue
            if time not in dict_time2level:
                dict_time2level[time] = []
            dict_time2level[time].append(level)
        dict_return = {'x': [], 'y': []}
        for time in sorted(dict_time2level):
            dict_return['x'].append(time)
            dict_return['y'].append(pick_intensity(dict_time2level[time], method_arg))
        return dict_return
    else:
        dict_time2pep2level = {}
        for i in range(len(x_arg)):
            time = x_arg[i]
            level = y_arg[i]
            tr_id_local = tr_id_list[i]
            try:
                # For replicate data
                peptide_local = get_pepseq_from_ion_series(tr_id_local)
                rep_name = extract_rep_name_tr_id(tr_id_local)
                peptide = make_unique_replicate_transition_id(peptide_local, rep_name)
            except IndexError:
                peptide = get_pepseq_from_ion_series(tr_id_local)
            if level == numpy.inf:
                continue
            if time not in dict_time2pep2level:
                dict_time2pep2level[time] = {}
            if peptide not in dict_time2pep2level[time]:
                dict_time2pep2level[time][peptide] = []
            dict_time2pep2level[time][peptide].append(level)
        dict_return = {'x': [], 'y': [], 'pep_avg': {}}
        for time in sorted(dict_time2pep2level):
            list_pep_medians = []
            for pept in sorted(dict_time2pep2level[time]):
                list_pep_medians.append(pick_intensity(dict_time2pep2level[time][pept], method_arg))
            # For median trend line
            dict_return['x'].append(time)
            dict_return['y'].append(pick_intensity(list_pep_medians, "AVG"))
            # For average enrichment of each peptide
            dict_return['pep_avg'][time] = []
            dict_return['pep_avg'][time] = list_pep_medians
        return dict_return


def extract_rep_name_tr_id(tr_id_arg):
    return tr_id_arg.split(delimeter_replicates)[1]


def find_index_in_list(x_arg, list_x):
    """
    Find 'x' is i-th number in a list_x
    :param x_arg: a numeric value
    :param list_x: a list containing x
    :return: integer
    """
    for j in range(len(list_x)):
        x_from_list = list_x[j]
        if x_arg == x_from_list:
            return j
    return 'NA'


def sort_x_y(x_arg, y_arg):
    dict_x2y = {}
    for i in range(len(x_arg)):
        dict_x2y[x_arg[i]] = y_arg[i]
    return_x = []
    return_y = []
    for x_sorted in sorted(dict_x2y):
        return_x.append(x_sorted)
        return_y.append(dict_x2y[x_sorted])
    return {'x': return_x, 'y': return_y}


def get_smooth_line(x_arg, y_arg, num_x=300):
    xnew = np.linspace(numpy.min(x_arg), numpy.max(x_arg), num_x)
    y_smooth = spline(x_arg, y_arg, xnew)
    return xnew, y_smooth


def draw_scatter_plot_med_line(x_arg, y_arg, ion_series, plot_title, destination, list_total_x,
                               x_axis_label, y_axis_label, dpi_user, max_length_fragment=5, 
                               is_trend_line=False, y_upper_lim=-1):
    """
    Draw scatter plot with median line for multiple PRM fragments
    :param x_arg: time points
    :param y_arg: ratio
    :partrendarg: expression value
    :param ion_series: fragment ions
    :param plot_title: peptide sequence with charge
    :param destination: directory path to save an image
    :param list_total_x: a list containing every time points
    :param x_axis_label: X axis label
    :param y_axis_label: Y axis label
    :param dpi_user: DPI
    :param max_length_fragment: maximum length of fragment ions
    :param is_trend_line: True - draw trend line
    :return: none
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    # set y limits
    if y_upper_lim > 0.0:
        ax.set_ylim(0, y_upper_lim)
    ax.set_xlim(numpy.min(list_total_x), numpy.max(list_total_x))
    # Link x and y to ion-series
    dict_ion2xy = {}
    for i in range(len(ion_series)):
        ion_series_name = ion_series[i]
        x1 = x_arg[i]
        y1 = y_arg[i]
        if y1 == numpy.inf:
            continue
        if ion_series_name not in dict_ion2xy:
            dict_ion2xy[ion_series_name] = {'x': [], 'y': []}
        dict_ion2xy[ion_series_name]['x'].append(x1)
        dict_ion2xy[ion_series_name]['y'].append(y1)
    if len(dict_ion2xy) > 20:
        fig.set_size_inches(10, 15)
    # Draw scatter plot with colors and shapes
    # http://www.w3schools.com/tags/ref_colorpicker.asp
    shapes = {'b': 'o', 'y': '*', 'a': 'H', 'x': 'x'}
    # http://matplotlib.org/examples/color/colormaps_reference.html
    spectral_scheme = plt.get_cmap('Spectral')
    c_norm = colors.Normalize(vmin=1, vmax=max_length_fragment * 2)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    list_plots_y = []
    list_plots_b = []
    list_ions = list(dict_ion2xy.keys())
    list_ions.sort()
    # Plot data on x-y plane
    for ion_series_name in sorted(list_ions):
        dict_sorted_xy = sort_x_y(dict_ion2xy[ion_series_name]['x'], dict_ion2xy[ion_series_name]['y'])
        x1 = dict_sorted_xy['x']
        y1 = dict_sorted_xy['y']
        ion_type = ion_series_name[0]
        try:
            ion_number = int(ion_series_name[1])
        except IndexError:
            ion_number = int(ion_series_name.replace(ion_type, ''))
        except ValueError:
            ion_number = int(ion_series_name.replace(ion_type, ''))
        if ion_type == 'a' or ion_type == 'b':
            color_value = scalar_map.to_rgba(ion_number)
            tmp = plt.scatter(x1, y1, marker=shapes[ion_type], color=color_value, s=circle_size, zorder=1)
            list_plots_b.append(tmp)
            plt.plot(x1, y1, zorder=2, lw=1, color=color_value)
        elif ion_type == 'x' or ion_type == 'y':
            color_value = scalar_map.to_rgba(ion_number + max_length_fragment)
            tmp = plt.scatter(x1, y1, marker=shapes[ion_type], color=color_value, s=circle_size, zorder=1)
            list_plots_y.append(tmp)
            plt.plot(x1, y1, zorder=2, lw=1, color=color_value)
    # Legends
    list_plots = list_plots_b + list_plots_y
    plt.legend(list_plots, list_ions, scatterpoints=1, bbox_to_anchor=(1.02, 1), loc=2, borderaxespad=0.,
               ncol=1, fontsize=font_size_ticks_2d)
    # Get median y
    dict_tmp = get_medians_replicates(x_arg, y_arg)
    x_med = dict_tmp['x']
    y_med = dict_tmp['y']
    ax.set_title("%s" % plot_title)
    # If is_trend_line is True, draw scatter plot and median line
    if is_trend_line:
        ax.plot(x_med, y_med, 'r-')
    if flag_figure_size:
        adjust_figure_size(fig)
        ax.get_legend().remove()
    else:
        # axis label and ticks
        ax.set_xlabel(x_axis_label, fontsize=font_size_label_2d)
        ax.set_ylabel(y_axis_label, fontsize=font_size_label_2d)
        ax.tick_params(axis='both', labelsize=font_size_ticks_2d)
    # Save the plot
    img_file_name = change_filename("{0}_fragments.png".format(plot_title))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def check_hdl_ylim(prot_name, ax_arg):
    ax_arg.set_xlim(-1, 75)
    # Set y-axis ticks
    if prot_name.lower() == 'apoe':
        ax_arg.set_ylim(0, 8.5)
        return 3
    elif prot_name.lower() == 'lcat':
        ax_arg.set_ylim(0, 1)
        return 0.6
    elif prot_name.lower() == 'apom':
        ax_arg.set_ylim(0, 2)
    elif prot_name.lower() == 'apoa1':
        ax_arg.set_ylim(0, 1)
        # ax_arg.set_ylim(0, 1)
        return 0.65
    elif prot_name.lower() == 'apoa2':
        ax_arg.set_ylim(0, 0.5)
        # ax_arg.set_ylim(0, 1)
        return 0.33
    elif prot_name.lower() == 'apod':
        ax_arg.set_ylim(0, 2)
    elif prot_name.lower() == 'apoc3':
        ax_arg.set_ylim(0, 2)
    elif prot_name.lower() == 'apoa4':
        ax_arg.set_ylim(0, 3)


def check_hdl_ylim_ion_filtering(prot_name, ax_arg):
    if prot_name == "APOE":
        max_y = 50.1
        # max_y = 20.1  # For LFD-ApoA1 protein data
    elif prot_name == "LCAT":
        max_y = 5.1  # For LFD-ApoA1 protein data
    elif prot_name == "APOA1":
        max_y = 25.1
        # max_y = 4.1  # For LFD-ApoA1 protein data
        # max_y = 3.1  # For Resolution data
    else:
        max_y = 1.0
    ax_arg.set_ylim(0, max_y)
    return max_y


def draw_scatter_plot_med_line_empty(protein, destination):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if flag_figure_size:
        adjust_figure_size(fig)
    else:
        ax.set_title("%s" % protein)
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    plt.text(0.3, 0.8, s="Not Detected")
    # Save the plot
    img_file_name = change_filename("{0}.png".format(protein))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=100)
    fig.clf()
    plt.close(fig)


def draw_scatter_plot_med_line_4protein(x_arg, y_arg, ion_series, plot_title, destination, x_axis_label, y_axis_label,
                                        dpi_user, is_trend_line=False, regression_flag=False, average_peptide_flag=True, 
                                        y_upper_lim=-1):
    """
    Draw scatter plot with median line for multiple PRM fragments
    :param average_peptide_flag: True - draw average point of each peptide, False - draw every data points
    :param regression_flag: True - draw ascending and descending regression line, False - non
    :param is_trend_line: True - draw trend line
    :param dpi_user: DPI
    :param y_axis_label:  Y axis label
    :param x_axis_label:  X axis label
    :param x_arg: time points
    :param y_arg: expression value
    :param ion_series: fragment ions
    :param plot_title: peptide sequence with charge
    :param destination: directory path to save an image
    :return: none
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    """
    # Set up for the paper figures only for HDL analysis
    if is_hdl:
        location_text = check_hdl_ylim(plot_title, ax)
    else:
        location_text = 0.33
    """
    # set y limits
    if y_upper_lim > 0.0:
        ax.set_ylim(0, y_upper_lim)
    location_text = 0.33

    # Link x and y to ion-series
    dict_ion2xy = {}
    list_uniq_peps = []
    for i in range(len(ion_series)):
        ion_series_name = ion_series[i]
        x1 = x_arg[i]
        y1 = y_arg[i]
        if ion_series_name not in dict_ion2xy:
            dict_ion2xy[ion_series_name] = {'x': [], 'y': []}
        dict_ion2xy[ion_series_name]['x'].append(x1)
        dict_ion2xy[ion_series_name]['y'].append(y1)
        # uniq_pep = ion_series_name.split("+")[0]
        try:
            # For replicate data
            peptide_local = get_pepseq_from_ion_series(ion_series_name)
            rep_name = extract_rep_name_tr_id(ion_series_name)
            uniq_pep = make_unique_replicate_transition_id(peptide_local, rep_name)
        except IndexError:
            uniq_pep = get_pepseq_from_ion_series(ion_series_name)
        if uniq_pep not in list_uniq_peps:
            list_uniq_peps.append(uniq_pep)
    # find median points for every time points
    dict_median_points = get_medians_replicates(x_arg, y_arg, tr_id_list=ion_series)
    if average_peptide_flag:
        # Draw peptide average points
        for x_local in dict_median_points['pep_avg']:
            ys = dict_median_points['pep_avg'][x_local]
            for y_local in ys:
                plt.scatter(x_local, y_local, facecolors='none', edgecolors='k', marker='o', s=circle_size, zorder=1)
    else:
        # Extract ion ID
        list_ions = list(dict_ion2xy.keys())
        list_ions.sort()
        # Plot all the data
        for i in range(len(list_ions)):
            ion_series_name = list_ions[i]
            x1 = dict_ion2xy[ion_series_name]['x']
            y1 = dict_ion2xy[ion_series_name]['y']
            plt.scatter(x1, y1, facecolors='none', edgecolors='k', marker='o', s=circle_size, zorder=1)

    # len(set(list_uniq_peps)) --> # of peptides.  If # of peptides is not over 2, skip drawing median points.
    if is_trend_line and len(set(list_uniq_peps)) >= 2:
        # Draw red median line
        x_med = dict_median_points['x']
        y_med = dict_median_points['y']
        # Write median enrichment
        fp_med = open(os.path.join(destination, change_filename("medians_enrichments_{0}.txt".format(plot_title))), "w")
        fp_med.write("X\tY\n")
        for xp, yp in zip(x_med, y_med):
            fp_med.write("{0}\t{1}\n".format(xp, yp))
        fp_med.close()
        ax.scatter(x_med, y_med, marker='o', color='r', s=circle_size - 30, zorder=2)
        if regression_flag:
            #################################
            # START: Linear regression part #
            #################################
            # Get highest enrichment time point before mid_section (22 hour)
            first_section = numpy.min(x_med)
            # These numbers were set up for the XPI paper.
            # If you want to use this part, you may need to change the numbers
            mid_section = 22
            last_section = 70
            highest_enrich = 0
            highest_enrich_time = 0

            if mid_section not in x_med:
                mid_section = 18

            for p in range(len(x_med)):
                time = x_med[p]
                enrich = y_med[p]
                # For ignoring too much initial increase
                # if time < 12: continue
                if time >= mid_section:
                    continue
                if enrich - highest_enrich >= 0:
                    highest_enrich = enrich
                    highest_enrich_time = time
                else:
                    pass
                    # For finding the initial increase
                    # if highest_enrich_time != first_section: break
            # Get values for each sections
            dict_sections = {'1.Ascending': {'x': [], 'y': []},
                             '2.Descending': {'x': [], 'y': []},
                             '3.Convergent': {'x': [], 'y': []}}
            for p in range(len(x_med)):
                time = x_med[p]
                enrich = y_med[p]
                if time == first_section:
                    dict_sections['1.Ascending']['x'].append(time)
                    dict_sections['1.Ascending']['y'].append(enrich)
                elif time == highest_enrich_time:
                    dict_sections['1.Ascending']['x'].append(time)
                    dict_sections['1.Ascending']['y'].append(enrich)
                    dict_sections['2.Descending']['x'].append(time)
                    dict_sections['2.Descending']['y'].append(enrich)
                elif time == mid_section:
                    dict_sections['2.Descending']['x'].append(time)
                    dict_sections['2.Descending']['y'].append(enrich)
                    dict_sections['3.Convergent']['x'].append(time)
                    dict_sections['3.Convergent']['y'].append(enrich)
                elif time == last_section:
                    dict_sections['3.Convergent']['x'].append(time)
                    dict_sections['3.Convergent']['y'].append(enrich)
                else:
                    pass
            # Calculate regression line and slope
            slopes_text = ''
            for section in sorted(dict_sections):
                times = dict_sections[section]['x']
                enriches = dict_sections[section]['y']
                co = numpy.polyfit(times, enriches, 1)
                slope = co[0]
                # p = numpy.poly1d(co)
                # plt.plot(times, p(enriches), 'b-', zorder=3)
                # plt.plot(times, enriches, 'b-', zorder=3)
                slopes_text = append_string("{0}: {1:.4f}".format(section, slope), slopes_text, "\n")
            plt.text(17, location_text, s=slopes_text)
            # plt.text(17, 0.6, s=slopes_text)

            fp = open(os.path.join(destination, change_filename("slopes_{0}.txt".format(plot_title))), "w")
            fp.write("{0}\nHighest enrichment time point\t{1}\n".format(slopes_text, highest_enrich_time))
            fp.close()
            #################################
            # E-N-D: Linear regression part #
            #################################
    ax.set_title("%s" % plot_title)
    if flag_figure_size:
        adjust_figure_size(fig)
    else:
        # axis label and ticks
        ax.set_xlabel(x_axis_label)
        ax.set_ylabel(y_axis_label)
    # Save the plot
    img_file_name = change_filename("{0}.png".format(plot_title))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def count_ions(list_ions_arg, ion_type_arg='b'):
    num_ions = 0
    for ion_series_local in list_ions_arg:
        ion_series_only = ion_series_local.split('-')[1]
        ion_type = ion_series_only[0]
        if ion_type == ion_type_arg:
            num_ions += 1
    return num_ions


def draw_protein_plot_by_peptides(x_arg, y_arg, ion_series, plot_title, destination, x_axis_label, y_axis_label,
                                  dpi_user):
    """
    Draw scatter plot with median line for multiple PRM fragments
    :param y_axis_label: Y axis label
    :param x_axis_label: X axis label
    :param x_arg: time points
    :param y_arg: expression value
    :param ion_series: fragment ions
    :param plot_title: peptide sequence with charge
    :param destination: directory path to save an image
    :param dpi_user: dpi rate
    :param is_hdl: if it is True, apply special y range for HDL paper
    :return: none
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if flag_figure_size:
        adjust_figure_size(fig)
    # Set up for the paper figures only for HDL analysis
    if is_hdl:
        check_hdl_ylim(plot_title, ax)
    xx = x_arg
    # Link x and y to ion-series
    dict_ion2xy = {}
    for i in range(len(ion_series)):
        ion_series_name = ion_series[i]
        x1 = xx[i]
        y1 = y_arg[i]
        if ion_series_name not in dict_ion2xy:
            dict_ion2xy[ion_series_name] = {'x': [], 'y': []}
        dict_ion2xy[ion_series_name]['x'].append(x1)
        dict_ion2xy[ion_series_name]['y'].append(y1)
    # Extract ion ID
    list_ions = list(dict_ion2xy.keys())
    list_ions.sort()
    # For differently colored and shaped scatter plot
    list_plots = []
    # Draw scatter plot with colors and shapes
    # http://www.w3schools.com/tags/ref_colorpicker.asp
    shapes = {'b': 'o', 'y': '*', 'a': 'H', 'x': 'x'}
    # http://matplotlib.org/examples/color/colormaps_reference.html
    spectral_scheme = plt.get_cmap('Set1')
    c_norm = colors.Normalize(vmin=1, vmax=len(list_ions) + 1)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # Plot data on x-y plane
    for i in range(len(list_ions)):
        ion_series_name = list_ions[i]
        dict_sorted_xy = sort_x_y(dict_ion2xy[ion_series_name]['x'], dict_ion2xy[ion_series_name]['y'])
        x1 = dict_sorted_xy['x']
        y1 = dict_sorted_xy['y']
        # plt.scatter(x1, y1, facecolors='none', edgecolors='k')
        ion_series_only = ion_series_name.split('-')[1]
        ion_type = ion_series_only[0]
        if ion_type == 'a' or ion_type == 'b':
            color_value = scalar_map.to_rgba(i)
            tmp = plt.scatter(x1, y1, marker=shapes[ion_type], color=color_value, s=10, zorder=1)
            plt.plot(x1, y1, zorder=2, lw=1, color=color_value)
            list_plots.append(tmp)
        elif ion_type == 'x' or ion_type == 'y':
            color_value = scalar_map.to_rgba(i)
            tmp = plt.scatter(x1, y1, marker=shapes[ion_type], color=color_value, s=10, zorder=1)
            plt.plot(x1, y1, zorder=2, lw=1, color=color_value)
            list_plots.append(tmp)
        else:
            pass
    # Locate legend outside of the plot
    # http://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-the-plot
    # Shrink current axis by 20%
    box_local = ax.get_position()
    ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
    # Legends for colored plots
    plt.legend(list_plots, list_ions, scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1, fontsize=7,
               borderaxespad=0.)
    # axis label and ticks
    ax.set_xlabel(x_axis_label)
    ax.set_ylabel(y_axis_label)
    ax.set_title("%s" % plot_title)
    # Save the plot
    img_file_name = change_filename("{0}.png".format(plot_title))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def draw_median_line_with_error_bar(dict_data_local, plot_title, destination, x_axis_label, y_axis_label, dpi_user,
                                    list_pepseq, r2_threshold=0.0, y_upper_lim=-1):
    """
    Draw scatter plot with median line for multiple PRM fragments
    :param r2_threshold: PCC threshold
    :param dict_data_local: a dictionary = {pep_seq:{x:y}}
    :param plot_title: peptide sequence with charge
    :param destination: directory path to save an image
    :param x_axis_label: x axis label
    :param y_axis_label: y axis label
    :param dpi_user: dpi rate
    :param list_pepseq: a peptide sequence list
    :return:
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if flag_figure_size:
        adjust_figure_size(fig)
    if len(dict_data_local) > 20:
        fig.set_size_inches(10, 15)
    # set y limits
    if y_upper_lim > 0.0:
        ax.set_ylim(0, y_upper_lim)    
    # Color map
    spectral_scheme = plt.get_cmap('jet')
    c_norm = colors.Normalize(vmin=1, vmax=len(list_pepseq) * 2)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # for plot legend
    list_plots = []
    list_found_pep = []
    # Get medians
    dict_x2y = {}
    for pep_seq_local in list_pepseq:
        if pep_seq_local not in dict_data_local:
            continue
        for x_local in sorted(dict_data_local[pep_seq_local]):
            if x_local not in dict_x2y:
                dict_x2y[x_local] = []
            dict_x2y[x_local].extend(dict_data_local[pep_seq_local][x_local])
    # Draw median line
    x_med = []
    y_med = []
    for x_local in sorted(dict_x2y):
        x_med.append(x_local)
        y_med.append(numpy.median(dict_x2y[x_local]))
    """
    # Draw median line
    tmp, = ax.plot(x_med, y_med, color='r')
    list_plots.append(tmp)
    list_found_pep.append("Total Median")
    """
    # Get min and max for each time points
    for i in range(len(sorted(list_pepseq))):
        # Error bar plot
        pep_seq_local = list_pepseq[i]
        if pep_seq_local not in dict_data_local:
            continue
        list_x_local = []
        y = []
        y_min = []
        y_max = []
        y_ste = []
        for x_local in sorted(dict_data_local[pep_seq_local]):
            list_y = dict_data_local[pep_seq_local][x_local]
            list_x_local.append(x_local)
            y.append(numpy.median(list_y))
            y_min.append(numpy.percentile(list_y, 0))
            y_max.append(numpy.percentile(list_y, 100))
            y_ste.append(numpy.std(list_y) / numpy.sqrt(len(list_y)))
        y_med_shortened = y_med[:len(y)]
        if False:
            r2 = get_pcc(y_med_shortened, y)
            if r2 < r2_threshold:
                continue
        color_value = scalar_map.to_rgba(i * 2)
        # line with error bar
        if y_min != y_max:
            """
            # For making error-bar with y max and min
            asymmetric_error = [y_min, y_max]
            ax.errorbar(x, y, yerr=asymmetric_error, fmt='o-', color=color_value, alpha=0.7)
            """
            ax.errorbar(list_x_local, y, yerr=y_ste, fmt='o-', color=color_value, alpha=1)
        else:
            ax.plot(list_x_local, y, color=color_value)
        list_plots.append(Rectangle((0, 0), 1, 1, fc=color_value, alpha=1))
        list_found_pep.append(pep_seq_local)
        """
        # Simple colored line
        tmp, = ax.plot(x, y, color=color_value)
        list_plots.append(tmp)
        list_found_pep.append(pep_seq_local)
        """
    # Locate legend outside of the plot
    box_local = ax.get_position()
    ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
    # Legends for colored plots
    plt.legend(list_plots, list_found_pep, scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1, fontsize=8,
               borderaxespad=0.)
    ax.set_title("%s" % plot_title)
    if flag_figure_size:
        ax.get_legend().remove()
    else:
        # axis label and ticks
        ax.set_xlabel(x_axis_label)
        ax.set_ylabel(y_axis_label)
    # Save the plot
    img_file_name = change_filename("{0}_errors.png".format(plot_title))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def adjust_figure_size(fig):
    # fig.set_size_inches(3, 2)
    fig.set_size_inches(fixed_width, fixed_height)


def draw_filling_plot_by_peptides(dict_data_local, plot_title, destination, x_axis_label, y_axis_label, dpi_user,
                                  list_pepseq, r2_threshold=0.0, y_upper_lim=-1):
    """
    Draw scatter plot with median line for multiple PRM fragments
    :param r2_threshold: PCC threshold
    :param dict_data_local: a dictionary = {pep_seq:{x:y}}
    :param plot_title: peptide sequence with charge
    :param destination: directory path to save an image
    :param x_axis_label: x axis label
    :param y_axis_label: y axis label
    :param dpi_user: dpi rate
    :param list_pepseq: a peptide sequence list
    :return:
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if flag_figure_size:
        adjust_figure_size(fig)
    if len(dict_data_local) > 20:
        fig.set_size_inches(10, 15)
    # set y limits
    if y_upper_lim > 0.0:
        ax.set_ylim(0, y_upper_lim)
    # Color map
    spectral_scheme = plt.get_cmap('jet')
    c_norm = colors.Normalize(vmin=1, vmax=len(list_pepseq) * 2)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # for plot legend
    list_plots = []
    list_found_pep = []
    # If peptide filling plot, transparency is false
    if len(list_pepseq) == 1:
        transparency = 1
    else:
        transparency = 0.4
    # Get min and max for each time points
    for i in range(len(sorted(list_pepseq))):
        pep_seq_local = list_pepseq[i]
        if pep_seq_local not in dict_data_local:
            continue
        list_x_local = []
        y_min = []
        y_max = []
        y_med = []
        for x_local in sorted(dict_data_local[pep_seq_local]):
            list_y = dict_data_local[pep_seq_local][x_local]
            list_x_local.append(x_local)
            y_min.append(numpy.percentile(list_y, 0))
            y_max.append(numpy.percentile(list_y, 100))
            y_med.append(numpy.median(list_y))
        color_value = scalar_map.to_rgba(i * 2)
        if y_min != y_max:
            ax.fill_between(list_x_local, y_min, y_max, facecolor=color_value, alpha=transparency)
        else:
            ax.plot(list_x_local, y_max, color=color_value)
        list_plots.append(Rectangle((0, 0), 1, 1, fc=color_value, alpha=transparency))
        list_found_pep.append(pep_seq_local)
    # Locate legend outside of the plot
    # http://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-the-plot
    # Shrink current axis by 20%
    box_local = ax.get_position()
    ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
    # Legends for colored plots
    plt.legend(list_plots, list_found_pep, scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1, fontsize=8,
               borderaxespad=0.)
    # axis label and ticks
    ax.set_title("%s" % plot_title)
    if flag_figure_size:
        ax.legend().set_visible(False)
    else:
        ax.set_xlabel(x_axis_label)
        ax.set_ylabel(y_axis_label)
    # Save the plot
    img_file_name = change_filename("{0}_filling.png".format(plot_title))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def draw_intensity_ratio_individuals(dict_data_arg, destination, protein, y_axis_label, dpi_user, standard_label,
                                     flag_log, ratio_min_threshold, m0_intensity_min_threshold, ratio_max_threshold,
                                     list_whole_tr_id, flag_show_valid_frags_only=False):
    """
    Draw validation plot via scatter plot of fragment ions in ratio_sum-m0_intensty plane
    :param dict_data_arg: x, y data
    :param destination: the directory for storing a result image
    :param protein: protein name
    :param y_axis_label: y axis label
    :param dpi_user: DPI
    :param standard_label: usually Light
    :param flag_log: True - get log 10 to x and y values
    :param ratio_min_threshold: ratio sum minimum threshold
    :param m0_intensity_min_threshold: m0 intensity threshold
    :param ratio_max_threshold: ratio sum maximum threshold
    :param list_whole_tr_id: transition ID list
    :return: none
    """
    # True: show the filtered ions only
    # False: show all the ions
    # flag_show_valid_frags_only = False

    fig = plt.figure()
    if flag_figure_size:
        adjust_figure_size(fig)
        # fig.set_size_inches(3, 4)     # Used size for the XPI paper
        plt.locator_params(axis='y', nbins=5)
        plt.locator_params(axis='x', nbins=5)
    else:
        if len(dict_data_arg) > 20:
            fig.set_size_inches(8, 12)
        elif len(dict_data_arg) > 40:
            fig.set_size_inches(10, 15)
        elif len(dict_data_arg) > 60:
            fig.set_size_inches(14, 21)
    # Normal tick shape
    ax = fig.add_subplot(111)

    list_ions = sorted(list_whole_tr_id)
    list_ions_4legend = []
    # Color map
    spectral_scheme = plt.get_cmap('jet')
    c_norm = colors.Normalize(vmin=1, vmax=len(list_ions))
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # Plot points
    list_plots = []
    list_x = []
    list_y = []
    for i in range(len(list_ions)):
        ion_local = list_ions[i]
        if ion_local not in dict_data_arg:
            continue
        x_local = dict_data_arg[ion_local]['m0_intensity']
        y = dict_data_arg[ion_local]['y']
        y_m3 = dict_data_arg[ion_local]['y_m3']
        if flag_show_valid_frags_only:
            if x_local < m0_intensity_min_threshold:
                continue
            if y < ratio_min_threshold:
                continue
            if y > ratio_max_threshold:
                continue
        color_value = scalar_map.to_rgba(i)
        filled_markers = ['o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd', 'x']
        used_markers = []
        tp = 0
        for x_tmp, y_tmp, y_m3_tmp in zip(x_local, y, y_m3):
            marker = filled_markers[tp]
            # For enrichment_filter_log option, replace y (ratio) to y_m3 (2HM3 intensities) and get log 10.
            x_selected = x_tmp
            y_selected = y_tmp
            if flag_log:
                x_selected = numpy.log10(x_tmp)
                if y_m3_tmp != 0.0:
                    y_selected = numpy.log10(y_m3_tmp)
                else:
                    y_selected = 0
            tmp = plt.scatter(x_selected, y_selected, marker=marker, color=color_value, s=circle_size_small, zorder=1)
            if tp == 0:
                list_plots.append(tmp)
                list_ions_4legend.append(ion_local)
            tp += 1
            list_x.append(x_selected)
            list_y.append(y_selected)
    # x and y limits
    try:
        min_x = numpy.min(list_x)
        max_x = numpy.max(list_x)
        max_y = numpy.max(list_y)
        min_y = numpy.min(list_y)
        ax.set_xlim(min_x, max_x)
        ax.set_ylim(min_y, max_y)
    except ValueError:
        print("No valid fragment ions passed the noise-filter thresholds: {0}".format(protein))
        return 0

    if not flag_log:
        if not flag_show_valid_frags_only:
            # plot threshold line and add to legend
            if ratio_min_threshold != 0:
                line_threshold, = plt.plot([min_x, max_x], [ratio_min_threshold, ratio_min_threshold], 'r-')
                list_plots.append(line_threshold)
                list_ions_4legend.append("Ratio Sum Threshold")
            if m0_intensity_min_threshold != 0:
                line_threshold2, = plt.plot([m0_intensity_min_threshold, m0_intensity_min_threshold], [0, max_y], 'b-')
                list_plots.append(line_threshold2)
                list_ions_4legend.append("Intensity Sum Threshold")
            if ratio_max_threshold <= max_y:
                line_threshold, = plt.plot([min_x, max_x], [ratio_max_threshold, ratio_max_threshold], 'y-')
                list_plots.append(line_threshold)
                list_ions_4legend.append("Ratio Sum Max Threshold")
    if not flag_figure_size:
        # Locate legend outside of the plot and shrink current axis by 20%
        box_local = ax.get_position()
        ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
        try:
            plt.legend(list_plots, list_ions_4legend, scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1,
                       fontsize=7, borderaxespad=0.)
        except ValueError:
            pass
        # axis label and ticks
        if flag_log:
            ax.set_xlabel("Log10(Intensity {0})".format(standard_label))
            ax.set_ylabel("Log10(Intensity {0})".format(y_axis_label.split("(")[1].split("/")[0].split("+")[0]))
        else:
            ax.set_xlabel("Intensity {0}".format(standard_label))
            ax.set_ylabel("{0}".format(y_axis_label))
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        ax.set_title("{0}".format(protein))
    # Save the plot
    if flag_log:
        out_log = "_log10"
    else:
        out_log = ""
    img_file_name = change_filename("{0}_intensity_vs_ratio{1}.png".format(protein, out_log))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def draw_intensity_ratio_plot(dict_data_arg, destination, protein, y_axis_label, dpi_user, standard_label, flag_log,
                              ratio_sum_threshold, intensity_sum_threshold, ratio_sum_max_threshold, list_whole_tr_id,
                              flag_show_valid_frags_only=False):
    """
    Draw validation plot via scatter plot of fragment ions in ratio_sum-m0_intensty plane
    :param dict_data_arg: x, y data
    :param destination: the directory for storing a result image
    :param protein: protein name
    :param y_axis_label: y axis label
    :param dpi_user: DPI
    :param standard_label: usually Light
    :param flag_log: True - get log 10 to x and y values
    :param ratio_sum_threshold: ratio sum minimum threshold
    :param intensity_sum_threshold: m0 intensity threshold
    :param ratio_sum_max_threshold: ratio sum maximum threshold
    :param list_whole_tr_id: transition ID list
    :return: none
    """
    fig = plt.figure()
    if flag_figure_size:
        adjust_figure_size(fig)
        # fig.set_size_inches(3, 4)     # Used size for the XPI paper
        # fig.set_size_inches(6.5, 8)   # Used size for the XPI paper
        plt.locator_params(axis='y', nbins=5)
        plt.locator_params(axis='x', nbins=5)
    else:
        if len(dict_data_arg) > 20:
            fig.set_size_inches(8, 12)
        elif len(dict_data_arg) > 40:
            fig.set_size_inches(10, 15)
        elif len(dict_data_arg) > 60:
            fig.set_size_inches(14, 21)
    """
    # To make the ticks out and ticklabel inside
    ax = axisartist.Subplot(fig, 111)
    fig.add_subplot(ax)
    ax.axis[:].invert_ticklabel_direction()
    """
    # Normal tick shape
    ax = fig.add_subplot(111)
    list_ions = sorted(list_whole_tr_id)
    list_ions_4legend = []
    # Color map
    spectral_scheme = plt.get_cmap('jet')
    """
    # Used for the XPI paper, not now
    if is_hdl and protein == "APOA1":
        spectral_scheme = plt.get_cmap('BrBG')
    else:
        spectral_scheme = plt.get_cmap('jet')
    """
    c_norm = colors.Normalize(vmin=1, vmax=len(list_ions))
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    
    # Only for Sasha's HDL study to match APOA1 ion colors used for the XPI paper
    dict_tr2color_ss = {}
    if is_hdl and protein == "APOA1":
        for row in csv.DictReader(open("./Accessories/apoa1_paper_color_idx.csv")):
            rgba = [float(row["rgba_1"]), float(row["rgba_2"]), float(row["rgba_3"]), float(row["rgba_4"])]
            dict_tr2color_ss[row["transition_id"]] = numpy.array(rgba)
            dict_tr2color_ss["{0}_{1}".format(row["transition_id"],row["rep_num"])] = numpy.array(rgba)
    # Plot points
    list_plots = []
    list_x = []
    list_y = []
    for i in range(len(list_ions)):
        ion_local = list_ions[i]
        if ion_local not in dict_data_arg:
            continue
        # Get sum of m0, m3 and ratio detected in mzML files.
        x_local = numpy.sum(dict_data_arg[ion_local]['m0_intensity'])
        y = numpy.sum(dict_data_arg[ion_local]['y'])
        y_m3 = numpy.sum(dict_data_arg[ion_local]['y_m3'])
        flag_sum_intensity = False
        flag_sum_ratio = False
        flag_sum_ratio_max = False
        if flag_show_valid_frags_only:
            if x_local < intensity_sum_threshold:
                flag_sum_intensity = True
            if y < ratio_sum_threshold:
                flag_sum_ratio = True
            if ratio_sum_max_threshold != default_ratio_max_threshold and y > ratio_sum_max_threshold:
                flag_sum_ratio_max = True
        # For enrichment_filter_log option, replace y (ratio) to y_m3 (2HM3 intensities)
        # And get log 10.
        if flag_log:
            y = y_m3
            if y <= 0.0:
                y = 1.0
            if x_local <= 0.0:
                x_local = 1.0
            y = numpy.log10(y)
            x_local = numpy.log10(x_local)
        list_x.append(x_local)
        list_y.append(y)
        if is_hdl and protein == "APOA1":
            if ion_local in dict_tr2color_ss:
                color_value = dict_tr2color_ss[ion_local]
            else:
                color_value = scalar_map.to_rgba(i)
        else:
            color_value = scalar_map.to_rgba(i)
        tmp = plt.scatter(x_local, y, marker='o', color=color_value, s=circle_size_small, zorder=1)
        list_plots.append(tmp)
        list_ions_4legend.append(ion_local)
    # x and y limits
    try:
        min_x = numpy.min(list_x)
        max_x = numpy.max(list_x)
        max_y = numpy.max(list_y)
        min_y = numpy.min(list_y)
        ax.set_xlim(min_x, max_x)
        ax.set_ylim(min_y, max_y)
    except ValueError:
        print("No valid fragment ions passed the noise-filter thresholds: {0}".format(protein))
        return 0

    ### For the paper.
    ### Should be removed in the distribution version
    # min_y, max_y = ax.get_ylim()
    # min_x, max_x = ax.get_xlim()
    """
    # Temporary code for Sasha
    if is_hdl and protein == "APOA1":
        ax.set_ylim(0, 4.1)
        ax.set_xlim(0, 31000000000)
    """
    if is_hdl:
        if not flag_log:
            max_y = check_hdl_ylim_ion_filtering(protein, ax)
    if flag_log:
        # calculate Pearson's correlation coefficient
        r2 = get_pcc(list_x, list_y)
        # Draw regression line and add to legend
        co = numpy.polyfit(list_x, list_y, 1)
        p = numpy.poly1d(co)
        line_reg, = plt.plot(list_x, p(list_x), 'k-')
        list_plots.append(line_reg)
        list_ions_4legend.append("Regression line (R2={0:.2f})".format(r2))

    if not flag_show_valid_frags_only and not flag_log:
        # plot threshold line and add to legend
        if ratio_sum_threshold != 0:
            line_threshold, = plt.plot([min_x, max_x], [ratio_sum_threshold, ratio_sum_threshold], 'r-')
            list_plots.append(line_threshold)
            list_ions_4legend.append("Ratio Sum Threshold")
        if intensity_sum_threshold != 0:
            line_threshold2, = plt.plot([intensity_sum_threshold, intensity_sum_threshold], [0, max_y], 'b-')
            list_plots.append(line_threshold2)
            list_ions_4legend.append("Intensity Sum Threshold")
        if ratio_sum_max_threshold <= max_y:
            if ratio_sum_max_threshold != default_ratio_max_threshold:
                line_threshold, = plt.plot([min_x, max_x], [ratio_sum_max_threshold, ratio_sum_max_threshold], 'y-')
                list_plots.append(line_threshold)
                list_ions_4legend.append("Ratio Sum Max Threshold")

    if not flag_figure_size:
        # Locate legend outside of the plot and shrink current axis by 20%
        box_local = ax.get_position()
        ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
        plt.legend(list_plots, list_ions_4legend, scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1, fontsize=7,
                   borderaxespad=0.)
        # axis label and ticks
        if flag_log:
            ax.set_xlabel("Log10(Intensity Sum of {0})".format(standard_label))
            ax.set_ylabel("Log10(Intensity Sum of {0})".format(y_axis_label.split("(")[1].split("/")[0]))
        else:
            ax.set_xlabel("Intensity Sum of {0}".format(standard_label))
            ax.set_ylabel("Sum of {0}".format(y_axis_label))
        ax.set_title("{0}".format(protein))
    # Save the plot
    if flag_log:
        out_log = "_log10"
    else:
        out_log = ""
    img_file_name = change_filename("{0}_intensity_vs_ratio{1}.png".format(protein, out_log))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def draw_scatter_plot_reg_line(x_arg, y, transit_id, destination, x_axis_label, y_axis_label, dpi_user, poly_level=1):
    """
    Draw scatter plot with regression line
    :param x_arg: time points
    :param y: expression value
    :param transit_id: PRM ion name
    :param destination: directory path to save an image
    :param y_axis_label: Y label
    :param x_axis_label: X label
    :param dpi_user: DPI
    :param poly_level: polyfit level
    :return: none
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if flag_figure_size:
        adjust_figure_size(fig)
    plt.scatter(x_arg, y, marker='o', s=circle_size)
    # plt.scatter(x, y, facecolors='none', edgecolors='k')
    co = numpy.polyfit(x_arg, y, poly_level)
    p = numpy.poly1d(co)
    # calculate Pearson's correlation coefficient
    r2 = get_pcc(x_arg, y)
    ax.plot(x_arg, p(x_arg), 'r-')
    ax.set_title("%s (R2=%.4f)" % (transit_id, r2))
    # axis label and ticks
    ax.set_xlabel(x_axis_label, fontsize=font_size_label_2d)
    ax.set_ylabel(y_axis_label, fontsize=font_size_label_2d)
    ax.tick_params(axis='both', labelsize=font_size_ticks_2d)
    ax.set_xlim([min(x_arg) * 0.9 - 0.05, max(x_arg) * 1.01])
    # ax.set_ylim([min(y), max(y) * 1.01])
    # save the image
    img_file_name = change_filename("{0:.4f}_{1}.png".format(r2, transit_id))
    img_file_path = os.path.join(destination, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)
    return r2


def change_filename(filename):
    """
    Remove special characters that are not allowed to be file name
    :param filename: a file name
    :return: a string
    """
    for c in r'[]/\;,><&*:%=+@!#^()|?^ ':
        filename = filename.replace(c, '_')
    return filename


def get_ratio_name(dict_run2info):
    """
    Get labeling name saved in PRM Run information
    :param dict_run2info: PRM Run information
    :return: a string
    """
    return_name = ''
    for prm_run in dict_run2info:
        r_name = dict_run2info[prm_run]['RatioName']
        if return_name == '':
            return_name = r_name
        elif return_name != r_name:
            print("Found multiple labeling pairs: {0}\t{1}".format(r_name, return_name))
    return return_name


def check_intensity_picking_method(method):
    """
    Check intensity-picking method
    :param method: selected method name
    :return: a boolean
    """
    if method.lower() == 'all':
        return False
    elif method.upper() in list_intensity_picking_method:
        return False
    else:
        return True


def check_aa_in_sequence(list_aa, sequence):
    """
    Check amino acids in given peptide sequence
    :param list_aa: a list containing amino acids
    :param sequence: a peptide sequence
    :return: a boolean
    """
    for aa in list_aa:
        # '-' ignores this check process
        if aa != "-":
            if not re.search(pattern=aa, string=sequence):
                return False
    return True


def rep(string, times):
    """
    replicate 'string' 'times' times
    :param string: string
    :param times: integer
    :return: a list
    """
    list_return = []
    for i in range(0, times):
        list_return.append(string)
    return list_return


def count_outliers(list_data, threshold=1.96):
    """
    Count outliers that is over the threshold
    :param list_data: a list containing float numbers
    :param threshold: default is 1.96
    :return: integer number
    """
    try:
        list_zscores = list(stats.zscore(list_data))
        num_outliers = 0
        for zscore in list_zscores:
            if zscore >= threshold:
                num_outliers += 1
        return num_outliers
    except TypeError:
        return 0


def get_zscores(list_data):
    """
    Count outliers that is over the threshold
    :param list_data: a list containing float numbers
    :return: integer number
    """
    try:
        return list(numpy.absolute(stats.zscore(list_data)))
    except TypeError:
        return 0


def get_zscore_for_all_methods(data_directory, config_xpiviz, ttr):
    """
    Get the number of outliers (Z-score >= 1.96) for every intensity-picking methods
    :param ttr: True - calculate TTR
    :param data_directory: Data directory containing XPI output
    :param config_xpiviz: An object of XPIViz_config class
    :return: none
    """
    # parse xpi_4check file
    file_all_methods_path = os.path.join(data_directory, file_check)
    check_file(file_all_methods_path)
    ratio_name = get_ratio_name(config_xpiviz.data)
    xpi_4check = XPI4Check(file_all_methods_path, ratio_name, config_xpiviz.standard_label, ttr)
    # Get the number of outliers for every quantification methods
    dict_tr_id_method_zscore = {}
    dict_method_zscore = {}
    for method in list_intensity_picking_method:
        # Collect ratios from all the data
        dict_observed_ratio = xpi_4check.data
        dict_method_zscore[method] = []
        dict_pep2ratios = {}
        for tr_id in dict_observed_ratio[method]:
            pep_seq = tr_id.split("+")[0]
            for prm_run in dict_observed_ratio[method][tr_id]:
                ratio = dict_observed_ratio[method][tr_id][prm_run][ratio_name]
                if pep_seq not in dict_pep2ratios:
                    dict_pep2ratios[pep_seq] = []
                dict_pep2ratios[pep_seq].append(ratio)
        # Get the number of outliers for every peptides
        for pep_seq in dict_pep2ratios:
            num_outliers = get_zscores(dict_pep2ratios[pep_seq])
            if num_outliers == 0:
                continue
            dict_method_zscore[method].extend(num_outliers)
            # make dictionary for printing out PCC results
            if pep_seq not in dict_tr_id_method_zscore:
                dict_tr_id_method_zscore[pep_seq] = {}
            if method not in dict_tr_id_method_zscore[pep_seq]:
                dict_tr_id_method_zscore[pep_seq][method] = {}
            dict_tr_id_method_zscore[pep_seq][method] = num_outliers
    return {'method2pcc': dict_method_zscore, 'tr_id_pcc': dict_tr_id_method_zscore}


def get_pcc_for_all_methods(data_directory, config_xpiviz, ttr, dict_tr2sum_ri):
    """
    Get PCC for every intensity-picking methods
    :param dict_tr2sum_ri: a dictionary containing transition id 2 sum ratio and sum intensity
    :param ttr: True - TTR
    :param data_directory: Data directory containing XPI output
    :param config_xpiviz: An object of XPIViz_config class
    :return: none
    """
    # parse xpi_4check file
    file_all_methods_path = os.path.join(data_directory, file_check)
    check_file(file_all_methods_path)
    xpi_4check = XPI4Check(file_all_methods_path, get_ratio_name(config_xpiviz.data), config_xpiviz.standard_label, ttr)
    # Get Pearson correlation coefficients (PCC) for each of intensity-picking methods
    dict_tr_id_method_pcc = {}
    dict_method_pcc = {}
    for method in list_intensity_picking_method:
        dict_observed_ratio = xpi_4check.data[method]
        standard_curve_obj = Scatterplot(dict_observed_ratio=dict_observed_ratio,
                                         dict_run2info=config_xpiviz.data,
                                         data_directory=data_directory,
                                         is_log=config_xpiviz.flag_log,
                                         dpi=config_xpiviz.dpi)
        standard_curve_obj.process_ratio_data(dict_tr2sum_ri, mode='pcc')
        # Get PCC for each of transition PRM ions
        dict_method_pcc[method] = []
        for tr_id in standard_curve_obj.dict_transit2pcc:
            pcc = float(standard_curve_obj.dict_transit2pcc[tr_id]['PCC'])
            pcc = abs(pcc)
            dict_method_pcc[method].append(pcc)
            # make dictionary for printing out PCC results
            if tr_id not in dict_tr_id_method_pcc:
                dict_tr_id_method_pcc[tr_id] = {}
            if method not in dict_tr_id_method_pcc[tr_id]:
                dict_tr_id_method_pcc[tr_id][method] = {}
            dict_tr_id_method_pcc[tr_id][method] = pcc
    return {'method2pcc': dict_method_pcc, 'tr_id_pcc': dict_tr_id_method_pcc}


def draw_box_plot(data_directory, img_path, config_xpiviz, ttr, dict_tr2sum_ri, metrics='pcc'):
    """
    Draw box-plots and write pcc for every intensity-picking methods
    :param metrics: pcc or zscore
    :param dict_tr2sum_ri: a dictionary
    :param ttr: True - TTR
    :param img_path: image path to generate
    :param data_directory: Data directory containing XPI output
    :param config_xpiviz: An object of XPIViz_config class
    :return: none
    """
    if metrics == 'pcc':
        dict_data = get_pcc_for_all_methods(data_directory, config_xpiviz, ttr, dict_tr2sum_ri)
        dict_method_pcc = dict_data['method2pcc']
        dict_tr_id_method_pcc = dict_data['tr_id_pcc']
    else:
        dict_data = get_zscore_for_all_methods(data_directory, config_xpiviz, ttr)
        dict_method_pcc = dict_data['method2pcc']
        dict_tr_id_method_pcc = dict_data['tr_id_pcc']
    # calculate median pcc
    dict_median_pcc_method = {}
    for method in list_intensity_picking_method:
        pcc_median = numpy.median(dict_method_pcc[method])
        if pcc_median in dict_median_pcc_method:
            dict_median_pcc_method[pcc_median] = "{0}{1}{2}".format(dict_median_pcc_method[pcc_median],
                                                                    delimiter_residue, method)
        else:
            dict_median_pcc_method[pcc_median] = method
    # Get data list for box-plots
    list_data4box = []
    list_method_sorted = []
    list_pcc_sorted = list(dict_median_pcc_method.keys())
    if metrics == 'pcc':
        list_pcc_sorted.sort(reverse=True)
    else:
        list_pcc_sorted.sort()
    header = "Transition_ID"  # for PCC output
    for median_pcc in list_pcc_sorted:
        # Get method labels sorted
        method = dict_median_pcc_method[median_pcc]
        list_methods = method.split(":")
        for q_meth in list_methods:
            list_method_sorted.append(q_meth)
            list_data4box.append(dict_method_pcc[q_meth])
            # Print out median PCC for each quantification method
            # print("{0}\t{1}".format(q_meth, median_pcc))
            header = append_string(short_string=q_meth, template=header, glue=delimiter_tab)
    # Draw Box-plot
    # http://matplotlib.org/api/pyplot_api.html
    fig = plt.figure()
    ax = fig.add_subplot(111)
    bp = plt.boxplot(list_data4box, labels=list_method_sorted, showmeans=True)
    if metrics == 'pcc':
        ax.set_title("Pearson Correlation Coefficient")
    elif metrics == 'zscore':
        ax.set_title("Z-Score")
    else:
        Message.print_message("Wrong option. Choose 'pcc' or 'zscore': {0}".format(metrics))
    plt.setp(bp['boxes'], color='black')
    plt.setp(bp['whiskers'], color='red')
    plt.setp(bp['fliers'], color='royalblue', marker='+')
    if metrics == 'pcc':
        plt.ylabel('Pearson Correlation Coefficient', fontsize=font_size_label_2d)
    else:
        ax.set_ylim(-1, numpy.max(list_data4box))
        plt.ylabel('Z-score', fontsize=font_size_label_2d)
    img_file_path = os.path.join(img_path, "Box_plot_{0}.png".format(metrics))
    plt.savefig(img_file_path, dpi=config_xpiviz.dpi)
    fig.clf()
    plt.close(fig)
    # print("Box_plot_all_methods.png was generated.")

    # Print PCC output
    pcc_out_file = os.path.join(img_path, "Median_{0}.txt".format(metrics))
    fp_out = open(pcc_out_file, "w")
    fp_out.write(header + "\n")
    for tr_id in dict_tr_id_method_pcc:
        out_string = tr_id
        for median_pcc in list_pcc_sorted:
            method = dict_median_pcc_method[median_pcc]
            list_methods = method.split(":")
            for q_meth in list_methods:
                if q_meth in dict_tr_id_method_pcc[tr_id]:
                    if isinstance(dict_tr_id_method_pcc[tr_id][q_meth], list):
                        pcc = "{0:.4f}".format(pick_intensity(dict_tr_id_method_pcc[tr_id][q_meth], 'avg'))
                    else:
                        pcc = "{0:.4f}".format(dict_tr_id_method_pcc[tr_id][q_meth])
                else:
                    pcc = 'NA'
                out_string = append_string(short_string=pcc, template=out_string, glue=delimiter_tab)
        fp_out.write(out_string + "\n")
    fp_out.close()


def find_data_directory_from_time_file(data_directory):
    """
    Get the path of config file for XPIQuant script
    :param data_directory:  data directory containing XPI outputs
    :return: a string
    """
    path_time_file = os.path.join(data_directory, file_time)
    fp = open(path_time_file, "rU")
    for data_line in fp:
        list_data = data_line.split('\t')
        if list_data[0] == colname_data_directory:
            return list_data[1].strip()
    return 'NA'


def draw_3d_plot_all(dict_peptide2data, data_directory, ms2_window, plot_title, dpi_user, list_label_3d, dict_peak,
                     file_name, is_z_same=True):
    """
    Draw 3d plots for different labellings of one peptide
    :param dict_peptide2data: dict_ion2data: labeling_group -> ion_series -> {'intensity': [], 'mz_diff': [], 'rt': []}
    :param data_directory: data directory
    :param ms2_window: MS/MS window
    :param plot_title: Title of the plot
    :param is_z_same: True - Make z axis to have same scalibility
    :param dpi_user: DPI
    :param file_name: File name
    :param list_label_3d: labels
    :param dict_peak: peak information
    :return: none
    """
    # make list_ion_series from all the labeling groups
    list_ion_series = []
    for lab in list_label_3d:
        if lab in dict_peptide2data:
            for ion_series in dict_peptide2data[lab]:
                if ion_series not in list_ion_series:
                    list_ion_series.append(ion_series)
    # Sort lists to make organized plot
    list_label_3d.sort()
    list_ion_series.sort()
    # True: small figure for the paper
    if flag_figure_size:
        grid_for_3d_plots = 10
        grid_for_color_bar = 2
        grid_for_space = 2
        row_number = 4
        col_number = 3
        font_size_label = 10
        font_size_ticks = 10
    else:
        # basic plot set up
        grid_for_3d_plots = 50
        grid_for_color_bar = 5
        grid_for_space = 5
        row_number = 15
        col_number = 10
        font_size_label = 30
        font_size_ticks = 20
    # Calculate total y grid number
    grid_y = len(list_label_3d) * grid_for_3d_plots + grid_for_color_bar + grid_for_space * 3
    #  Set figure size and location of 3d plots for each labellings
    fig = plt.figure(figsize=(len(list_label_3d) * row_number, col_number))
    grid_ax = gs.GridSpec(1, grid_y)
    color_bar = fig.add_subplot(grid_ax[0, (grid_for_color_bar + grid_for_space) * -1:grid_for_space * -1])
    # Color scales
    spectral_scheme = plt.get_cmap('winter')
    c_norm = colors.Normalize(vmin=0, vmax=ms2_window)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # Draw 3d plots of every ion_series and every labellings for one peptide
    total_rt = ''
    max_z = 0.0
    list_axes = []
    count_filtered = {}
    for i in range(len(list_label_3d)):
        label_name = list_label_3d[i]
        init_grid = i * (grid_for_3d_plots + grid_for_space)
        ax = fig.add_subplot(grid_ax[0, init_grid:init_grid + grid_for_3d_plots], projection='3d')
        list_axes.append(ax)
        # set ticks
        list_x_tick = []
        list_y_tick = [0]
        fig.suptitle(plot_title.replace('_', ' (') + ")", fontsize=font_size_label)
        # set font sizes
        matplotlib.rc('font', size=font_size_ticks)
        ax.tick_params(axis='y', labelsize=font_size_ticks-2)
        # Sometimes default font sizes are applied to x and z axis
        # These are redundant, but to make it sure for making x and z axis ticks
        ax.tick_params(axis='x', labelsize=font_size_ticks)
        ax.tick_params(axis='z', labelsize=font_size_ticks)
        ax.zaxis.offsetText.set_fontsize(font_size_ticks)
        plt.locator_params(nbins=6)
        ##############################
        # Change angles of the 3d plot
        # ax.azim, ax.elev = (115, 10)
        ##############################
        # x axis label
        ax.set_xlabel('RT (min.)', fontsize=font_size_label)
        ax.xaxis.labelpad = 30
        # z axis label
        # ax.set_zlabel('Intensity', fontsize=font_size_label)
        ax.zaxis.labelpad = 30
        # Z axis ticks will be shown like scientific numbers
        ax.ticklabel_format(style='sci', axis='z', scilimits=(0, 0))
        # Without this line some of x axis will be scientific numbers
        ax.ticklabel_format(axis='x', useOffset=False)
        # plot tile
        ax.set_title("{0}".format(label_name), fontsize=font_size_label)
        # Check if there are ion-series...
        list_rt_all = []
        for ion_series in list_ion_series:
            list_rt = []
            if label_name in dict_peptide2data and ion_series in dict_peptide2data[label_name]:
                list_rt = dict_peptide2data[label_name][ion_series]['rt']
            list_rt_all.extend(list_rt)
        # If there is no valid ion-series, skip drawing plots
        if len(list_rt_all) == 0:
            ax.text(x=0.3, y=0.3, z=0.5, s="Not Detected")
        elif label_name not in dict_peak:
            ax.text(x=0.3, y=0.3, z=0.5, s="Not Detected")
        else:
            rt_4blank = numpy.mean(list_rt_all)
            # plot blank for making space at the first position of ion series
            ax.plot([rt_4blank, rt_4blank], [0, 0], [0, 0], '-', alpha=alpha)
            for j in range(len(list_ion_series)):
                ion_series = list_ion_series[j]
                if ion_series in dict_peptide2data[label_name]:
                    # retrieve data from scan data
                    list_rt = dict_peptide2data[label_name][ion_series]['rt']
                    list_intensity = dict_peptide2data[label_name][ion_series]['intensity']
                    list_mz_diff = dict_peptide2data[label_name][ion_series]['mz_diff']
                    total_rt = dict_peptide2data[label_name][ion_series]['total_rt']
                    if len(total_rt) > 0:
                        total_rt = total_rt[0]
                    # retrieve data from peak selection data
                    if ion_series in dict_peak[label_name]:
                        peak_range = dict_peak[label_name][ion_series]['peak_range']
                        list_rt_selected = peak_range.split(delimiter_residue)
                        rt_start = float(list_rt_selected[0])
                        rt_end = float(list_rt_selected[2])
                        max_delta_mass = dict_peak[label_name][ion_series]['max_delta_mass']
                    else:
                        rt_start = total_rt.split('_')[0]
                        rt_end = total_rt.split('_')[1]
                        max_delta_mass = ms2_window
                else:
                    list_rt = []
                    list_intensity = []
                    list_mz_diff = []
                    rt_start = 0
                    rt_end = 0
                    max_delta_mass = ms2_window
                list_x_tick.extend(list_rt)
                list_y_tick.append(j + 1)
                count_filtered[ion_series] = 0
                for rt, intensity, mz_diff in zip(list_rt, list_intensity, list_mz_diff):
                    # Filter out ion intensities by retention time and delta m/z
                    if check_range(rt_start, rt_end, rt):
                        continue
                    if mz_diff > max_delta_mass:
                        continue
                    color_value = scalar_map.to_rgba(mz_diff)
                    ax.plot([rt, rt], [j + 1, j + 1], [0, intensity], '-', color=color_value, alpha=alpha)
                    count_filtered[ion_series] += 1
            # If no peak passed mz_diff and rt filters, print NA
            num_intensities = 0
            for ion_series in count_filtered:
                num_intensities += count_filtered[ion_series]
            if num_intensities == 0:
                ax.text(x=rt_4blank, y=int(len(list_ion_series) / 2), z=0, s="No Valid Peaks")
            # plot blank for making space at the end of ion series
            ax.plot([rt_4blank, rt_4blank], [j + 2, j + 2], [0, 0], '-', alpha=alpha)
            list_y_tick.append(j + 2)
            # y axis label
            # Since y axis requires ion_series information, this is separated from other axis parameter sets
            # ax.set_ylabel('Ion Series', fontsize=font_size_label)
            ax.yaxis._axinfo['label']['space_factor'] = 2
            ax.set_yticks(list_y_tick)
            ax.set_yticklabels([''] + list_ion_series + [''])
            # RT lower and upper limit
            low = float(total_rt.split("_")[0])
            high = float(total_rt.split("_")[1])
            ax.set_xlim([low, high])
            # Get Maximum intensity for making same z axis scale
            lim = ax.get_zlim()
            max_z_tmp = lim[1]
            if max_z_tmp > max_z:
                max_z = max_z_tmp
    # If is_z_same is True, make every plots to have same intensity ticks
    if is_z_same:
        for ax in list_axes:
            ax.set_zlim([0, max_z])
    # Sometimes we need tight_layout() for better margin control, but it usualy returns bunch of warnings.
    plt.tight_layout()
    # Plot color bar
    cb.ColorbarBase(color_bar, norm=c_norm, cmap=spectral_scheme)
    color_bar.set_ylabel('Delta m/z [Da]')
    # Save the image to the designated directory
    img_file_path = os.path.join(data_directory, "{0}_{1}.png".format(plot_title.replace(' ', '_'),
                                                                      os.path.basename(file_name)))
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)
    return count_filtered


def draw_2d_mass_profile_for_two(dict_peptide2data, data_directory, plot_title, dpi_user):
    """
    Draw 2d mass profile for different labelings
    :param dict_peptide2data: dict_ion2data: labeling_group -> {'intensity': [], 'mz_diff': [], 'rt': []}
    :param data_directory: data directory
    :param plot_title: Title of the plot
    :param dpi_user: DPI
    :return: none
    """
    list_labels = list(dict_peptide2data.keys())
    list_labels.sort()
    fig, ax1 = plt.subplots()
    if flag_figure_size:
        adjust_figure_size(fig)
    else:
        fig.set_size_inches(5, 3)
    # fig.suptitle(plot_title.replace("_", " "))
    ax1.set_xlabel('Retention Time')
    ax1.ticklabel_format(axis='x', useOffset=False)

    # color
    color_label1 = 'orange'
    color_label2 = 'gray'

    dict_peptide2data_cleaned = {}
    for label_local in dict_peptide2data:
        dict_peptide2data_cleaned[label_local] = {'rt': [], 'intensity': [], 'mz_diff': []}
        rt_start = dict_peptide2data[label_local]['rt_start']
        rt_end = dict_peptide2data[label_local]['rt_end']
        max_delta_mass = dict_peptide2data[label_local]['max_del']
        zipped_array = zip(dict_peptide2data[label_local]['rt'],
                           dict_peptide2data[label_local]['intensity'],
                           dict_peptide2data[label_local]['mz_diff'])
        for rt_local, int_local, delt_mz_local in zipped_array:
            if check_range(rt_start, rt_end, rt_local):
                continue
            if delt_mz_local > max_delta_mass:
                continue
            dict_peptide2data_cleaned[label_local]['rt'].append(rt_local)
            dict_peptide2data_cleaned[label_local]['intensity'].append(int_local)
            dict_peptide2data_cleaned[label_local]['mz_diff'].append(delt_mz_local)

    # For first label group
    label1 = list_labels[0]
    ax1.bar(dict_peptide2data_cleaned[label1]['rt'], dict_peptide2data_cleaned[label1]['intensity'], 0.01,
            color=color_label1, edgecolor='none', alpha=alpha)
    ax1.set_ylabel("Intensity ({0})".format(label1), color=color_label1)
    ax1.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    for tl in ax1.get_yticklabels():
        tl.set_color(color_label1)

    # For first label group
    label2 = list_labels[1]
    ax2 = ax1.twinx()
    ax2.bar(dict_peptide2data_cleaned[label2]['rt'], dict_peptide2data_cleaned[label2]['intensity'], 0.01,
            color=color_label2, edgecolor='none', alpha=alpha)
    ax2.set_ylabel("Intensity ({0})".format(label2), color=color_label2)
    ax2.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    for tl in ax2.get_yticklabels():
        tl.set_color(color_label2)

    """
    # RT lower and upper limit
    try:
        total_rt = dict_peptide2data[label1]['total_rt'][0]
        low = float(total_rt.split("_")[0])
        high = float(total_rt.split("_")[1])
        ax1.set_xlim([low, high])
        ax2.set_xlim([low, high])
    except IndexError:
        pass
    """
    # Save the plot
    img_file_name = change_filename("{0}.png".format(plot_title))
    img_file_path = os.path.join(data_directory, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)


def draw_line_graphs(dict_peptide2data, data_directory, ms2_window, plot_title, dpi_user, list_labels, dict_peak,
                     file_name, all_ion_series):
    """
    Draw line graphs for different labellings of one peptide
    :param dict_peptide2data: dict_ion2data: labeling_group -> ion_series -> {'intensity': [], 'mz_diff': [], 'rt': []}
    :param data_directory: data directory
    :param ms2_window: MS/MS window
    :param plot_title: Title of the plot
    :param dpi_user: DPI
    :param file_name: File name
    :param list_labels: labels
    :param dict_peak: peak information
    :return: a dictionary containing ion -> labeling -> qunatification
    """
    # Sort lists to make organized plot
    list_labels.sort()
    all_ion_series.sort()
    num_labelings = len(list_labels)
    # Start drawing
    fig = plt.figure()
    span_for_plots = 30
    # color sets
    spectral_scheme = plt.get_cmap('jet')
    c_norm = colors.Normalize(vmin=1, vmax=len(all_ion_series))
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    # To return the refined quantification
    dict_return = {}
    # Draw smooth line graphs of every ion_series and every labellings for one peptide
    for i in range(len(list_labels)):
        ax = plt.subplot2grid((num_labelings, num_labelings + span_for_plots), (i, 0), colspan=span_for_plots)
        ax.tick_params(axis='both', labelsize=font_size_ticks_2d)
        label_group = list_labels[i]
        for j in range(len(all_ion_series)):
            ion_series = all_ion_series[j]
            if ion_series not in dict_peptide2data[label_group]:
                continue
            total_rt = dict_peptide2data[label_group][ion_series]['total_rt']
            dict_return[ion_series] = {}
            if label_group in dict_return[ion_series]:
                Message.print_message("Found redundant ion series in the quantification process")
            dict_return[ion_series][label_group] = []
            if len(total_rt) == 0:
                dict_return[ion_series][label_group] = get_statistics_intensity([])
            else:
                rt_selected = dict_peak[label_group][ion_series]['peak_range'].split(delimiter_residue)
                rt1 = float(rt_selected[0])
                rt2 = float(rt_selected[2])
                zipped_array = zip(dict_peptide2data[label_group][ion_series]['rt'],
                                   dict_peptide2data[label_group][ion_series]['intensity'])
                x_rt = []
                y_int = []
                for r,i in zipped_array:
                    if rt1 <= r <= rt2:
                        x_rt.append(r)
                        y_int.append(i)
                # Quantify ions
                dict_return[ion_series][label_group] = get_statistics_intensity(y_int)
                # Fill outside of the selected RT with zeros
                # Without the filling, line graphs will be funky
                list_total_rt = total_rt[0].split('_')
                for rt in numpy.arange(float(list_total_rt[0]),float(list_total_rt[1]),0.01):
                    if rt1 <= rt <= rt2:
                        continue
                    else:
                        x_rt.append(rt)
                        y_int.append(0.0)
                # For coloring
                color_value = scalar_map.to_rgba(j)
                # Drwaing raw intensity lines
                # ax.plot(x_rt, y_int, linewidth=.8, color=color_value, label=ion_series)
                # For drawing smooth lines
                smoothed_data = lowess(y_int, x_rt, frac=0.1, it=0)
                rt_lowess = smoothed_data[0:len(smoothed_data), 0]
                intensity_lowess = smoothed_data[0:len(smoothed_data), 1]
                intensity_lowess[intensity_lowess < 0] = 0
                ax.plot(rt_lowess, intensity_lowess, linewidth=.8, color=color_value, label=ion_series)
    # Locate legend outside of the plot and shrink current axis by 20%
    box_local = ax.get_position()
    ax.set_position([box_local.x0, box_local.y0, box_local.width * 0.8, box_local.height])
    ax.legend(scatterpoints=1, bbox_to_anchor=(1.01, 1), loc=2, ncol=1, fontsize=10, borderaxespad=0.)
    ax.set_xlabel("RT")
    ax.set_ylabel("Normalized Intensity")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    ax.set_title(plot_title)
    # ax.set_ylim([0, 0.01])
    img_file_path = os.path.join(data_directory, "{0}_{1}.png".format(plot_title.replace(' ', '_'), os.path.basename(file_name)))
    fig.set_size_inches(10,7)
    plt.savefig(img_file_path, dpi=dpi_user)
    fig.clf()
    plt.close(fig)
    return dict_return

def draw_2d_mass_profile_multi(dict_peptide2data, data_directory, plot_title, dpi_user, ms2_window, list_label_groups):
    """
    Draw 2d mass profile for different labelings
    :param dict_peptide2data: dict_ion2data: labeling_group ->
                                                {'intensity': [], 'mz_diff': [], 'rt': [], "total_rt": []}
    :param data_directory: data directory
    :param plot_title: Title of the plot
    :param dpi_user: DPI
    :param ms2_window: MS/MS window
    :param list_label_groups: List label groups
    :return: none
    """
    list_label_groups.sort()
    num_labelings = len(list_label_groups)
    span_for_plots = 30
    # Color scales
    spectral_scheme = plt.get_cmap('winter')
    c_norm = colors.Normalize(vmin=0, vmax=ms2_window)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)
    plt.figure()
    for i in range(num_labelings):
        ax = plt.subplot2grid((num_labelings, num_labelings + span_for_plots), (i, 0), colspan=span_for_plots)
        label_group = list_label_groups[i]
        total_rt = dict_peptide2data[label_group]['total_rt']
        rt_start = dict_peptide2data[label_group]['rt_start']
        rt_end = dict_peptide2data[label_group]['rt_end']
        max_delta_mass = dict_peptide2data[label_group]['max_del']
        zipped_array = zip(dict_peptide2data[label_group]['rt'],
                           dict_peptide2data[label_group]['intensity'],
                           dict_peptide2data[label_group]['mz_diff'])
        # Tick size
        ax.tick_params(axis='both', labelsize=font_size_ticks_2d)
        if len(total_rt) == 0:
            ax.text(x=0.4, y=0.5, s="Not Detected")
            ax.set_ylabel('Peak intensity', fontsize=font_size_label_2d)
            plt.text(0, 0, label_group)
        else:
            count_filtered = 0
            for rt, intensity, mz_diff in zipped_array:
                # Filter out ion intensities by retention time and delta m/z
                if check_range(rt_start, rt_end, rt):
                    continue
                if mz_diff > max_delta_mass:
                    continue
                color_value = scalar_map.to_rgba(mz_diff)
                ax.plot([rt, rt], [0, intensity], '-', color=color_value)
                count_filtered += 1
            # Set X axis limit
            total_rt = total_rt[0]
            low = float(total_rt.split("_")[0])
            high = float(total_rt.split("_")[1])
            ax.set_xlim([low, high])
            if count_filtered == 0:
                ax.text(x=numpy.mean([low, high]), y=0.5, s="No Valid Peaks")
            # Y axis ticks will be shown like scientific numbers
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            ax.ticklabel_format(axis='x', useOffset=False)
            # Set labels
            ax.set_ylabel('Peak intensity', fontsize=font_size_label_2d)
            # Set subplot title
            plt.text(low, 0, label_group)
        if i == 0:
            # Set plot title
            ax.set_title(plot_title.replace('_', ' '))
        if i == num_labelings - 1:
            # Set X-axis label
            ax.set_xlabel('Retention Time', fontsize=font_size_label_2d)
    # Plot color bar
    color_bar = plt.subplot2grid((num_labelings, num_labelings + span_for_plots), (0, span_for_plots),
                                 rowspan=num_labelings)
    cb.ColorbarBase(color_bar, norm=c_norm, cmap=spectral_scheme)
    # Save the plot
    img_file_name = change_filename("{0}.png".format(plot_title))
    img_file_path = os.path.join(data_directory, img_file_name)
    plt.savefig(img_file_path, dpi=dpi_user)
    plt.close()


def get_ion_series_from_tr_id(tr_id):
    tmp = tr_id.split('+')
    tmp2 = tmp[1].split('-')
    ion_series = tmp2[1]
    peptide_sequence = "{0}+{1}".format(tmp[0], tmp2[0])
    return ion_series, peptide_sequence


def parse_xpi_all_obj(xpi_all_obj):
    # xpi_all_obj.data[transition_id][file_name][labeling_group] = {'intensity': [], 'mz_diff': []}
    # peptide -> file_name (time or mixing ratio) -> labeling_group -> ion_series ->
    #                                                                       {'intensity': [], 'mz_diff': [], 'rt': []}
    dict_pep_intensities = {}
    for tr_id in xpi_all_obj.data:
        ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
        for file_name in xpi_all_obj.data[tr_id]:
            # ratio_name = config_XPIViz.data[file_name]['RatioName']
            # Generate a dictionary dict_pep_intensities from XPI_output_all object
            for labeling_group in xpi_all_obj.data[tr_id][file_name]:
                list_ints = xpi_all_obj.data[tr_id][file_name][labeling_group]['intensity']
                list_mz_diff = xpi_all_obj.data[tr_id][file_name][labeling_group]['mz_diff']
                list_rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['rt']
                list_total_rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['total_rt']
                # If there is no data, skip it
                if len(list_ints) == 0:
                    continue
                # Initialize dict_pep_intensities
                if peptide_sequence not in dict_pep_intensities:
                    dict_pep_intensities[peptide_sequence] = {}
                if file_name not in dict_pep_intensities[peptide_sequence]:
                    dict_pep_intensities[peptide_sequence][file_name] = {}
                if labeling_group not in dict_pep_intensities[peptide_sequence][file_name]:
                    dict_pep_intensities[peptide_sequence][file_name][labeling_group] = {}
                if ion_series not in dict_pep_intensities[peptide_sequence][file_name][labeling_group]:
                    dict_pep_intensities[peptide_sequence][file_name][labeling_group][ion_series] = \
                        {'intensity': [], 'mz_diff': [], 'rt': []}
                # Assign intensities, mz_diff and rt
                dict_pep_intensities[peptide_sequence][file_name][labeling_group][ion_series]['intensity'] = list_ints
                dict_pep_intensities[peptide_sequence][file_name][labeling_group][ion_series]['mz_diff'] = list_mz_diff
                dict_pep_intensities[peptide_sequence][file_name][labeling_group][ion_series]['rt'] = list_rt
                dict_pep_intensities[peptide_sequence][file_name][labeling_group][ion_series]['total_rt'] = list_total_rt
    return dict_pep_intensities


def get_path_for_plot(data_directory, dir_plot):
    return_dir = os.path.join(data_directory, dir_plot)
    if os.path.isdir(return_dir):
        for root, dirs, files in os.walk(return_dir, topdown=False):
            for name_local in files:
                os.remove(os.path.join(root, name_local))
    else:
        os.mkdir(return_dir)
    return return_dir


def check_range(low_limit, high_limit, value):
    """
    Check something is in the range of low limit to high limit
    :param low_limit:  Low limit
    :param high_limit: high limit
    :param value: value for checking
    :return: boolean type
    """
    try:
        num_lower_limit = float(low_limit)
        num_value = float(value)
        num_high_limit = float(high_limit)
    except ValueError:
        num_lower_limit = 0
        num_value = 0
        num_high_limit = 0
        Message.print_message("Please check your inclusion list. XPI found non-number retention time")
    if num_lower_limit <= num_value <= num_high_limit:
        return False
    else:
        return True


def check_sum_ratio_of_fragment(data_directory, xpi_check_data, config_xpiviz):
    ratio_name = config_xpiviz.ratio_name
    flag_log = config_xpiviz.flag_log
    dict_prot2filter = config_xpiviz.filters
    m0_intensity_col_name = 'm0_intensity'
    m3_intensity_col_name = 'm3_intensity'
    # construct a dictionary to collect required information from XPI_output.txt
    return_dict = {}
    list_metrics = []
    dict_tr2time2m0int = {}
    dict_tr2time2m3int = {}
    dict_tr2time2ttr = {}
    dict_tr2prot = {}
    for tr_id in xpi_check_data:
        for file_name in xpi_check_data[tr_id]:
            row = xpi_check_data[tr_id][file_name]
            metrics = config_xpiviz.data[file_name]['MR']   # e.g. hours
            dict_tr2prot[tr_id] = row['Protein_Name']
            m0_intensity = row[m0_intensity_col_name]
            m3_intensity = row[m3_intensity_col_name]
            if row[ratio_name] == 'NA':
                ratio = 0.0
            else:
                try:
                    if flag_log:
                        ratio = Scatterplot.get_log10(float(row[ratio_name]))
                    else:
                        if config_xpiviz.percent_ratio:
                            ratio = float(row[ratio_name])*100
                        else:
                            ratio = float(row[ratio_name])
                except ValueError:
                    Message.print_message("{0} is not number".format(row[ratio_name]))
            if tr_id not in dict_tr2time2m0int:
                dict_tr2time2m0int[tr_id] = {}
                dict_tr2time2m3int[tr_id] = {}
                dict_tr2time2ttr[tr_id] = {}
            if metrics not in dict_tr2time2m0int[tr_id]:
                dict_tr2time2m0int[tr_id][metrics] = float(m0_intensity)
                dict_tr2time2m3int[tr_id][metrics] = float(m3_intensity)
                dict_tr2time2ttr[tr_id][metrics] = ratio
            else:
                dict_tr2time2m0int[tr_id][metrics] += float(m0_intensity)
                dict_tr2time2m3int[tr_id][metrics] += float(m3_intensity)
                dict_tr2time2ttr[tr_id][metrics] += ratio
            if metrics not in list_metrics:
                list_metrics.append(metrics)
    # write the filtered data
    fp_out = open(os.path.join(data_directory, file_filtered), "w")
    fp_out.write("{0},Protein,sum_intensity,flag_sum_intensity,sum_ratio,flag_sum_ratio,flag_sum_ratio_max\n".format(colname_transit_id))
    # calculate sum_intensity and sum_ratio
    for tr_id in sorted(dict_tr2time2m0int):
        prot_name = dict_tr2prot[tr_id]
        threshold_inten = dict_prot2filter[prot_name]["sum_intensity"]
        threshold_ratio = dict_prot2filter[prot_name]["sum_ratio"]
        threshold_ratio_max = dict_prot2filter[prot_name]["sum_ratio_max"]
        list_ttr = []
        sum_m0_intensity = 0.0
        for metrics in sorted(list_metrics):
            if metrics in dict_tr2time2m0int[tr_id]:
                sum_m0_intensity += dict_tr2time2m0int[tr_id][metrics]
                list_ttr.append(dict_tr2time2ttr[tr_id][metrics])
        sum_m3_intensity = 0.0
        for metrics in sorted(list_metrics):
            if metrics in dict_tr2time2m3int[tr_id]:
                sum_m3_intensity += dict_tr2time2m3int[tr_id][metrics]
        # Remove infinite value
        while True:
            if numpy.inf in list_ttr:
                list_ttr.remove(numpy.inf)
            else:
                break
        # Check if tr_id is in return_dict
        if tr_id in return_dict:
            Message.print_message("Redundant transition id in transition file: {0}".format(tr_id))
        return_dict[tr_id] = {}
        # If m0_intensity_sum is less than threshold
        if threshold_inten > sum_m0_intensity:
            return_dict[tr_id][colname_sum_intensity] = True
        else:
            return_dict[tr_id][colname_sum_intensity] = False
        # Threshold for ratio sum
        # If ratio_sum is smaller than lower limit
        if threshold_ratio > numpy.sum(list_ttr):
            return_dict[tr_id][colname_sum_ratio] = True
        else:
            return_dict[tr_id][colname_sum_ratio] = False
        # If ratio_sum is bigger than upper limit
        if threshold_ratio_max == default_ratio_max_threshold:  # ignore threshold_ratio_max
            return_dict[tr_id][colname_sum_ratio_max] = False
        elif threshold_ratio_max < numpy.sum(list_ttr):
            return_dict[tr_id][colname_sum_ratio_max] = True
        else:
            return_dict[tr_id][colname_sum_ratio_max] = False
        fp_out.write("{0},{1},{2},{3},{4},{5},{6}\n".format(
        tr_id, prot_name, sum_m0_intensity, return_dict[tr_id][colname_sum_intensity], 
        str(numpy.sum(list_ttr)), return_dict[tr_id][colname_sum_ratio],
        return_dict[tr_id][colname_sum_ratio_max]))
    fp_out.close()
    return return_dict

def get_tr_id_selection_info(data_directory):
    return_dict = {}
    rows = csv.DictReader(open(os.path.join(data_directory, file_filtered), "rU"))
    for row in rows:
        tr_id = row[colname_transit_id]
        return_dict[tr_id] = {}
        return_dict[tr_id][colname_sum_intensity] = str2bool(row["flag_{0}".format(colname_sum_intensity)])
        return_dict[tr_id][colname_sum_ratio] = str2bool(row["flag_{0}".format(colname_sum_ratio)])
        return_dict[tr_id][colname_sum_ratio_max] = str2bool(row["flag_{0}".format(colname_sum_ratio_max)])
    return return_dict


def str2bool(s):
    s = str(s).lower()
    # 'VKDLATVYVDVLK+3-y4': {'sum_intensity': False, 'sum_ratio': False, 'sum_ratio_max': True},
    if s == 'true':
         return True
    elif s == 'false':
         return False
    else:
         raise ValueError # evil ValueError that doesn't tell you what the wrong value was
        

def make_numbers_same_significant_digits(num_a, num_b):
    sig_digits_num_a = len(str(num_a)) - 2
    sig_digits_num_b = len(str(num_b)) - 2
    if sig_digits_num_a < sig_digits_num_b:
        return format(num_a, '.{0}f'.format(sig_digits_num_a)), format(num_b, '.{0}f'.format(sig_digits_num_a))
    else:
        return format(num_a, '.{0}f'.format(sig_digits_num_b)), format(num_b, '.{0}f'.format(sig_digits_num_b))


def get_filtered_intensities(dict_out_all_arg, dict_narrow_rt_arg, peptide_sequence_local, file_name_local,
                             label_group_name_local, ion_series_local, ms2_window=-1):
    """
    Get RT, intensity and mz_diff of ions that passed mz_diff and narrowed down RT filters
    :param dict_out_all_arg: xpi_all_obj.data[tr_id][file_name][label_group_name]
    :param dict_narrow_rt_arg: dict_narrowedRT
    :param peptide_sequence_local: peptide sequence
    :param file_name_local: file name
    :param label_group_name_local: label group name
    :param ion_series_local: ion series
    :return: a dictionary
    """
    zipped_array_local = zip(dict_out_all_arg['rt'], dict_out_all_arg['intensity'], dict_out_all_arg['mz_diff'],
                             dict_out_all_arg['mz'])
    local_intensity_list = []
    local_rt_list = []
    local_mz_diff_list = []
    local_mz_list = []

    flag_narrow_rt = True
    if len(dict_out_all_arg['intensity']) == 0:
        flag_narrow_rt = False
    elif label_group_name_local not in dict_narrow_rt_arg[peptide_sequence_local][file_name_local]:
        flag_narrow_rt = False
    elif ion_series_local not in dict_narrow_rt_arg[peptide_sequence_local][file_name_local][label_group_name_local]:
        flag_narrow_rt = False
    if flag_narrow_rt:
        # Retrieve data from dict_narrwedRT
        dict_narrow_rt_local = \
            dict_narrow_rt_arg[peptide_sequence_local][file_name_local][label_group_name_local][ion_series_local]
        peak_range_local = dict_narrow_rt_local['peak_range']
        max_delta_mass_local = dict_narrow_rt_local['max_delta_mass']
        if ms2_window > 0:
            if max_delta_mass_local > ms2_window:
                max_delta_mass_local = ms2_window
        list_rt_local = peak_range_local.split(delimiter_residue)
        rt_start_local = float(list_rt_local[0])
        rt_end_local = float(list_rt_local[2])
        # Filtering ions by new RT window and new mass difference limit
        for rt_local, intensity_local, mz_diff_local, mz_local in zipped_array_local:
            if check_range(rt_start_local, rt_end_local, rt_local):
                continue
            mz_diff_f, max_delta_mass_f = make_numbers_same_significant_digits(mz_diff_local, max_delta_mass_local)
            if mz_diff_f > max_delta_mass_f:
                continue
            local_intensity_list.append(intensity_local)
            local_rt_list.append(rt_local)
            local_mz_diff_list.append(mz_diff_local)
            local_mz_list.append(mz_local)
    return {"intensity": local_intensity_list, "rt": local_rt_list, "mz_diff": local_mz_diff_list, 'mz': local_mz_list}


def make_transition_id(pep_seq, pep_charge, pep_ion):
    return "{0}{3}{1}{4}{2}".format(pep_seq, pep_charge, pep_ion, delimiter_config, delimiter_dash)


############################
# START XPIQuant libraries #
############################
xpi_quant_config = "XPIQuant configurations:\tSee the manual section 7"


# for informing how to use this script
class UserMessageXPIQuant(Message):
    def __init__(self, message):
        self.usage = """This script reads mzML files and a transition file to quantify PRM transitions.
Usage:
    $ python XPIQuant.py [path of config file]
    e.g.) $ python XPIQuant.py XPIQuent_config.txt

# In the [data directory] the following files should be there
    1. XPIQuant_config.txt
    2. mzML files
    3. Inclusion list

{0}
""".format(xpi_quant_config)
        self.print_message(message, usage=self.usage)


# Class for configuration file
class XPIQuantConfig(File):
    def __init__(self, filename, delim=delimiter_tab, delim_label=delimiter_label,
                 delim_residues=delimiter_residue, delim_config=delimiter_config):
        File.__init__(self, filename)
        self.list_required_items = list_requirements
        self.delim = delim
        self.delim_label = delim_label
        self.delim_residues = delim_residues
        self.delim_config = delim_config
        self.modification = {}
        self.data = {}

        self.check_file()
        self.parse_file()
        self.check_items()

        self.labeling = Labeling(self.data['Labeling'], self.delim_config,
                                 self.delim_label, self.delim_residues, self.data[colname_max_l_sites])
        if colname_modification in self.data:
            self.modification = Modification(self.data[colname_modification])

        self.print_file()

    def print_multiple_items(self, item_name):
        list_ratio_labels = self.data[item_name].split(delimiter_config)
        print("{0:17s}:".format(item_name), end=" ")
        flag_first = True
        for ratio_label in list_ratio_labels:
            if flag_first:
                print("{0}".format(ratio_label))
                flag_first = False
            else:
                print("{0:17s}: {1}".format("", ratio_label))

    def print_file(self):
        """
        Print Configurations
        :return: none
        """
        for config_item in self.list_required_items:
            if config_item == "Labeling":
                self.labeling.print_labeling()
            else:
                if config_item not in self.data:
                    continue
                if re.search("\+", str(self.data[config_item])):
                    self.print_multiple_items(config_item)
                else:
                    print("{0:17s}: {1}".format(config_item, self.data[config_item]))

    def parse_file(self):
        """
        Parse config file and assign config information to a dictionary
        :return: a dictionary containing configuration information
        """
        fp = open(self.filename, "rU")
        for data_line in fp:
            if data_line.strip() == "":
                continue
            if re.match("#", data_line):
                continue
            if re.match("\s+", data_line):
                continue
            list_data = data_line.strip().split(self.delim)
            if len(list_data) != 2:
                print("Found wrong format of config item \"{0}\"".format(data_line.strip()))
                continue
            item = list_data[0]
            value = list_data[1]
            if item not in self.data:
                if item == colname_ttr:
                    self.data[item] = XPIVizConfig.read_flag(value)
                # For old version XPIQuant_config
                elif item == "TTR":
                    self.data[colname_ttr] = XPIVizConfig.read_flag(value)
                elif item == colname_normalize:
                    self.data[item] = XPIVizConfig.read_flag(value)
                elif item == colname_background:
                    self.data[item] = XPIVizConfig.read_flag(value)
                elif item == colname_skip_inclusion:
                    self.data[item] = XPIVizConfig.read_flag(value)
                elif item == colname_ax_ion:
                    self.data[item] = XPIVizConfig.read_flag(value)
                else:
                    self.data[item] = value
            else:
                self.data[item] = "{0}{1}{2}".format(self.data[item], self.delim_config, value)
        fp.close()

    def check_items(self):
        """
        Check items in the config file
        :return:
        """
        for config_item in self.list_required_items:
            if config_item not in self.data:
                if config_item == colname_modification:
                    print("No modification found")
                elif config_item == colname_max_peak:
                    print("No Max peak number, so use {0} by default".format(default_max_num_peaks))
                    self.data[config_item] = 1
                elif config_item == colname_ttr:
                    print("Simple ratio will be calculated")
                    self.data[config_item] = False
                elif config_item == colname_max_l_sites:
                    print("All the possible labeling sites will be considered")
                    self.data[config_item] = 'all'
                elif config_item == colname_ax_ion:
                    print("XPI will quantify 'b' and 'y' ions only")
                    self.data[config_item] = False
                else:
                    Message.print_message("\"%s\" item is missing in your config file \"%s\"\n%s" % (
                        config_item, self.filename, xpi_quant_config))


# Class to handle labeling information
class Labeling:
    def __init__(self, labeling_info, delim_config=delimiter_config, delim_label=delimiter_label,
                 delim_residues=delimiter_residue, max_labeling_sites='all'):
        # Labeling.group --> key: label name = {'Residues':[],'Shifts':[]}
        self.config = labeling_info
        self.data = {}
        self.delim_config = delim_config
        self.delim_label = delim_label
        self.delim_residues = delim_residues
        self.max_labeling_sites = max_labeling_sites

        self.parse_labeling()

    def print_labeling(self):
        """
        Print obtained labeling information
        :return: string type
        """
        flag_first = True
        return_string = "{0:17s}".format("Labeling")
        for label_group_name in self.data:
            out_string_labeling = ""
            for i in range(len(self.data[label_group_name]['Residues'])):
                out_string_labeling += "{0}({1})  ".format(
                    self.data[label_group_name]['Residues'][i],
                    self.data[label_group_name]['Shifts'][i])
            if flag_first:
                space = ""
                flag_first = False
            else:
                space = "{0:17s}".format("")
            return_string += "{2}: {0}\t{1}\n".format(label_group_name, out_string_labeling, space)
        print(return_string, end="")

    def parse_labeling(self):
        """
        Parse Labeling information formatted as
        "Heavy, K:8.02683872, R:10.0335484" (Label name, AA residue:Mass shift, ...)
        :return: none
        """
        # Light, K:0.0, R:0.0+Medium, K:4.01341936, R:6.02012904+Heavy, K:8.02683872, R:10.0335484
        list_label_groups = self.config.strip().split(self.delim_config)
        for label_group in list_label_groups:
            # Heavy, K:8.02683872, R:10.0335484
            list_label_info = label_group.split(self.delim_label)
            label_group_name = list_label_info[0]
            self.data[label_group_name] = {'Residues': [], 'Shifts': []}
            for i in range(1, len(list_label_info)):
                label_info = list_label_info[i]
                if not re.search(self.delim_residues, label_info):
                    Message.print_message("Found wrong format of labeling information at {0}".format(label_info))
                # K:8.02683872, R:10.0335484
                list_tmp = label_info.strip().split(self.delim_residues)
                label_aa = list_tmp[0]
                ms_shift = float(list_tmp[1])
                self.data[label_group_name]['Residues'].append(label_aa)
                self.data[label_group_name]['Shifts'].append(ms_shift)

    def check_labeling(self, sequence):
        list_labeling_aa = []
        for label_name in self.data:
            list_residues = self.data[label_name]['Residues']
            for aa in list_residues:
                if aa not in list_labeling_aa:
                    list_labeling_aa.append(aa)
        return check_aa_in_sequence(list_labeling_aa, sequence)

    def get_labeling(self, fragment_mass, fragment_seq, fragment_ion_series):
        """
        colname_max_l_sites
        Returns theoretical m/z of labeled ions
        :param fragment_mass: m/z of fragmented ion
        :param fragment_seq: sequence of fragmented ion
        :param fragment_ion_series: ion series, b- or y-ion, of fragmented one
        :return: a dictionary containing shifted m/z of each labeling groups
        """
        dict_return = {}
        for label_name in self.data:
            list_residues = self.data[label_name]['Residues']
            list_shifts = self.data[label_name]['Shifts']
            fragment_shift = 0.0
            # For multiple labeling sites
            for i in range(len(list_residues)):
                labeling_aa = list_residues[i]
                aa_shift = list_shifts[i]
                num_labeling_sites = self.get_num_labeling_sites(fragment_seq, fragment_ion_series, labeling_aa)
                fragment_shift += self.calculate_mass_shift(num_labeling_sites, aa_shift)
            dict_return[label_name] = fragment_mass + fragment_shift
        return dict_return

    def calculate_mass_shift(self, num_theoretical_sites, aa_shift):
        if self.max_labeling_sites == 'all':
            return num_theoretical_sites * aa_shift
        else:
            self.max_labeling_sites = int(self.max_labeling_sites)
            if self.max_labeling_sites <= num_theoretical_sites:
                return self.max_labeling_sites * aa_shift
            else:
                return num_theoretical_sites * aa_shift

    @staticmethod
    def get_num_labeling_sites(fragment_seq, fragment_ion_series, labeling_aa):
        """
        Calculates mass shifts by labeled residues in a fragmented ion
        :param fragment_seq: sequence of fragmented ion
        :param fragment_ion_series: ion series, b- or y-ion, of fragmented one
        :param labeling_aa: amino acid residue labeled
        :return: float type of labeled ion's m/z
        """
        pep_ion_location = int(fragment_ion_series[1])
        if re.match('b', fragment_ion_series):
            return int(fragment_seq[:pep_ion_location].count(labeling_aa))

        elif re.match('y', fragment_ion_series):
            return int(fragment_seq[-1 * pep_ion_location:].count(labeling_aa))

        elif re.match('a', fragment_ion_series):
            return int(fragment_seq[:pep_ion_location].count(labeling_aa))

        elif re.match('x', fragment_ion_series):
            return int(fragment_seq[-1 * pep_ion_location:].count(labeling_aa))


class Modification:
    def __init__(self, mods):
        self.mods = mods
        self.data = {}

        self.parse_mods()

    def parse_mods(self):
        list_mods = self.mods.split(delimiter_config)
        for modify in list_mods:
            list_mod = modify.split(delimiter_residue)
            if len(list_mod) != 2:
                Message.print_message("Wrong format in Modification")
            residue = list_mod[0]
            mass_mod = list_mod[1]
            try:
                self.data[residue] = float(mass_mod)
            except ValueError:
                Message.print_message("Wrong format in Modification: mass should be a number")


class InclusionList(File):
    def __init__(self, config_object):
        File.__init__(self, config_object.data[colname_inclusion_list])
        self.data_directory = config_object.data[colname_data_directory]
        self.config = config_object
        self.fp_data = self.get_file()
        self.mod_mass = {}
        self.data = {}
        self.aa_comp_mod = mass.std_aa_comp.copy()

        # Check requirements for inclusion list file
        self.check_headers()
        # If there is Modification, is_simple = True, otherwise False
        self.is_simple = self.set_flag()
        # Parse Inclusion file and make Transition file
        self.make_transition_file()

    def check_headers(self):
        tmp_data = self.get_file()
        for row in tmp_data:
            for required_col_name in list_requirements_inclusion:
                if required_col_name not in row:
                    m = "Inclusion list file doesn't have required information: {0}".format(required_col_name)
                    Message.print_message(m)
            break

    def set_flag(self):
        """
        Set a flag indicating whether there is modifications or not
        :return:
        """
        if colname_modification in self.config.data:
            self.mod_mass = self.config.modification.data
            return False
        else:
            self.set_mods_silac()
            return True

    def set_mods_silac(self):
        """
        Add Oxidized Methionine in SILAC labeling (http://www.unimod.org/modifications_view.php?editid1=885)
        Add Carbamidomethyl Cysteine
        :return: none
        """
        db = mass.Unimod()
        carbamidomethyl = db.by_title('Carbamidomethyl')
        self.aa_comp_mod['c'] = self.aa_comp_mod['C'] + carbamidomethyl['composition']
        oxidation = db.by_title('Oxidation')
        self.aa_comp_mod['m'] = self.aa_comp_mod['M'] + oxidation['composition']

    def find_modification(self, sequence):
        """
        Find modifications, user added, from the input sequence
        :param sequence:    peptide sequence
        :return:            a dictionary
        """
        dict_return = {}
        for modi in self.mod_mass:
            if re.search(modi, sequence):
                count = len(re.findall(modi, sequence))
                dict_return[modi] = count
                sequence = sequence.replace(modi, '')
        dict_return['without_mod'] = sequence
        return dict_return

    def calculate_peptide_mass(self, sequence, charge, ion_type):
        """
        Calculate peptide molecular mass
        :param ion_type:    ion type
        :param sequence:    peptide sequence
        :param charge:      precursor ion's charge
        :return:            float type
        """
        # If user did not add any modifications
        if self.is_simple:
            # Calculate molecular mass using pyteomics
            return mass.calculate_mass(sequence=sequence, ion_type=ion_type, charge=charge, aa_comp=self.aa_comp_mod)

        # If user added any modifications
        else:
            dict_mod_info = self.find_modification(sequence)
            sequence_without_mod = dict_mod_info['without_mod']
            # Calculate molecular mass without modifications
            mz_without_mod = mass.calculate_mass(sequence=sequence_without_mod, ion_type=ion_type, charge=1)
            # Calculate hydrogen molecular mass
            hydrogen = mass.calculate_mass(composition={'H': 1})
            mz_without_mod -= hydrogen
            # Calculate molecular mass of modified residues
            modified_mass = 0.0
            for modd in dict_mod_info:
                if modd == 'without_mod':
                    continue
                count = dict_mod_info[modd]
                modified_mass += self.mod_mass[modd] * count
            # Return m/z of precursor ion
            return (mz_without_mod + modified_mass + hydrogen * charge) / charge

    def make_transition_file(self):
        """
        Write Transition file from Inclusion list file
        :return: none
        """
        out_file = os.path.join(self.data_directory, file_transition)
        fp_out = open(out_file, "w")
        fp_out.write(delimiter_tab.join(transition_file_header) + "\n")
        for row in self.fp_data:
            # Mass [m/z]
            # Formula [M]
            # Formula type
            # Species
            # CS [z]
            # Polarity
            # Start [min]
            # End [min]
            # NCE
            # Sequence
            # Protein
            pep_seq = row['Sequence']
            sym_prot = row['Protein']
            charge = row['CS [z]']
            rt_start = row['Start [min]']
            rt_end = row['End [min]']
            filter_string = row['Mass [m/z]']
            rt = "{0}_{1}".format(rt_start, rt_end)

            # Precursor ion_typ information
            precursor_theoretical_mz = self.calculate_peptide_mass(pep_seq, int(charge), 'M')
            # 'Precursor_MZ', 'RT', 'Peptide Sequence', 'Protein Name', 'precursor_charge', 'filter string'
            precursor_info = "{1}{0}{2}{0}{3}{0}{4}{0}{5}{0}{6}".format(delimiter_tab,
                                                                        precursor_theoretical_mz, rt,
                                                                        pep_seq, sym_prot, charge, filter_string)
            # Set self.data for new transition ID
            transition_id = "{0}+{1}".format(pep_seq, charge)
            if transition_id not in self.data:
                self.data[transition_id] = {}
            # Get fragment ions series
            max_length = int(self.config.data[colname_max_length_fragment])
            min_length = PRM_fragment_min_length
            for b in range(len(pep_seq)):
                y = len(pep_seq) - b
                if min_length <= b <= max_length:
                    self.assign_transition('b', b, pep_seq, transition_id)
                if min_length <= y <= max_length:
                    self.assign_transition('y', y, pep_seq, transition_id)
                if self.config.data[colname_ax_ion]:
                    a = b
                    if min_length <= a <= max_length:
                        self.assign_transition('a', a, pep_seq, transition_id)
                """
                # 'x' ion
                x = y
                if min_length <= x <= max_length:
                    self.assign_transition('x', x, pep_seq, transition_id)
                """
            # 'Fragment_MZ', 'ion-series', 'Fragment_Sequence'
            for ion_typ in self.data[transition_id]:
                fp_out.write("{1}{0}{2}{0}{3}{0}{4}\n".format(delimiter_tab, precursor_info,
                                                              self.data[transition_id][ion_typ]['mz'], ion_typ,
                                                              self.data[transition_id][ion_typ]['seq']))
        fp_out.close()

    def assign_transition(self, ion_type, n, pep_seq, transition_id):
        ion_types = ion_type + str(n)
        seq = ''
        if ion_type == 'b' or ion_type == 'a':
            seq = ''.join(''.join(map(str, l)) for l in pep_seq[:n])
        elif ion_type == 'x' or ion_type == 'y':
            seq = ''.join(''.join(map(str, l)) for l in pep_seq[len(pep_seq) - n:])
        else:
            Message.print_message("Wrong ion type: {0}".format(ion_type))
        if self.config.labeling.check_labeling(seq):
            if ion_types not in self.data[transition_id]:
                self.data[transition_id][ion_types] = {}
            self.data[transition_id][ion_types]['seq'] = seq
            self.data[transition_id][ion_types]['mz'] = str(self.calculate_peptide_mass(seq, 1, ion_type))

# python XPIQuant.py ./mzML/paper_revision/06_Merged/05_Intraday/pt243/XPIQuant_config.txt
# Class for transition file
class Transition(File):
    def __init__(self, filename, labeling):
        File.__init__(self, filename)
        self.labeling = labeling
        self.data = {}
        self.list_mz_fragments = []

        self.check_file()
        self.parse_file()

    def get_mz_fragments(self):
        """
        Get MS/MS fragment m/z for screening while reading mzML file
        :return:
        """
        for tr_id in self.data:
            for label_id in self.data[tr_id]['Labeling']:
                mz = self.data[tr_id]['Labeling'][label_id]
                try:
                    mz = int(float(mz))
                except ValueError:
                    Message.print_message("ValueError, wrong number in your transition file")
                self.list_mz_fragments.append(mz)

    def parse_file(self):
        """
        Parse the transition file and assign the file contents to a dictionary, "data"
        :require: csv
        :return: a dictionary containing parsed transition file data
        """
        num_transitions = 0
        fp_data = self.get_file()
        for row in fp_data:
            fragment_mz = float(row[colname_fragment_mz])
            pep_seq = row[colname_pep_seq]
            pep_ion = row[colname_ion_series]
            pep_charge = row[colname_precursor_charge]
            retention = row[colname_rt]
            frag_seq = row["Fragment_Sequence"]

            transition_id = make_transition_id(pep_seq, pep_charge, pep_ion)
            if transition_id in self.data:
                print("Redundant transition is found!\t{0}".format(transition_id))

            self.data[transition_id] = {}
            self.data[transition_id][colname_fragment_mz] = fragment_mz
            self.data[transition_id]['Fragment_Sequence'] = frag_seq
            self.data[transition_id]['Peptide_Sequence'] = pep_seq
            self.data[transition_id]['Ion_Series'] = pep_ion
            self.data[transition_id]['Precursor_Charge'] = pep_charge
            self.data[transition_id][colname_theoretical_mz] = row[colname_theoretical_mz]
            self.data[transition_id]['Protein_Name'] = row[colname_prot_id]
            self.data[transition_id]['Low_RT'] = retention.split('_')[0]
            self.data[transition_id]['High_RT'] = retention.split('_')[1]
            if 'filter string' in row:
                self.data[transition_id]['Filter_String'] = "{0:.2f}".format(float(row[colname_filter_string]))
            self.data[transition_id]['Labeling'] = {}
            self.data[transition_id]['Labeling'] = self.labeling.get_labeling(fragment_mz, pep_seq, pep_ion)
            num_transitions += 1
        print("Found {0} transitions".format(num_transitions))


# Class for mzML file using pymzml package
class MZML(File):
    def __init__(self, filename, list_targets=False, tic_normalize=False, background=False):
        File.__init__(self, filename)
        self.normalize = tic_normalize
        self.background = background
        self.list_targets = list_targets
        self.data = {}
        self.check_file()
        self.parse_file()
        self.write_file()

    @staticmethod
    def get_noise_level(scan, method='med'):
        list_intensities = []
        for mz, intensity in scan.centroidedPeaks:
            list_intensities.append(intensity)
        return pick_intensity(list_intensities, method)

    def write_file(self):
        fp_out = open("{0}.extracted.csv".format(self.filename), "w")
        fp_out.write("Scan_ID,RT,MS2_M/Z,Intensity_Normalized,Intensity_Raw,Filter_String,MS1_M/Z,Charge\n")
        for ion_id in self.data:
            fp_out.write("{0},{1},{2},{3},{4},{5},{6},{7}\n".format(self.data[ion_id].scan_id,
                                                                    self.data[ion_id].retention_time,
                                                                    self.data[ion_id].mz_ms2,
                                                                    self.data[ion_id].intensity,
                                                                    self.data[ion_id].raw_intensity,
                                                                    self.data[ion_id].filter_string_value,
                                                                    self.data[ion_id].mz_precursor,
                                                                    self.data[ion_id].charge_precursor))
            # self.data[ion_id].tic
            # self.data[ion_id].noise_level
        fp_out.close()

    def parse_file(self):
        """
        Parse the mzML file to collect the following information
            1. 'ScanID'
            2. 'ms2_m/z'
            3. 'intensity'
            4. 'Precursor_m/z'
            5. 'Precursor_charge'
            6. 'RT'
            7. 'TIC'
            8. 'Filter_String_Value'
        Then, generate PRMscan objects
        """
        mzml_obj = pymzml.run.Reader(self.filename, measured_precision=5e-06)
        # https://pymzml.readthedocs.io/en/latest/pymzml_spec.html
        for scan in mzml_obj:
            # Extract ms level 2
            if scan.ms_level == 2:
                if self.background:
                    # noise_level = scan.estimated_noise_level(mode='median')
                    noise_level = self.get_noise_level(scan)
                    noise_level_for_report = noise_level
                else:
                    noise_level_for_report = self.get_noise_level(scan)
                    noise_level = 0
                tic = scan.TIC
                # MS1 precursor ion mass used in the inclusion list
                # 'FTMS + p NSI Full ms2 406.8850@hcd25.00 [100.0000-1280.0000]'
                filter_string = scan['filter string'].split(" ")[6]
                filter_string = filter_string[:filter_string.find("@")]
                ## For processing centroided peaks
                for mz, intensity in scan.centroidedPeaks:
                    # If there is target fragment m/z list, filter out unmatched m/z to speed up the further processes
                    # list_targets contains integer m/z of the theoretically calculated fragments
                    if self.list_targets:
                        tmp_mz = int(float(mz))
                        if tmp_mz not in self.list_targets:
                            continue
                    # we only take i that is more than 0
                    if intensity > noise_level:
                        # ion_id = scan number + mz
                        ion_id = "{0}-{1}".format(str(scan.ID), str(mz))
                        if ion_id in self.data:
                            print("{0} is redundantly matched.".format(ion_id))
                            ion_id = "{0}1".format(ion_id)
                        intensity_normalized = intensity
                        if self.normalize:
                            intensity_normalized = intensity_normalized / tic
                        # scan.selected_precursors
                        # [{'mz': 406.885009765625, 'i': 4190.69580078125, 'charge': 3}]
                        self.data[ion_id] = MS2ScanInfo(scan.ID, mz, intensity_normalized,
                                                        scan.selected_precursors[0]["mz"],
                                                        scan.selected_precursors[0]['charge'],
                                                        scan.scan_time[0],
                                                        tic, filter_string, intensity, noise_level_for_report)


# Class for PRM scan
class MS2ScanInfo:
    def __init__(self, scan_id, mz_ms2, intensity, mz_precursor, charge_precursor, retention_time, tic,
                 filter_string_value, raw_intensity, noise_level):
        self.scan_id = scan_id
        self.mz_ms2 = mz_ms2
        self.intensity = intensity
        self.mz_precursor = mz_precursor
        self.charge_precursor = charge_precursor
        self.retention_time = retention_time
        self.tic = tic
        self.filter_string_value = filter_string_value
        self.raw_intensity = raw_intensity
        self.noise_level = noise_level

        # Calculated by calling 'add_ppm' function
        self.mass_difference = 0.0
        self.ppm = 0.0

    def set_mass_difference(self, theoretical_mass):
        """
        Add Parts per million information
        :param theoretical_mass: float type
        :return: none
        """
        if not isinstance(theoretical_mass, float):
            theoretical_mass = float(theoretical_mass)
        self.mass_difference = abs(theoretical_mass - self.mz_ms2)
        self.ppm = self.calculate_ppm(theoretical_mass, self.mz_ms2)

    @staticmethod
    def calculate_ppm(theoretical_mass, observed_mass):
        """
        Calculate parts per million of mass difference between theoretical and observed mass
        :param theoretical_mass: float
        :param observed_mass: float
        :return: float
        """
        if not isinstance(theoretical_mass, float):
            theoretical_mass = float(theoretical_mass)
        if not isinstance(observed_mass, float):
            observed_mass = float(observed_mass)
        return ((observed_mass - theoretical_mass) / observed_mass) * 1000000


# Class for PRM data analysis
class PRMDataset:
    def __init__(self, data_dir, config_object, transit_obj, file_type='mzML'):
        self.data_directory = data_dir
        self.file_type = file_type
        self.config = config_object
        self.transition = transit_obj
        self.target_ext = "%s/*.%s" % (self.data_directory, self.file_type)
        self.list_PRM_files = glob.glob(self.target_ext)
        self.data = {}

        self.check_file()
        self.parse_file()

    def check_file(self):
        """
        Check how many MS Run data is in the designated directory
        :return: none
        """
        if len(self.list_PRM_files) == 0:
            UserMessageXPIQuant("No %s file was found in the directory, \"%s\"" % (self.file_type, self.data_directory))
        else:
            print("Found {0} file(s) in {1}".format(len(self.list_PRM_files), self.data_directory))

    def parse_file(self):
        """
        Call a class for parsing MS Run data, such as calling mzML class for a .mzML file.
        :return: none
        """
        for raw_data in self.list_PRM_files:
            t = Timer()
            t.start_timer("Reading {0}".format(raw_data))
            flag_normalization = self.config.data[colname_normalize]
            flag_background = self.config.data[colname_background]
            if self.file_type.lower() == 'mzml':
                self.data[raw_data] = MZML(raw_data, self.transition.list_mz_fragments,
                                           flag_normalization, flag_background)
            else:
                UserMessageXPIQuant("%s is not acceptable" % self.file_type)
            t.stop_timer()
            t.print_elapsed()

    def print_data(self):
        """
        Print out whole PRM data-set
        :return: string type
        """
        return_string = ""
        for raw_data in self.data:
            for ion_ID in self.data[raw_data].data:
                # ion_id = scan number + mz
                scan_info = self.data[raw_data].data[ion_ID]
                return_string += "{0},{1},{2},{3},{4},{5},{6},{7},{8}\n" \
                                 "".format(raw_data, scan_info.scan_id, scan_info.mz_ms2, scan_info.intensity,
                                           scan_info.mz_precursor, scan_info.charge_precursor,
                                           scan_info.retention_time, scan_info.tic, scan_info.filter_string_value)
        return return_string


# Class for extracting target ions
class PRMTargets:
    def __init__(self, transition, scans, config):
        self.transition = transition
        self.scan_info = ''
        self.scans = scans
        self.config = config
        self.ms1_window = 5.0
        self.ms2_window = float(self.config.data['MS2 window'])
        self.data = {}
        self.data_refined = {}

        self.read_scans()

    def read_scans(self):
        """
        Read PRM scans parsed by a class 'PRM_dataset'
        :return: none
        """
        # for showing progress bar
        for MSRun in self.scans.data:
            pg = Progress(len(self.scans.data[MSRun].data), MSRun.split('/')[-1])
            num = 1
            # ion_id = scan number + mz
            for ion_ID in self.scans.data[MSRun].data:
                # for showing progress bar
                pg.print_progress(num)
                num += 1

                # compare PRM ions to transitions
                scan_info = self.scans.data[MSRun].data[ion_ID]
                self.find_targets(scan_info, MSRun)

    def find_targets(self, scan_info, msrun):
        """
        Find user's target ions described in "transition" class
        by comparing fragmented ion m/z within the obtained window
        :return: none
        """
        self.scan_info = scan_info
        fragment_mz_prm = float(self.scan_info.mz_ms2)
        fragment_rt_prm = self.scan_info.retention_time
        # In some cases, charge is not a number, "None"
        try:
            precursor_charge_prm = int(self.scan_info.charge_precursor)
        except TypeError:
            precursor_charge_prm = "NA"
        except ValueError:
            precursor_charge_prm = "NA"

        for transition_id in self.transition.data:
            transition_info = self.transition.data[transition_id]

            # Check PRM ions over labeling groups
            if transition_id not in self.data:
                self.data[transition_id] = {}
            labeling_info = transition_info['Labeling']
            for label_group_name in labeling_info:
                theoretical_mass = labeling_info[label_group_name]
                low_limit = abs(theoretical_mass - self.ms2_window)
                high_limit = abs(theoretical_mass + self.ms2_window)

                # Generate sub dictionary and list for msrun file and labeling group
                if msrun not in self.data[transition_id]:
                    self.data[transition_id][msrun] = {}
                if label_group_name not in self.data[transition_id][msrun]:
                    self.data[transition_id][msrun][label_group_name] = []

                # If PRM m/z is within the allowance window of target transition
                if check_range(low_limit, high_limit, fragment_mz_prm):
                    continue

                # Check precursor ion's retention time
                if check_range(transition_info['Low_RT'], transition_info['High_RT'], fragment_rt_prm):
                    continue

                # Check charge of precursor ion
                if precursor_charge_prm != "NA":
                    if precursor_charge_prm != int(transition_info['Precursor_Charge']):
                        continue
                # Check MS1
                # Pick exact scans using precursor mass (two options)
                try:
                    # 1. Compare filter string of mzML with mass of inclusion list
                    precursor_mz_prm = float(self.scan_info.filter_string_value)
                except ValueError:
                    # 2. Compare filter string of mzML with theoretical mass
                    precursor_mz_prm = float(self.scan_info.mz_precursor)
                    # precursor_mz_transition = float(transition_info[colname_theoretical_mz])
                precursor_mz_transition = float(transition_info['Filter_String'])
                high_limit = abs(precursor_mz_transition + self.ms1_window)
                low_limit = abs(precursor_mz_transition - self.ms1_window)
                if check_range(low_limit, high_limit, precursor_mz_prm):
                    continue

                # Add scans to transition target DB
                self.scan_info.set_mass_difference(theoretical_mass)
                self.data[transition_id][msrun][label_group_name].append(self.scan_info)

    @staticmethod
    def remove_redundant_hits(target_list):
        """
        If there are multiple hits within one scan ID, pick only one that has smaller mass difference
        :param target_list: a list containing objects of class 'PRM_scan'
        :return: a dictionary {scan id: iteration number}
        """
        dict_scan_i = {}
        dict_scan_delta_mz = {}
        for i in range(len(target_list)):
            scan_obj = target_list[i]
            scan_id = scan_obj.scan_id
            mass_difference = scan_obj.mass_difference
            if scan_id not in dict_scan_delta_mz:
                dict_scan_delta_mz[scan_id] = mass_difference
                dict_scan_i[scan_id] = i
            else:
                if mass_difference < dict_scan_delta_mz[scan_id]:
                    dict_scan_delta_mz[scan_id] = mass_difference
                    dict_scan_i[scan_id] = i
        return dict_scan_i

    def get_targets(self, target_list):
        """
        Get MS/MS fragments data selected by 'find_targets' function
        :param target_list: a list containing objects of class 'PRM_scan'
        :return: string type
        """
        dict_return = {}
        dict_scan_i = self.remove_redundant_hits(target_list)
        for i in range(len(target_list)):
            scan_obj = target_list[i]
            mz_ms2 = float(scan_obj.mz_ms2)
            scan_id = scan_obj.scan_id
            intensity = scan_obj.intensity
            mz_precursor = scan_obj.mz_precursor
            charge_precursor = scan_obj.charge_precursor
            retention_time = scan_obj.retention_time
            tic = scan_obj.tic
            filter_string_value = scan_obj.filter_string_value
            mass_difference = scan_obj.mass_difference
            ppm = scan_obj.ppm
            raw_intensity = scan_obj.raw_intensity
            noise_level = scan_obj.noise_level

            if i == dict_scan_i[scan_id]:
                if scan_obj not in dict_return:
                    dict_return[scan_id] = {'intensity': 0.0, 'print': ''}
                dict_return[scan_id]['print'] = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}".format(
                    scan_id, mz_precursor, charge_precursor, retention_time, mz_ms2, intensity, ppm, mass_difference,
                    tic, filter_string_value, raw_intensity, noise_level)
                dict_return[scan_id]['intensity'] = intensity
                dict_return[scan_id]['mz_ms2'] = mz_ms2
                dict_return[scan_id]['RT'] = retention_time
                dict_return[scan_id]['delMZ'] = mass_difference
        return dict_return

    def get_transition_info(self, transition_id):
        """
        Make output of transition information
        :param transition_id:   transition ID of class Transition
        :return: string type
        """
        precursor_rt = "{0}_{1}".format(
            self.transition.data[transition_id]['Low_RT'],
            self.transition.data[transition_id]['High_RT'])

        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(
            self.transition.data[transition_id]['Protein_Name'],
            self.transition.data[transition_id][colname_theoretical_mz],
            self.transition.data[transition_id]['Precursor_Charge'],
            precursor_rt,
            self.transition.data[transition_id]['Peptide_Sequence'],
            self.transition.data[transition_id]['Ion_Series'])

    def get_transition(self, transition_id, intensity_transition, fragment_mz):
        """
        Make output of transition information
        :param transition_id:           transition ID of class Transition
        :param intensity_transition:    collected intensity of the corresponding Transition ID
        :param fragment_mz:             Fragment m/z of the transition ID
        :return: string type
        """
        filter_string = ""
        filter_string_header = ""
        if 'Filter_String' in self.transition.data[transition_id]:
            filter_string = self.transition.data[transition_id]['Filter_String']
            filter_string_header = "\tFilter_String"

        # make header for transition intensities of all the available picking methods
        header_intensities = get_header_intensity_picking_methods("PRM_Intensity:")

        header_transition = "Protein_Name\tTheoretical_MZ\tPrecursor_charge\tRetention_Time\tPeptide_Sequence\t" \
                            "Ion_Series\tTarget_Fragment_MZ\t{1}{0}".format(filter_string_header, header_intensities)

        intensity_tr = "\t".join(map(str, intensity_transition))
        transition_info = self.get_transition_info(transition_id)
        return_string = "{0}\n{1}\t{2}\t{3}\t{4}".format(header_transition, transition_info,
                                                         fragment_mz, intensity_tr, filter_string)
        return return_string

    def print_whole_data(self, out_file_name):
        """
        Write the quantified transition ions with all the detected PRM peaks into a tab-delimited text file
        :return: none
        """
        fp_out = open(out_file_name, "w")
        header = "Scan_ID\tPrecursor_MZ\tPrecursor_Charge\tRetention_Time\tFragment_MZ\t" \
                 "Intensity\tPPM\tMass_Difference\tTIC\tFilter_String\tRaw_Intensity\tNoise_Level"
        for transition_id in sort_dict_keys(self.transition.data):
            transition_info = "{0}+{1}-{2}".format(
                self.transition.data[transition_id]['Peptide_Sequence'],
                self.transition.data[transition_id]['Precursor_Charge'],
                self.transition.data[transition_id]['Ion_Series']
            )
            self.data_refined[transition_id] = {}
            for MSRun in sort_dict_keys(self.data[transition_id]):
                self.data_refined[transition_id][MSRun] = {}
                # ## Code for making every labeling groups to have same scan numbers.
                # flag_null_scan = True
                dict_scan2count = {}
                dict_label2info = {}
                list_label_group = sort_dict_keys(self.data[transition_id][MSRun])
                for label_group_name in list_label_group:
                    dict_label2info[label_group_name] = self.get_targets(
                        self.data[transition_id][MSRun][label_group_name])
                    # ## Code for making every labeling groups to have same scan numbers.
                    # if len(dict_label2info[label_group_name]) == 0:
                    #    flag_null_scan = False
                    # count detected scans in different labellings
                    for scan_id in dict_label2info[label_group_name]:
                        if scan_id not in dict_scan2count:
                            dict_scan2count[scan_id] = 1
                        else:
                            dict_scan2count[scan_id] += 1

                for label_group_name in list_label_group:
                    list_intensities = []
                    out_string = ""
                    self.data_refined[transition_id][MSRun][label_group_name] = {}
                    for scan_id in dict_scan2count:
                        # ## Code for making every labeling groups to have same scan numbers.
                        # if at least one labeling doesn't have any detected scans...
                        # if flag_null_scan:
                        #    # Check if all the labeled ions were detected in a MS/MS scan
                        #    if dict_scan2count[scan_id] != len(list_label_group):
                        #        continue

                        # Collect intensities and detected peak information
                        if scan_id in dict_label2info[label_group_name]:
                            self.data_refined[transition_id][MSRun][label_group_name][scan_id] = \
                                dict_label2info[label_group_name][scan_id]
                            intensity = dict_label2info[label_group_name][scan_id]['intensity']
                            print_out = dict_label2info[label_group_name][scan_id]['print']
                            list_intensities.append(intensity)
                            out_string = append_string(short_string=print_out, template=out_string, glue='\n')
                    intensity_transition = get_statistics_intensity(list_intensities)
                    transition_output = self.get_transition(
                        transition_id,
                        intensity_transition,
                        self.transition.data[transition_id]['Labeling'][label_group_name])
                    fp_out.write("{4}\t{3}\t{0}\n{1}\n{5}\n{2}\n\n".format(
                        transition_info, transition_output, out_string, label_group_name, MSRun, header))
        fp_out.close()


##########################
# END XPIQuant libraries #
##########################


###########################
# START XPIPeak libraries #
###########################
class UserMessageXPIPeak(Message):
    def __init__(self, message):
        """
        Print out message to users
        :param message: user message
        :return: none
        """
        self.pgm_name = "XPIPeak.py"
        self.usage = """
{0} is designed to select commonly and strongly deteced peaks over different ion-series

Usage:
    $ python XPIPick.py [data directory]
        e.g.) $ python {0} data_directory

# In the [data directory] the following files should be there
    1. {1}
    2. {2}
    3. XPIViz_config.txt
    4. XPIPeak_config.txt
    5. XPIQuant configuration file

See the manual section 8 for more information.
""".format(self.pgm_name, file_all, file_time, file_viz_config, file_peak_config)
        self.print_message(message, usage=self.usage)


def remove_noisy_tails_by_ratio(rt_lowess, min_peak_span, intensity_lowess, rt1, rt2, rt3, ratio_threshold=0.1,
                                gap_threshold=0.2, flag_debugging=False):
    """
    Find new local minima that by trimming out intensities less than 5% of peak.
    Also, if a gap between intensities is larger than 0.2 min, trim it.
    :param intensity_lowess: a list of lowes
    :param min_peak_span: minimum peak range
    :param rt_lowess: retention time of lowess data
    :param ratio_threshold: 0.01 is default
    :param rt1: starting local minima
    :param rt2: local maxima
    :param rt3: ending local minima
    :return: new boundary of peaks
    """
    start_rt_local = rt_lowess[rt1]
    end_rt_local = rt_lowess[rt3]
    if end_rt_local - start_rt_local < min_peak_span:
        return rt1, rt3

    new_start = rt1
    new_end = rt3
    for p in range(int(rt1), int(rt3)+1):
        ratio_ions = intensity_lowess[p] / intensity_lowess[int(rt2)]
        ######################
        # Code for debugging #
        if flag_debugging: print(p, new_start, new_end, ratio_ions, ratio_threshold)
        ######################        
        if p < rt2:
            if ratio_ions < ratio_threshold:
                new_start = p
        else:
            if ratio_ions >= ratio_threshold:
                new_end = p
    try:
        rt_lowess_end = rt_lowess[new_start+1]
    except IndexError:
        rt_lowess_end = rt_lowess[new_start]
    if numpy.abs(rt_lowess[new_start] - rt_lowess_end) > gap_threshold:
        new_start += 1
        if numpy.abs(rt_lowess[new_start] - rt_lowess_end) > gap_threshold:
            new_start += 1
    if numpy.abs(rt_lowess[new_end] - rt_lowess[new_end-1]) > gap_threshold:
        new_end -= 1
        if numpy.abs(rt_lowess[new_end] - rt_lowess[new_end-1]) > gap_threshold:
            new_end -= 1
    return new_start, new_end


def get_max_delta_mass_and_intensities(xpi_all_obj, tr_id, file_name, labeling_group, 
                                       mz_diff_threshold, ms2_window, rt_first, rt_last, 
                                       mode='gap_diff'):
    """
    Get maximum delate mass within given retention time
    Changed this module to return intensity and delta_mass together on 12/6/18 by Lang Ho Lee
    :param ms2_window: Delta MS/MS window
    :param mz_diff_threshold: Delta MS/MS window threshold
    :param labeling_group: Labeling group
    :param file_name: file name to get XPI_All data
    :param tr_id: transition ID to get XPI_All data
    :param xpi_all_obj: XPI_All object
    :param mode: gap_diff is default
    :param rt_first: starting retention time point
    :param rt_last: ending retention time point
    :return: int
    """
    zipped_array_local = zip(xpi_all_obj.data[tr_id][file_name][labeling_group]['rt'],
                             xpi_all_obj.data[tr_id][file_name][labeling_group]['intensity'],
                             xpi_all_obj.data[tr_id][file_name][labeling_group]['mz_diff'])
    # get mz_diff between selected retention time range
    list_mz_diff_local = []
    list_mz_intensity_local = []
    for rt_local, intensity_local, mz_diff_local in zipped_array_local:
        if float(rt_first) <= rt_local <= float(rt_last):
            list_mz_diff_local.append(mz_diff_local)
            if mz_diff_local <= ms2_window:
                list_mz_intensity_local.append(intensity_local)
    return {'mz_diff':list_mz_diff_local, 'intensity':list_mz_intensity_local}
    # This part is to detect sudden gap in one peak
    # Not used after 12/6/18 by Lang Ho Lee
    """
    # ./mzML/HDL/pt243_alpha3/01_ss_mDIA_pt243-HDL_DietB_alpha3_0-5hr.mzML, ALVQQMEQLR+2-y6
    if mode == 'gap_diff':
        # get mz_diff that have changed bigger than ms2_window/5
        list_mz_diff_local = sorted(list_mz_diff_local)
        for ii in range(1, len(list_mz_diff_local)):
            gap_difference = numpy.abs(list_mz_diff_local[ii] - list_mz_diff_local[ii - 1])
            if gap_difference >= mz_diff_threshold:
                return list_mz_diff_local[ii - 1]
    # if there's no big gap difference, return max mz_diff
    if len(list_mz_diff_local) == 0.0:
        return ms2_window
    elif numpy.max(list_mz_diff_local) > ms2_window:
        return ms2_window
    else:
        return numpy.max(list_mz_diff_local)
    """

def check_redundant_peaks(rt_window, dict_peak_detected_arg, peak_rt_arg, peak_info_arg):
    """
    If there's redundant peaks, merge them
    :param rt_window: Retention time window
    :param dict_peak_detected_arg: a dictionary containing detected peak information
    :param peak_rt_arg: retention time at maximum peak
    :param peak_info_arg: peak information being compared
    :return: a dictionary containing merged peaks
    """
    peak_rt_local = float(peak_rt_arg)
    # If there's no data ...
    if len(dict_peak_detected_arg) == 0:
        dict_peak_detected_arg[peak_rt_local] = peak_info_arg
    # Check if there's peak_rt that is within rt_window ...
    else:
        redundant_rt_within_window = 0.0
        for stored_peak_rt in dict_peak_detected_arg:
            # Check if previous stored_peak_rt is within rt_window of current peak_rt ...
            if check_retention_time(rt_window, stored_peak_rt, peak_rt_local):
                continue
            # If there is close one within rt_window...
            else:
                redundant_rt_within_window = float(stored_peak_rt)
        if redundant_rt_within_window == 0.0:
            dict_peak_detected_arg[peak_rt_local] = peak_info_arg
        else:
            merged_peak_info_local = merge_two_peaks(peak_info_arg,
                                                     dict_peak_detected_arg[redundant_rt_within_window])
            del dict_peak_detected_arg[redundant_rt_within_window]
            merged_peak_rt_local = float(merged_peak_info_local.split(delimiter_residue)[1])
            dict_peak_detected_arg[merged_peak_rt_local] = merged_peak_info_local
    return dict_peak_detected_arg


def merge_two_peaks(previous_info, current_info):
    """
    Merge two peak information
    :param previous_info: Peak information string
    :param current_info: Peak information string
    :return: string
    """
    list_peak_info_previous = previous_info.split(delimiter_residue)
    start_rt_previous = float(list_peak_info_previous[0])
    peak_rt_previous = float(list_peak_info_previous[1])
    end_rt_previous = float(list_peak_info_previous[2])
    peak_rank_previous = int(list_peak_info_previous[3])

    list_peak_info_current = current_info.split(delimiter_residue)
    start_rt_current = float(list_peak_info_current[0])
    peak_rt_current = float(list_peak_info_current[1])
    end_rt_current = float(list_peak_info_current[2])
    peak_rank_current = int(list_peak_info_current[3])

    if peak_rank_previous < peak_rank_current:
        return_intensity_rank = str(peak_rank_previous)
        return_peak_rt = str(peak_rt_previous)
    else:
        return_intensity_rank = str(peak_rank_current)
        return_peak_rt = str(peak_rt_current)
    if start_rt_previous < start_rt_current:
        return_start_rt = str(start_rt_previous)
    else:
        return_start_rt = str(start_rt_current)
    if end_rt_previous < end_rt_current:
        return_end_rt = str(end_rt_current)
    else:
        return_end_rt = str(end_rt_previous)
    return "{1}{0}{2}{0}{3}{0}{4}".format(delimiter_residue, return_start_rt, return_peak_rt, return_end_rt,
                                          return_intensity_rank)


def check_retention_time(rt_window, rt1, rt_target):
    """
    Check retention time if it is located within given rt_window
    :param rt_window: Retention time window
    :param rt1: retention time being compared
    :param rt_target: target retention time
    :return: boolean
    """
    upper_limit = float(rt1) + rt_window
    lower_limit = float(rt1) - rt_window
    return check_range(lower_limit, upper_limit, rt_target)


def merge_minima_maxima(local_maxima, local_minima):
    """
    Extend a list containing local minima data
    :return: a list
    """
    peak_list = []
    peak_list.extend(local_maxima)
    peak_list.extend(local_minima)
    peak_list.sort()
    return peak_list


def print_nd_text(ax):
    """
    Draw text "Not Detected"
    :return: none
    """
    ax.text(x=0.4, y=0.5, s="Not Detected")


class XPIPeakConfig:
    def __init__(self, file_path):
        self.file_path = file_path
        self.data = {}

        self.parse_file()

    def get_data(self, key_name):
        if key_name not in self.data:
            UserMessageXPIPeak.print_message("{0} is missing in your XPIPeak config file".format(key_name))
        else:
            return self.data[key_name]

    def parse_file(self):
        """
        Parse a config file for visualization of XPI data
        """
        fp = open(self.file_path, "rU")
        for data_line in fp:
            if data_line.strip() == "":
                continue
            if re.match("#", data_line):
                continue
            list_data = data_line.strip().split('\t')
            item = list_data[0]
            value = list_data[1]
            # LOWESS fraction
            if item == 'LOWESS fraction':
                self.data['lowess_fraction'] = float(value)
            # LOWESS it weight
            elif item == 'LOWESS weight':
                self.data['lowess_weight'] = int(value)
            # Reference label for determining retention time selection
            elif item == 'Standard label':
                self.data['reference_label'] = value
            # Retention time window to merge nearby peaks
            elif item == 'RT window':
                self.data['rt_window'] = float(value)
            # Reference label for determining retention time selection
            elif item == 'Long tail limit':
                self.data['long_tail'] = float(value)
            # Required the number of ion intensities to detect incomplete borderline peaks
            elif item == 'Borderline limit':
                self.data['borderline'] = int(value)
            # Plot option to show peak selection algorithm
            elif item == 'Show all peaks':
                self.data['debug'] = XPIVizConfig.read_flag(list_data[1])
            # If it is True, merge peaks located within rt_window
            elif item == 'Merge close peaks':
                self.data['merge'] = XPIVizConfig.read_flag(list_data[1])
            else:
                continue


def process_peak_detection(data_directory, xpi_all_obj, peak_config_obj, config_quant_obj, config_xpi_viz):
    # Create peak directory
    output_path = get_path_for_plot(data_directory, dir_peak_picking)
    # LOWESS set up
    fraction = peak_config_obj.get_data('lowess_fraction')
    it_weight = peak_config_obj.get_data('lowess_weight')
    # labeling name of unlabeled ion
    standard_label = peak_config_obj.get_data('reference_label')
    # Retention time window
    rt_window = peak_config_obj.get_data('rt_window')
    # If it is True, merge peaks located within rt_window
    flag_merge_close_peaks = peak_config_obj.get_data('merge')
    # ratio threshold to shorten long right tail
    ratio_threshold_long_tail = peak_config_obj.get_data('long_tail')
    # limit of the number of intensity peaks for borderline peak detection
    limit_borderline_peak = peak_config_obj.get_data('borderline')
    # It it is True, plot will include local maximan and minima and lowess lines
    flag_showing_selection_algorithm = peak_config_obj.get_data('debug')
    # print the program information
    ms2_window = float(config_quant_obj.data[colname_ms2_window])
    list_labeling = sort_dict_keys(config_quant_obj.labeling.data)
    # mz_diff threshold for detection of max_mz_diff
    mz_diff_threshold = ms2_window / 5
    # 'gap_diff' for detection of big gap mz_diff
    max_delta_mass_selection = 'gap_diff'
    # Default rt window to allow more retention time range to Heavy ions
    default_rt_window_for_heavy = 0.1
    # Peaks spanned longer than 'min_peak_span' will be considered for long tail removal
    min_peak_span = 0.1

    #####################################
    # Extract data from XPIOut_all file #
    #####################################
    dict_peaks = {}
    print("Processing peak data...", end='')
    for tr_id in xpi_all_obj.data:
        ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
        for file_name in xpi_all_obj.data[tr_id]:
            # Generate a dictionary dict_pep_intensities from XPI_output_all object
            for labeling_group in xpi_all_obj.data[tr_id][file_name]:
                # mz_diff = xpi_all_obj.data[tr_id][file_name][labeling_group]['mz_diff']
                # prot_name = xpi_all_obj.data[tr_id][file_name][labeling_group]['prot_name']
                intensity = xpi_all_obj.data[tr_id][file_name][labeling_group]['intensity']
                rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['rt']
                total_rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['total_rt']
                # Skip if any ions were not detected
                if len(intensity) == 0:
                    continue
                ### Added on 11/19/2018, but modified on 12/11/18
                # Detect retention time gaps (bigger than 'RT Window')
                # For example, 'RT Window' or 'rt_window' = 0.3
                # RT gap between two intensities at 15.374 and 15.785 minutes is 0.411
                # Since 0.411 is bigger than rt_window 0.3, it will be added to 'rt_gaps' list
                # 'rt_gaps' will be used for local minima addition
                prev_r = float(total_rt[0].split("_")[0])
                rt_gaps = []
                for r,i in zip(rt, intensity):
                    if numpy.abs(r - prev_r) >= rt_window:
                        for testing_rt in [prev_r,r]:
                            flag_addition = True
                            for rg in rt_gaps:
                                if numpy.abs(rg - testing_rt) < rt_window:
                                    flag_addition = False
                            if flag_addition:
                                rt_gaps.append(testing_rt)
                    prev_r = r
                #####################################################################
                # Peak smoothing, finding local maxima and minima and peak detection
                # http://statsmodels.sourceforge.net/devel/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html
                # Peak detection algorithm of other tools
                # http://strimmerlab.org/notes/mass-spectrometry.html
                #####################################################################
                smoothed_data = lowess(intensity, rt, frac=fraction, it=it_weight)
                rt_lowess = smoothed_data[0:len(smoothed_data), 0]
                intensity_lowess = smoothed_data[0:len(smoothed_data), 1]
                # Find local maxima and minima
                local_maxima = list(argrelextrema(intensity_lowess, numpy.greater)[0])
                local_minima = list(argrelextrema(intensity_lowess, numpy.less)[0])
                ######################################################################################################
                # Find if there's maxima on both sides where argrelextrema can't detect
                # 1. Add maxima for left borderline peak
                # 2. Add maxima for right borderline peak
                # 3. Add maxima by adding 0 intensity to both side that don't have any local minima and maxima points
                ######################################################################################################
                step_for_left_borderline_detection = 1
                step_for_right_borderline_detection = -1 * step_for_left_borderline_detection
                # Add maxima for left borderline peak
                if len(local_minima) > 0 and local_minima[0] > limit_borderline_peak:
                    flag_signal = True
                    prev_intensity = 0.0
                    for i in range(0, local_minima[0], step_for_left_borderline_detection):
                        if prev_intensity == 0.0:
                            prev_intensity = intensity_lowess[i]
                        else:
                            prev_intensity = intensity_lowess[i - step_for_left_borderline_detection]
                            if prev_intensity < intensity_lowess[i]:
                                flag_signal = False
                    if flag_signal:
                        local_minima.append(0)
                        local_maxima.append(1)                    
                # Add maxima for right borderline peak
                if len(local_minima) > 0 and local_minima[-1] < len(intensity_lowess) - limit_borderline_peak - 1:
                    flag_signal = True
                    prev_intensity = 0.0
                    for i in range(len(intensity_lowess) - 1, local_minima[-1], step_for_right_borderline_detection):
                        if prev_intensity == 0.0:
                            prev_intensity = intensity_lowess[i]
                        else:
                            prev_intensity = intensity_lowess[i - step_for_right_borderline_detection]
                            if prev_intensity < intensity_lowess[i]:
                                flag_signal = False
                    if flag_signal:
                        local_minima.append(len(intensity_lowess) - 1)
                        local_maxima.append(len(intensity_lowess) - 2)
                # Add maxima by adding 0 intensity to both side that don't have any local minima and maxima points
                if len(local_minima) == 0 and len(local_maxima) == 0:
                    temp = [0]
                    first_rt = rt_lowess[0]
                    # last_rt = rt_lowess[-1]
                    rt_gap = 0.001
                    rt_tmp_last = [first_rt + rt_gap]
                    rt_tmp_first = [first_rt - rt_gap]
                    intensity_lowess = numpy.append(temp, intensity_lowess)
                    intensity_lowess = numpy.append(intensity_lowess, temp)
                    rt_lowess = numpy.append(rt_tmp_first, rt_lowess)
                    rt_lowess = numpy.append(rt_lowess, rt_tmp_last)
                    local_maxima = list(argrelextrema(intensity_lowess, numpy.greater)[0])
                    local_minima = list(argrelextrema(intensity_lowess, numpy.less)[0])
                ### added on 11/19/2018, but modified on 12/11/18
                # Add local minima for the detected rt gaps
                for i in range(len(rt_lowess)):
                    r = rt_lowess[i]
                    if r in rt_gaps:
                        local_minima.append(i)
                local_minima.sort()
                ### End of update on 11/19/2018
                """
                # Debugging code
                if tr_id=="VKDLATVYVDVLK+3-b4" and labeling_group=="Light" and file_name=="./mzML/Lumos/a3/01_ss_pt286LFD_APOA1-APOE_4Da_alpha3_3hr.mzML":
                    print(rt_lowess[local_maxima])
                    print(rt_lowess[local_minima])
                    exit()
                """
                ###########################################################################################
                # START Peak detection process #
                # 1. If first merged point is not local minima, assign first retention time as local minima
                # 2. If last merged point is not local minima, assign last retention time as local minima
                # 3. Sort maxima by peak intensity
                # 4. Detect peaks by finding local maxima between local minimas
                ###########################################################################################
                # If at least one local maxima were found ...
                dict_peak_detected = {}
                if len(local_maxima) > 0:
                    # Merge minima and maxima
                    list_peaks = merge_minima_maxima(local_maxima, local_minima)
                    # If first merged point is not local minima, assign first retention time as local minima
                    if list_peaks[0] not in local_minima:
                        local_minima.append(0)
                        local_minima.sort()
                        list_peaks = merge_minima_maxima(local_maxima, local_minima)
                        list_peaks.sort()
                    # If last merged point is not local minima, assign last retention time as local minima
                    if list_peaks[-1] not in local_minima:
                        local_minima.append(len(rt_lowess) - 1)
                        local_minima.sort()
                        list_peaks = merge_minima_maxima(local_maxima, local_minima)
                        list_peaks.sort()
                    # Sort maxima by peak intensity
                    dict_maxima2order = {}
                    dict_int2max = {}
                    for i in local_maxima:
                        peak_int = intensity_lowess[i]
                        if peak_int in dict_int2max:
                            dict_int2max[peak_int] = str(i) + "," + dict_int2max[peak_int]
                        else:
                            dict_int2max[peak_int] = str(i)
                    j = 1
                    for peak_int in sorted(dict_int2max, reverse=True):
                        list_maxima = dict_int2max[peak_int].split(',')
                        for i_maxima in list_maxima:
                            dict_maxima2order[int(i_maxima)] = j
                        j += 1
                    flag_debugging = False
                    """
                    if tr_id=="VKDLATVYVDVLK+3-b4" and labeling_group=="Light" and file_name=="./mzML/Lumos/a3/01_ss_pt286LFD_APOA1-APOE_4Da_alpha3_3hr.mzML":
                        flag_debugging = True
                        ttt = []
                        for rr in list_peaks:
                            if rr in local_maxima:
                                ttt.append("max_{0}".format(rt_lowess[rr]))
                            elif rr in local_minima:
                                ttt.append("min_{0}".format(rt_lowess[rr]))
                            else:
                                ttt.append("err_{0}".format(rt_lowess[rr]))
                        print(ttt)
                    """
                    # Detect peaks by finding local maxima between local minimas
                    for i in range(len(list_peaks)):
                        if i + 1 == len(list_peaks):
                            continue
                        if list_peaks[i] not in local_maxima:
                            # print("ERROR {0} {1} {2} {3} {4} {5} {6}".format(
                            #    peptide_sequence, file_name, ion_series, labeling_group, local_minima,
                            #    local_maxima, merge_minima_maxima(local_maxima, local_minima)))
                            continue
                        if flag_debugging:
                            print(rt_lowess[list_peaks[i]])
                        # Initial location
                        i_start_rt = list_peaks[i - 1]
                        i_peak_rt = list_peaks[i]
                        i_end_rt = list_peaks[i + 1]
                        # Narrow down retention time by ratio (boundary intensity/peak intensity)
                        i_start_rt, i_end_rt = remove_noisy_tails_by_ratio(rt_lowess, min_peak_span, intensity_lowess,
                                                                           i_start_rt, i_peak_rt, i_end_rt,
                                                                           ratio_threshold_long_tail, gap_threshold=0.2, flag_debugging=False)
                        # Debugging code
                        if flag_debugging:
                            print(rt_lowess[i_start_rt], rt_lowess[i_peak_rt], rt_lowess[i_end_rt])
                            # intensity_lowess[i_start_rt], intensity_lowess[list_peaks[i]], intensity_lowess[i_end_rt])

                        # Get retention times
                        start_rt = rt_lowess[i_start_rt]
                        peak_rt = rt_lowess[i_peak_rt]
                        end_rt = rt_lowess[i_end_rt]
                        # Check if local maxima has intensity rank
                        if list_peaks[i] not in dict_maxima2order:
                            print("Missing {0} {1} {2}".format(list_peaks[i], local_maxima, dict_maxima2order))
                            peak_int = 100
                        else:
                            # rank by peak intensity at local maxima
                            peak_int = dict_maxima2order[list_peaks[i]]
                        peak_info = "{1}{0}{2}{0}{3}{0}{4}".format(delimiter_residue, start_rt, peak_rt, end_rt,
                                                                   peak_int)
                        if flag_merge_close_peaks:
                            dict_peak_detected = check_redundant_peaks(rt_window, dict_peak_detected, peak_rt,
                                                                       peak_info)
                        else:
                            if peak_rt not in dict_peak_detected:
                                dict_peak_detected[peak_rt] = peak_info
                            else:
                                dict_peak_detected = check_redundant_peaks(rt_window, dict_peak_detected, peak_rt,
                                                                           peak_info)
                # If there's no local maxima ...
                else:
                    j = 1
                    dict_int2order = {}
                    for peak_int in sorted(intensity, reverse=True):
                        dict_int2order[peak_int] = j
                        j += 1
                    for k1 in range(len(intensity)):
                        peak_int = intensity[k1]
                        peak_info = "{1}{0}{2}{0}{3}{0}{4}".format(delimiter_residue, rt[k1], rt[k1], rt[k1],
                                                                   dict_int2order[peak_int])
                        dict_peak_detected = check_redundant_peaks(rt_window, dict_peak_detected, rt[k1], peak_info)
                # Initiate dictionary for peak detection
                if peptide_sequence not in dict_peaks:
                    dict_peaks[peptide_sequence] = {}
                if file_name not in dict_peaks[peptide_sequence]:
                    dict_peaks[peptide_sequence][file_name] = {}
                if labeling_group not in dict_peaks[peptide_sequence][file_name]:
                    dict_peaks[peptide_sequence][file_name][labeling_group] = {}
                if ion_series not in dict_peaks[peptide_sequence][file_name][labeling_group]:
                    dict_peaks[peptide_sequence][file_name][labeling_group][ion_series] = {}
                # Store information at dict_peaks
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['tr'] = tr_id
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['total_rt'] = total_rt[0]
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['peaks'] = dict_peak_detected
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['rt_lowess'] = rt_lowess
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['int_sm'] = intensity_lowess
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['maxima'] = local_maxima
                dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]['minima'] = local_minima
    print(" Done!")

    """
    # For debugging
    tmp_dict = dict_peaks["VKDLATVYVDVLK+3"]["./mzML/Lumos/a3/01_ss_pt286LFD_APOA1-APOE_4Da_alpha3_3hr.mzML"]["Light"]
    for iss in tmp_dict:
        print(iss, tmp_dict[iss]['peaks'])
    exit()
    """

    #######################################################################################
    # Narrow down retention time by commonly found strong peaks & Write Peak output file
    #   1. Merge peaks that are located closely (peak rt difference < rt_window)
    #   2. Align all the peaks of every ion_series in same labeling group
    #   3. Select a peak detected most frequently and strongly
    #   4. Merge retention time range of the selected peak
    #   5. Find 'Heavy' ion peaks nearby the retention time range selected in 'Light' ions
    #######################################################################################
    print("Writing peak information and Picking effective peaks... ", end='')
    path_output = os.path.join(data_directory, file_peak_info)
    fp_out = open(path_output, "w")
    fp_out.write("File_Name\tLabeling_Group\tTransition\tIon_series\tTotal_RT\tSelected_RT\tMax_Delta_Mass\tIntensity Sum\n")

    dict_narrowed_rt = {}
    for peptide_sequence in sorted(dict_peaks):
        for file_name in sorted(dict_peaks[peptide_sequence]):
            #################################################################
            # START: Get exact retention time of standard label (Light ion) #
            #################################################################
            labeling_group = standard_label
            flag_first_entry = True
            dict_peak_tree = {}
            # If standard label is not found, skip the peptide
            if standard_label not in dict_peaks[peptide_sequence][file_name]:
                continue
            list_ion_series = sorted(dict_peaks[peptide_sequence][file_name][labeling_group])
            for ion_series in list_ion_series:
                # Retrieve data
                dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                # tr_id = dict_tmp['tr']
                # total_rt = dict_tmp['total_rt']
                # maximas = dict_tmp['maxima']
                # minimas = dict_tmp['minima']
                # rt_sorted = dict_tmp['rt_lowess']
                dict_peak_detected = dict_tmp['peaks']
                # Align peaks detected in different ion_series
                # If it is first ion_series, add all the peaks to dict_peak_tree
                if flag_first_entry:
                    for rt_peak in sorted(dict_peak_detected):
                        dict_peak_tree[rt_peak] = {}
                        dict_peak_tree[rt_peak][ion_series] = dict_peak_detected[rt_peak]
                    flag_first_entry = False
                else:
                    # If it is not first ion_series, find close rt_peak within rt_window.
                    for rt_peak in sorted(dict_peak_detected):
                        rt_peak_selected = 0.0
                        rt_peak_diff = -1
                        # Check if there is close rt_peak within rt_window
                        for stored_rt_peak in dict_peak_tree:
                            """
                            # Debugging
                            if peptide_sequence == "ATEHLSTLSEK+3" and file_name == "./mzML/Lumos/low/01_ss_pt286LFD_APOA1-APOE_4Da_low_10hr.mzML":
                                print(ion_series, stored_rt_peak, rt_peak, check_retention_time(rt_window, stored_rt_peak, rt_peak))
                            """
                            if check_retention_time(rt_window, stored_rt_peak, rt_peak):
                                continue
                            else:
                                # Get close pre-existing rt_peak within rt_window
                                diff = numpy.abs(stored_rt_peak - rt_peak)
                                if rt_peak_diff == -1:
                                    rt_peak_selected = stored_rt_peak
                                    rt_peak_diff = diff
                                elif rt_peak_diff > diff:
                                    rt_peak_selected = stored_rt_peak
                        # If there is no close rt_peak within rt_window
                        if rt_peak_selected == 0.0:
                            rt_peak_selected = rt_peak
                            dict_peak_tree[rt_peak_selected] = {}
                        # Store peak info to dict_peak_tree
                        if ion_series in dict_peak_tree[rt_peak_selected]:
                            dict_peak_tree[rt_peak_selected][ion_series] = merge_two_peaks(
                                dict_peak_tree[rt_peak_selected][ion_series], dict_peak_detected[rt_peak])
                        else:
                            dict_peak_tree[rt_peak_selected][ion_series] = dict_peak_detected[rt_peak]
            # Count # detected ion_series of peaks
            dict_freq2rt_peak = {}
            for rt_peak in dict_peak_tree:
                freq = len(dict_peak_tree[rt_peak])
                if freq not in dict_freq2rt_peak:
                    dict_freq2rt_peak[freq] = []
                dict_freq2rt_peak[freq].append(rt_peak)
            ## Modified 12/5/18 by Lang Ho Lee
            ## Choose peak retention time range that have higher intensity sum
            ## If there are two retention time ranges that have the same intensity sum, 
            ## Choose one with lower sum of delta mass difference in ion-series
            # get intensities within detected peak time ranges
            dict_peakrt_intensity = {}
            for freq in dict_freq2rt_peak:
                for rt_peak in dict_freq2rt_peak[freq]:
                    dict_peakrt_intensity[rt_peak] = {}
                    for ion_series_local in dict_peak_tree[rt_peak]:
                        peak_rt_range = dict_peak_tree[rt_peak][ion_series_local].split(delimiter_residue)
                        dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series_local]
                        dict_peakrt_intensity[rt_peak][ion_series_local] = get_max_delta_mass_and_intensities(
                            xpi_all_obj, dict_tmp['tr'], file_name, labeling_group,
                            mz_diff_threshold, ms2_window, peak_rt_range[0], peak_rt_range[2], 
                            mode=max_delta_mass_selection)
            # find one retention time range that have biggest intensity sum
            selected_rt_intensity_sum = 0
            selected_rt_deltamz_sum = 0
            selected_rt_peak = ''
            for peak_rt in dict_peakrt_intensity:
                tmp_intensity_sum = 0
                tmp_max_delta_mz_sum = 0
                for ion_series_local in dict_peakrt_intensity[peak_rt]:
                    tmp_intensity_sum += numpy.sum(dict_peakrt_intensity[peak_rt][ion_series_local]['intensity'])
                    tmp_max_delta_mz_sum  += numpy.sum(dict_peakrt_intensity[peak_rt][ion_series_local]['mz_diff'])
                ## Choose peak retention time range that have higher intensity sum
                if tmp_intensity_sum > selected_rt_intensity_sum:
                    selected_rt_peak = peak_rt
                    selected_rt_intensity_sum = tmp_intensity_sum
                    selected_rt_deltamz_sum = tmp_max_delta_mz_sum
                ## If there are two retention time ranges that have the same intensity sum, 
                ## Choose one with lower sum of delta mass difference in ion-series
                elif tmp_intensity_sum == selected_rt_intensity_sum:
                    if tmp_max_delta_mz_sum < selected_rt_deltamz_sum:
                        selected_rt_peak = peak_rt
                        selected_rt_intensity_sum = tmp_intensity_sum
                        selected_rt_deltamz_sum = tmp_max_delta_mz_sum
            """
            # Modified 02/08/16 by Lang Ho Lee
            # Find peak retention time that have more number of rank 1 intensity peak.
            # If there are more than two peak retention time that have same number of rank 1,
            #  compare sum of ranks and select peak retention time that have less sum of ranks.
            # Rank 1: most intense peak
            # Modified 03/03/16 by Lang Ho Lee
            # If the number of rank 1 is same, consider identification frequency in advance to rank sum
            selected_rt_freq = 0
            sum_of_peak_int_selected = 0.0
            max_num_rank1 = 0
            selected_rt_peak = ''
            for freq in sorted(dict_freq2rt_peak, reverse=True):
                for rt_peak in dict_freq2rt_peak[freq]:
                    sum_of_peak_int = 0
                    count_rank1 = 0
                    for ion_series in list_ion_series:
                        if ion_series in dict_peak_tree[rt_peak]:
                            peak_int = int(dict_peak_tree[rt_peak][ion_series].split(delimiter_residue)[3])
                            sum_of_peak_int += peak_int
                            if peak_int == 1:
                                count_rank1 += 1
                    if max_num_rank1 == 0.0:
                        max_num_rank1 = count_rank1
                        selected_rt_peak = rt_peak
                        sum_of_peak_int_selected = sum_of_peak_int
                        selected_rt_freq = freq
                    # Take rt_peak that have bigger rank1_count
                    elif count_rank1 > max_num_rank1:
                        max_num_rank1 = count_rank1
                        selected_rt_peak = rt_peak
                        # in_sum_of_peak = sum_of_peak_int
                        selected_rt_freq = freq
                    # If rank1_count is same,
                    elif count_rank1 == max_num_rank1:
                        # code added on 03/03/16 instead of 02/08/16 code
                        # Pick more frequently identified rt_peak
                        if selected_rt_freq < freq:
                            max_num_rank1 = count_rank1
                            selected_rt_peak = rt_peak
                            sum_of_peak_int_selected = sum_of_peak_int
                            selected_rt_freq = freq
                        elif selected_rt_freq == freq:
                            # If rt_freq is same, pick rt_pick of lower rank_sum
                            # sum_of_peak_int = sum of peak rank
                            # avg_intensity_current = sum_of_peak_int/freq
                            # avg_intensity_prev = sum_of_peak_int_selected/selected_rt_freq
                            if sum_of_peak_int_selected > sum_of_peak_int:
                                max_num_rank1 = count_rank1
                                selected_rt_peak = rt_peak
                                sum_of_peak_int_selected = sum_of_peak_int
                                selected_rt_freq = freq
                        # code added in 02/08/16
                        if sum_of_peak_int_selected > sum_of_peak_int:
                            max_num_rank1 = count_rank1
                            selected_rt_peak = rt_peak
                            sum_of_peak_int_selected = sum_of_peak_int
                            selected_rt_freq = freq
            """
            """
            # Original 01/22/16
            # Pick best peak that has best sum of intensity rank in most frequently detected peaks
            # Rank 1 is better than rank 3
            sum_of_peak_int_selected = 0.0
            selected_rt_peak = ''
            sorted_freq = sorted(dict_freq2rt_peak, reverse=True)
            for rt_peak in dict_freq2rt_peak[sorted_freq[0]]:
                sum_of_peak_int = 0
                for ion_series in list_ion_series:
                    if ion_series in dict_peak_tree[rt_peak]:
                        peak_int = int(dict_peak_tree[rt_peak][ion_series].split(delimiter_residue)[3])
                        sum_of_peak_int += peak_int
                if sum_of_peak_int_selected == 0.0:
                    sum_of_peak_int_selected = sum_of_peak_int
                    selected_rt_peak = rt_peak
                elif sum_of_peak_int < sum_of_peak_int_selected:
                    sum_of_peak_int_selected = sum_of_peak_int
                    selected_rt_peak = rt_peak
            """
            # Updated by Lang Ho Lee on 12/6/18
            # Replace the selected peak time ranges with the reference rt
            # The reference RT can't be over 10 times of "RT window", e.g. if RT window 0.05, it is 1 min.
            # This limit will be separately applied to starting and ending RT.
            reference_rt_start = -1.0
            reference_rt_end = -1.0
            # Get the reference RT
            for ion_series in dict_peak_tree[selected_rt_peak]:
                list_rt = dict_peak_tree[selected_rt_peak][ion_series].split(delimiter_residue)
                rt_start = float(list_rt[0])
                rt_end = float(list_rt[2])
                if reference_rt_start == -1.0:
                    reference_rt_start = rt_start
                elif reference_rt_start > rt_start:
                    reference_rt_start = rt_start
                if reference_rt_end == -1.0:
                    reference_rt_end = rt_end
                elif reference_rt_end < rt_end:
                    reference_rt_end = rt_end
            rt_limit = rt_window*10
            if reference_rt_start < selected_rt_peak - rt_limit:
                reference_rt_start = selected_rt_peak - rt_limit
            if reference_rt_end > selected_rt_peak + rt_limit:
                reference_rt_start = selected_rt_peak + rt_limit
            # write the peak picking results
            for ion_series in list_ion_series:
                if ion_series in dict_peak_tree[selected_rt_peak]:
                    list_rt = dict_peak_tree[selected_rt_peak][ion_series].split(delimiter_residue)
                    peak_rt_selected = list_rt[1]
                else:
                    peak_rt_selected = selected_rt_peak
                dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                tr_id = dict_tmp['tr']
                total_rt = dict_tmp['total_rt']
                # Get largest delta ms/ms intensity within selected retention time
                tmp_results = get_max_delta_mass_and_intensities(xpi_all_obj, tr_id, file_name, labeling_group,
                                                                    mz_diff_threshold, ms2_window,
                                                                    reference_rt_start, reference_rt_end, 
                                                                    mode=max_delta_mass_selection)
                try:
                    max_delta_mass = numpy.max(tmp_results['mz_diff'])
                    intensity_sum = numpy.sum(tmp_results['intensity'])
                # This is to handle a case with no ions in the given retention time range
                except ValueError:
                    max_delta_mass = ms2_window
                    intensity_sum = 0.0
                # Collect all the detected peaks for debugging
                # for prt in sorted(dict_tmp['peaks']):
                #     peak_all = peak_all + '\t' + dict_tmp['peaks'][prt]
                peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, reference_rt_start, peak_rt_selected, reference_rt_end)
                fp_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\n".format(file_name, labeling_group, tr_id,
                ion_series, total_rt, peak_range, str(max_delta_mass), str(intensity_sum)))
                # Save retention time into fo dict_narrowed_rt
                if peptide_sequence not in dict_narrowed_rt:
                    dict_narrowed_rt[peptide_sequence] = {}
                if file_name not in dict_narrowed_rt[peptide_sequence]:
                    dict_narrowed_rt[peptide_sequence][file_name] = {}
                if labeling_group not in dict_narrowed_rt[peptide_sequence][file_name]:
                    dict_narrowed_rt[peptide_sequence][file_name][labeling_group] = {}
                if ion_series not in dict_narrowed_rt[peptide_sequence][file_name][labeling_group]:
                    dict_narrowed_rt[peptide_sequence][file_name][labeling_group][ion_series] = \
                        {'peak_range': peak_range, 'max_delta_mass': max_delta_mass, 'intensity_sum':intensity_sum}
            
            """
            # Original algorithm (report different peak retention time by ion-series)
            # Write selected peak information and add peak information to dict_narrowed_rt
            # Get reference rt for other labelings by combining earlist rt and latest rt in every ion-series
            # Get reference rt based on the selected peak time ranges from ion series
            list_missed_ions = []
            for ion_series in list_ion_series:
                if ion_series in dict_peak_tree[selected_rt_peak]:
                    # Merge retention time of selected Light ion series
                    list_rt = dict_peak_tree[selected_rt_peak][ion_series].split(delimiter_residue)
                    rt_start = float(list_rt[0])
                    rt_end = float(list_rt[2])
                    if reference_rt_start == 0.0:
                        reference_rt_start = rt_start
                    elif reference_rt_start > rt_start:
                        reference_rt_start = rt_start
                    if reference_rt_end == 0.0:
                        reference_rt_end = rt_end
                    elif reference_rt_end < rt_end:
                        reference_rt_end = rt_end
                    # Write peak detection
                    dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                    tr_id = dict_tmp['tr']
                    total_rt = dict_tmp['total_rt']
                    # maximas = dict_tmp['maxima']
                    # minimas = dict_tmp['minima']
                    # rt_sorted = dict_tmp['rt_lowess']

                    # Get largest delta ms/ms intensity within selected retention time
                    max_delta_mass = get_max_delta_mass_and_intensities(xpi_all_obj, tr_id, file_name, labeling_group,
                                                        mz_diff_threshold, ms2_window,
                                                        list_rt[0], list_rt[2], mode=max_delta_mass_selection)
                    # Collect all the detected peaks for debugging
                    # for prt in sorted(dict_tmp['peaks']):
                    #     peak_all = peak_all + '\t' + dict_tmp['peaks'][prt]
                    peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, list_rt[0], list_rt[1], list_rt[2])
                    fp_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(file_name, labeling_group, tr_id,
                                                                              ion_series, total_rt, peak_range,
                                                                              str(max_delta_mass)))
                    # Save retention time into fo dict_narrowed_rt
                    if peptide_sequence not in dict_narrowed_rt:
                        dict_narrowed_rt[peptide_sequence] = {}
                    if file_name not in dict_narrowed_rt[peptide_sequence]:
                        dict_narrowed_rt[peptide_sequence][file_name] = {}
                    if labeling_group not in dict_narrowed_rt[peptide_sequence][file_name]:
                        dict_narrowed_rt[peptide_sequence][file_name][labeling_group] = {}
                    if ion_series not in dict_narrowed_rt[peptide_sequence][file_name][labeling_group]:
                        dict_narrowed_rt[peptide_sequence][file_name][labeling_group][ion_series] = \
                            {'peak_range': peak_range, 'max_delta_mass': max_delta_mass}
                else:
                    list_missed_ions.append(ion_series)

            # If there is missed ion (good peak is not detected within reference rt),
            # put reference retention time to missed ion series (Peak was not detected)
            for ion_series in list_missed_ions:
                # Write peak detection
                dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                tr_id = dict_tmp['tr']
                total_rt = dict_tmp['total_rt']
                # maximas = dict_tmp['maxima']
                # minimas = dict_tmp['minima']
                # rt_sorted = dict_tmp['rt_lowess']
                peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, str(reference_rt_start),
                                                      'NA', str(reference_rt_end))
                # Get largest delta ms/ms intensity within selected retention time
                max_delta_mass = get_max_delta_mass_and_intensities(xpi_all_obj, tr_id, file_name, labeling_group, mz_diff_threshold,
                                                    ms2_window,
                                                    reference_rt_start, reference_rt_end, mode=max_delta_mass_selection)
                fp_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(file_name, labeling_group, tr_id, ion_series,
                                                                          total_rt, peak_range, str(max_delta_mass)))
                # Save retention time into fo dict_narrowed_rt
                if peptide_sequence not in dict_narrowed_rt:
                    dict_narrowed_rt[peptide_sequence] = {}
                if file_name not in dict_narrowed_rt[peptide_sequence]:
                    dict_narrowed_rt[peptide_sequence][file_name] = {}
                if labeling_group not in dict_narrowed_rt[peptide_sequence][file_name]:
                    dict_narrowed_rt[peptide_sequence][file_name][labeling_group] = {}
                if ion_series not in dict_narrowed_rt[peptide_sequence][file_name][labeling_group]:
                    dict_narrowed_rt[peptide_sequence][file_name][labeling_group][ion_series] = \
                        {'peak_range': peak_range, 'max_delta_mass': max_delta_mass}
            """
            #####################################################
            # E-N-D: Get exact retention time of standard label #
            #####################################################

            #######################################################################################
            # START: Find exact retention time of Heavy ion considering RT of reference Light ion #
            #######################################################################################
            for labeling_group in dict_peaks[peptide_sequence][file_name]:
                if labeling_group == standard_label:
                    continue
                # for ion_series in dict_peaks[peptide_sequence][file_name][labeling_group]:
                for ion_series in list_ion_series:
                    if ion_series not in dict_peaks[peptide_sequence][file_name][labeling_group]:
                        continue
                    # Retrieve data of peak information
                    dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                    tr_id = dict_tmp['tr']
                    total_rt = dict_tmp['total_rt']
                    total_rt_low = float(total_rt.split("_")[0])
                    # total_rt_high = float(total_rt.split("_")[1])
                    dict_peak_detected = dict_tmp['peaks']
                    # Retrieve data from standard label
                    # if there's no standard label for this ion-series, use reference rt
                    if ion_series not in dict_narrowed_rt[peptide_sequence][file_name][standard_label]:
                        rt_start_standard = reference_rt_start
                        rt_end_standard = reference_rt_end
                    # if there's detected ion-series in standard label, use rt of standard label
                    else:
                        dict_nr_tmp = dict_narrowed_rt[peptide_sequence][file_name][standard_label][ion_series]
                        list_rt_standard = dict_nr_tmp['peak_range'].split(delimiter_residue)
                        rt_start_standard = float(list_rt_standard[0])
                        rt_end_standard = float(list_rt_standard[2])
                    # Find peaks within reference retention time range
                    peak_range = ''
                    if rt_window == 0.0:
                        rt_start_for_heavy = rt_start_standard - default_rt_window_for_heavy
                    else:
                        rt_start_for_heavy = rt_start_standard - rt_window
                    # rt_start_for_heavy cannot be lower than total_rt_low
                    if rt_start_for_heavy < total_rt_low:
                        rt_start_for_heavy = total_rt_low
                    rt_end_for_heavy = rt_end_standard
                    # Peak finding located between rt_start_for_heavy and rt_end_for_heavy
                    for peak_rt in dict_peak_detected:
                        list_rt = dict_peak_detected[peak_rt].split(delimiter_residue)
                        # rt_start = float(list_rt[0])
                        # rt_end = float(list_rt[2])
                        rt_peak = float(list_rt[1])
                        if rt_peak < rt_start_for_heavy:
                            continue
                        if rt_peak > rt_end_for_heavy:
                            continue
                        if peak_range == '':
                            peak_range = dict_peak_detected[peak_rt]
                        else:
                            peak_range = merge_two_peaks(peak_range, dict_peak_detected[peak_rt])

                    # Force to all the Heavy ions to rt_end_for_heavy and rt_start_for_heavy
                    # Modified by Lang Ho Lee on 02/10/16
                    if peak_range == '':
                        peak_rt_heavy = 'NA'
                    else:
                        peak_rt_heavy = peak_range.split(delimiter_residue)[1]
                    peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, rt_start_for_heavy,
                                                          peak_rt_heavy, rt_end_for_heavy)
                    """
                    # Original code finding retention time of the detected peak
                    #   between rt_end_for_heavy and rt_start_for_heavy
                    # Modified by Lang Ho Lee on 01/31/16
                    # If none of Heavy ion peak is found within reference retention time ...
                    if peak_range == '':
                        peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, str(rt_start_for_heavy),
                                                              "NA", str(rt_end_for_heavy))
                    # If peaks were detected,
                    # rearrange peak range by considering rt_end_for_heavy and rt_start_for_heavy
                    else:
                        list_rt = peak_range.split(delimiter_residue)
                        rt_start = float(list_rt[0])
                        rt_end = float(list_rt[2])
                        # To limit retention time to rt_start
                        if rt_start > rt_start_for_heavy:
                            rt_start_for_heavy = rt_start
                        # To limit retention time to rt_end
                        if rt_end < rt_end_for_heavy:
                            rt_end_for_heavy = rt_end
                        # To enlarge retention time to rt_start
                        # if rt_start < rt_start_for_heavy:
                        #     rt_start_for_heavy = rt_start
                        # To enlarge retention time to rt_end
                        # if rt_end > rt_end_for_heavy:
                        #     rt_end_for_heavy = rt_end
                        peak_range = "{1}{0}{2}{0}{3}".format(delimiter_residue, rt_start_for_heavy, list_rt[1],
                                                              rt_end_for_heavy)
                    """
                    # Get max_delta_mass of standard_label
                    max_delta_mass = \
                        dict_narrowed_rt[peptide_sequence][file_name][standard_label][ion_series]['max_delta_mass']
                    # max_delta_mass += 0.001
                    fp_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(file_name, labeling_group, tr_id,
                                                                              ion_series, total_rt, peak_range,
                                                                              str(max_delta_mass)))
                    # Save retention time into fo dict_narrowed_rt
                    if peptide_sequence not in dict_narrowed_rt:
                        dict_narrowed_rt[peptide_sequence] = {}
                    if file_name not in dict_narrowed_rt[peptide_sequence]:
                        dict_narrowed_rt[peptide_sequence][file_name] = {}
                    if labeling_group not in dict_narrowed_rt[peptide_sequence][file_name]:
                        dict_narrowed_rt[peptide_sequence][file_name][labeling_group] = {}
                    if ion_series not in dict_narrowed_rt[peptide_sequence][file_name][labeling_group]:
                        dict_narrowed_rt[peptide_sequence][file_name][labeling_group][ion_series] = \
                            {'peak_range': peak_range, 'max_delta_mass': max_delta_mass}
                        #######################################################################################
                        # E-N-D: Find exact retention time of Heavy ion considering RT of reference Light ion #
                        #######################################################################################
    fp_out.close()
    print("Done!")
    ######################################
    # E-N-D of RT Narrowing down process #
    ######################################


    ###########################################################################
    # START: Applying peak picking results to XPI outputs (file_check) #
    ###########################################################################
    file_check_path = os.path.join(data_directory, file_check)
    # file_path_check_prev_peaks = os.path.join(data_directory, file_check_prev_peaks)
    # Generate header for XPI_output_4check.txt
    # picking_method = config_xpi_viz.intensity_picking_method
    header_quant = ''
    for label_group_name in list_labeling:
        for method in list_intensity_picking_method:
            # Extend header for existing intensity picking methods
            header_quant = append_string(short_string="{0}:{1}".format(label_group_name, method),
                                         template=header_quant, glue='\t')
    fp_out = open(file_check_path, "w")
    fp_out.write("PRM_Run_Name\tTransition_ID\tProtein_Name\t" + header_quant + "\n")
    # Write XPI_output_4check.txt file with applying peak picking process
    for tr_id in sorted(xpi_all_obj.data):
        ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
        for file_name in sorted(xpi_all_obj.data[tr_id]):
            prot_name = ''
            out_string = ''
            # Check if the peptide was not found in the given XPIPeak output
            if peptide_sequence not in dict_narrowed_rt:
                continue
            # Check if the peptide was not found in the given file
            if not file_name in dict_narrowed_rt[peptide_sequence]:
                continue
            # Retrieve data from xpi_all_obj
            for label_group_name in list_labeling:
                dict_out_all = xpi_all_obj.data[tr_id][file_name][label_group_name]
                dict_tmp = get_filtered_intensities(dict_out_all, dict_narrowed_rt, peptide_sequence, file_name,
                                                    label_group_name, ion_series)
                prot_name = xpi_all_obj.tr2prot[tr_id]
                new_intensity_list = dict_tmp['intensity']
                # new_rt_list = dict_tmp['rt']
                # new_mz_diff_list = dict_tmp['mz_diff']
                # Quantification with various options
                out_string += "\t{0}".format(
                    "\t".join(map(str, get_statistics_intensity(new_intensity_list))))
            fp_out.write("{0}\t{1}\t{2}{3}\n".format(file_name, tr_id, prot_name, out_string))
    fp_out.close()
    ###########################################################################
    # E-N-D: Applying peak picking results to XPI outputs (file_check) #
    ###########################################################################


    #######################
    # START Drawing plots #
    #######################
    # Prepare progress bar
    pg_total = 0
    for peptide_sequence in dict_peaks:
        for file_name in dict_peaks[peptide_sequence]:
            pg_total += 1
    pg_num = 1
    pg = Progress(pg_total, "Drawing plots...")
    for peptide_sequence in sorted(dict_peaks):
        for file_name in sorted(dict_peaks[peptide_sequence]):
            metrics = "{0} {1}".format(config_xpi_viz.data[file_name]['MR'],
                                       config_xpi_viz.data[file_name]['Unit'])
            # Check if standard label ions were not detected in the peptide,
            if standard_label not in dict_peaks[peptide_sequence][file_name]:
                continue

            # Get how many plots will be drown in one figure
            i_th = 0
            num_labelings = 0
            for labeling_group in list_labeling:
                for ion_series in dict_peaks[peptide_sequence][file_name][standard_label]:
                    num_labelings += 1
            span_plots = 40
            span_colorbar = 1

            # Prepare color scales
            spectral_scheme = plt.get_cmap('winter')
            c_norm = colors.Normalize(vmin=0, vmax=ms2_window)
            tick_colorbar = [0, ms2_window]
            scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=spectral_scheme)

            # Prepare pyplot drawing
            plt.figure(figsize=(10, num_labelings * 3))
            for labeling_group in sorted(list_labeling, reverse=True):
                for ion_series in sorted(dict_peaks[peptide_sequence][file_name][standard_label]):
                    # Set plot panels
                    ax = plt.subplot2grid((num_labelings, span_colorbar + span_plots), (i_th, 0), colspan=span_plots)
                    # If labeling group is not detected
                    if labeling_group not in dict_peaks[peptide_sequence][file_name]:
                        print_nd_text(ax)
                    # If ion series is not detected ...
                    elif ion_series not in dict_peaks[peptide_sequence][file_name][labeling_group]:
                        print_nd_text(ax)
                    # If ion series is detected ...
                    elif ion_series in dict_peaks[peptide_sequence][file_name][labeling_group]:
                        tr_id = "{0}{1}{2}".format(peptide_sequence, delimiter_dash, ion_series)
                        dict_out_all = xpi_all_obj.data[tr_id][file_name][labeling_group]
                        # Set X axis limit
                        total_rt = dict_out_all['total_rt'][0]
                        low = float(total_rt.split("_")[0])
                        high = float(total_rt.split("_")[1])
                        # Show all the intensities, local miniman & maxima and lowess line
                        if flag_showing_selection_algorithm:
                            ###########################################################################
                            # START: Debugging code to check LOWESS line and local miniman and maxima #
                            ###########################################################################
                            if numpy.sum(dict_out_all['intensity']) > 0:
                                zipped_array = zip(dict_out_all['rt'], dict_out_all['intensity'],
                                                   dict_out_all['mz_diff'], dict_out_all['mz'])
                                for rt, intensity, mz_diff, mz in zipped_array:
                                    color_value = scalar_map.to_rgba(mz_diff)
                                    ax.plot([rt, rt], [0, intensity], '-', color=color_value)
                                # Draw Selected peak line
                                max_intensity = numpy.max(dict_out_all['intensity'])
                                dict_tmp = dict_narrowed_rt[peptide_sequence][file_name][labeling_group][ion_series]
                                rt_selected = dict_tmp['peak_range']
                                list_rt_selected = rt_selected.split(delimiter_residue)
                                rt_start = float(list_rt_selected[0])
                                rt_end = float(list_rt_selected[2])
                                ax.plot([rt_start, rt_start], [0, max_intensity], '-', color='r', alpha=0.5)
                                ax.plot([rt_end, rt_end], [0, max_intensity], '-', color='r', alpha=0.5)
                                # Draw Lowess line after converting negative value to zero
                                dict_tmp = dict_peaks[peptide_sequence][file_name][labeling_group][ion_series]
                                maximas = dict_tmp['maxima']
                                minimas = dict_tmp['minima']
                                rt_sorted = dict_tmp['rt_lowess']
                                smoothed_intensity = dict_tmp['int_sm']
                                # dict_peak_detected = dict_tmp['peaks']
                                smoothed_intensity_without_negative = []
                                for intensity_before in smoothed_intensity:
                                    if intensity_before < 0:
                                        smoothed_intensity_without_negative.append(0.0)
                                    else:
                                        smoothed_intensity_without_negative.append(intensity_before)
                                ax.plot(rt_sorted, smoothed_intensity_without_negative, 'b-')
                                # Plot local maxima points
                                for l in range(len(maximas)):
                                    j = maximas[l]
                                    max_int_before = smoothed_intensity[j]
                                    if smoothed_intensity[j] < 0:
                                        max_int_before = 0
                                    ax.scatter(rt_sorted[j], max_int_before, color='r')
                                # Plot local minima points
                                for l in range(len(minimas)):
                                    j = minimas[l]
                                    min_int_before = smoothed_intensity[j]
                                    if smoothed_intensity[j] < 0:
                                        min_int_before = 0
                                    ax.scatter(rt_sorted[j], min_int_before, color='y')
                                # Set X-axis min and max
                                ax.set_xlim([low, high])
                            else:
                                print_nd_text(ax)
                                ###########################################################################
                                # E-N-D: Debugging code to check LOWESS line and local miniman and maxima #
                                ###########################################################################
                        else:
                            # Get selected peak information
                            dict_filtered = get_filtered_intensities(dict_out_all, dict_narrowed_rt,
                                                                     peptide_sequence, file_name,
                                                                     labeling_group, ion_series)
                            zipped_array = zip(dict_filtered['rt'], dict_filtered['intensity'],
                                               dict_filtered['mz_diff'])
                            # If there's at least one peak that passed threshold ...
                            if len(dict_filtered['intensity']) > 0:
                                # Draw intensity peaks
                                for rt, intensity, mz_diff in zipped_array:
                                    color_value = scalar_map.to_rgba(mz_diff)
                                    ax.plot([rt, rt], [0, intensity], '-', color=color_value)
                                # Set X-axis min and max
                                ax.set_xlim([low, high])
                            else:
                                print_nd_text(ax)
                    # For other unexpected errors
                    else:
                        print_nd_text(ax)
                    # Y axis ticks will be shown like scientific numbers
                    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                    ax.ticklabel_format(axis='x', useOffset=False)
                    # Set labels
                    # ax.set_ylabel('Peak intensity', fontsize=font_size_label_2d)
                    ax.tick_params(axis='both', labelsize=font_size_ticks_2d)
                    # If it is first plot, set plot title
                    if i_th == 0:
                        ax.set_title("{0} {1}\n{2} {3}".format(peptide_sequence, metrics, ion_series,
                                                               labeling_group))
                    # If it is last plot, set X-axis label
                    elif i_th == num_labelings - 1:
                        ax.set_title("{0} {1}".format(ion_series, labeling_group))
                        ax.set_xlabel('Retention Time')
                    else:
                        ax.set_title("{0} {1}".format(ion_series, labeling_group))
                    color_bar = plt.subplot2grid((num_labelings, span_colorbar + span_plots), (i_th, span_plots),
                                                 colspan=span_colorbar)
                    cb.ColorbarBase(color_bar, norm=c_norm, cmap=spectral_scheme, ticks=tick_colorbar)
                    i_th += 1

            # Plot color bar and Save the plot
            img_file_name = change_filename("{0}_{1}.png".format(peptide_sequence, metrics))
            img_file_path = os.path.join(output_path, img_file_name)
            if os.path.exists(img_file_path):
                img_file_name = change_filename("{0}_{1}_{2}.png".format(peptide_sequence, metrics,
                                                                         file_name))
                img_file_path = os.path.join(output_path, img_file_name)
            plt.savefig(img_file_path, dpi=config_xpi_viz.dpi)
            plt.close()
            pg.print_progress(pg_num)
            pg_num += 1
    print("Drawing peak picking profile... Done!")
    #######################
    # E-N-D Drawing plots #
    #######################
#########################
# END XPIPeak libraries #
#########################


##########################
# START XPIViz libraries #
##########################
class UserMessageXPIViz(Message):
    def __init__(self, message):
        self.usage = """
{0}
{1}

Usage:
    $ python XPIViz.py [data directory] [mode]
        e.g.) $ python XPIViz.py data_directory {2}

# In the [data directory] the following files should be there
    1. {9}
    2. {10}
    3. {11}
    4. {12}
    5. {13}
    6. {14}
    7. {15}
    8. XPIQuant configuration file

# In the [mode] the following files should be there
    1. {2}: Draw 3D Mass profile for every ion-series in one plot
    2. {3}: Draw 2D Mass profile for each fragment ions
    3. {4}: Draw scatter, filling and error bar plots for each protein
    4. {5}: Draw filling and line graphs for each peptide
    5. {6}: Draw scatter plot with regression line with Pearson's r
    6. {7}: Draw validation plots (box plots and ion filtering plots)
    7. {8}: Write XPI output report
""".format(xpi_viz_config, xpi_quant_config,
           mode_for_3d_profile,
           mode_for_2d_profile,
           mode_for_scatter_protein,
           mode_for_scatter_peptide,
           mode_for_standard_curve,
           mode_for_filter,
           mode_for_output, 
           file_all, file_check, file_time, file_peak_info, file_transition, file_viz_config, file_peak_config)
        self.print_message(message, usage=self.usage)


def draw_line_mass_profiles(xpi_all_obj, config_xpi_viz, xpi_peak_obj, dir_line_graph, ms2_window, list_labels):
    file_check_path = os.path.join(dir_line_graph, file_refined)
    fp_out = open(file_check_path, "w")
    header_quant = ''
    for label_group_name in list_labels:
        for method in list_intensity_picking_method:
            # Extend header for existing intensity picking methods
            header_quant = append_string(short_string="{0}:{1}".format(label_group_name, method),
                                         template=header_quant, glue=',')
    fp_out.write("PRM_Run_Name,Transition_ID,Protein_Name," + header_quant + "\n")
    all_ion_series = []
    for tr_id in xpi_all_obj.tr2prot:
        ion_series = tr_id.split("-")[1]
        if ion_series not in all_ion_series:
            all_ion_series.append(ion_series)
    # Parse xpi_all_obj and get dictionary containing all the parsed information
    dict_pep_intensities = parse_xpi_all_obj(xpi_all_obj)
    for peptide_sequence in sorted(dict_pep_intensities):
        for file_name in sorted(dict_pep_intensities[peptide_sequence]):
            plot_title = "{0} {1} {2}".format(peptide_sequence, 
                                        config_xpi_viz.data[file_name]['MR'],
                                        config_xpi_viz.data[file_name]['Unit'])
            xpi_peak_data = xpi_peak_obj.data[peptide_sequence][file_name]
            dict_peptide2data = dict_pep_intensities[peptide_sequence][file_name]
            dict_quant_ions = draw_line_graphs(dict_peptide2data, dir_line_graph, ms2_window, plot_title,
                                            config_xpi_viz.dpi, list_labels, xpi_peak_data, file_name, all_ion_series)
            for ion_series in dict_quant_ions:
                # AGLQVYNK+2-b2
                tr_id = "{0}-{1}".format(peptide_sequence, ion_series)
                prot_name = xpi_all_obj.tr2prot[tr_id]
                out_string = ""
                for label_group_name in list_labels:
                    new_intensity_list = dict_quant_ions[ion_series][label_group_name]
                    out_string += ",{0}".format(
                        ",".join(map(str, get_statistics_intensity(new_intensity_list))))
                fp_out.write("{0},{1},{2}{3}\n".format(file_name, tr_id, prot_name, out_string))
    fp_out.close()


def draw_3d_mass_profile(xpi_all_obj, config_xpi_viz, xpi_peak_obj, dir_3d_mass_profile, ms2_window, list_labels):
    # Parse xpi_all_obj and get dictionary containing all the parsed information
    dict_pep_intensities = parse_xpi_all_obj(xpi_all_obj)
    # Calculate the total number of images
    progress_total = 0
    for peptide_sequence in dict_pep_intensities:
        progress_total += len(list(dict_pep_intensities[peptide_sequence].keys()))
    report = ""
    # Draw 3D plots
    pg_num = 1
    pg = Progress(progress_total, "Drawing 3D-plots...")
    for peptide_sequence in sorted(dict_pep_intensities):
        for file_name in sorted(dict_pep_intensities[peptide_sequence]):
            metrics = "{0} {1}".format(config_xpi_viz.data[file_name]['MR'],
                                       config_xpi_viz.data[file_name]['Unit'])
            # Check if a peptide was detected in specific sample
            if peptide_sequence not in xpi_peak_obj.data:
                continue
            if file_name not in xpi_peak_obj.data[peptide_sequence]:
                continue
            xpi_peak_data = xpi_peak_obj.data[peptide_sequence][file_name]
            pg.print_progress(pg_num)
            pg_num += 1
            dict_peptide2data = dict_pep_intensities[peptide_sequence][file_name]
            plot_title = "{0}_{1}".format(peptide_sequence, metrics)
            cnt = draw_3d_plot_all(dict_peptide2data, dir_3d_mass_profile, ms2_window, plot_title,
                                   config_xpi_viz.dpi, list_labels, xpi_peak_data, file_name,
                                   config_xpi_viz.flag_same_z)
            for ion_series in cnt:
                num_intensities = cnt[ion_series]
                report = append_string("{0}\t{1}\t{2}\t{3}\n".format(peptide_sequence, file_name, ion_series,
                                                                     num_intensities), report, "")
    # Print out # of PRM ions
    # fp_out = open(os.path.join(dir_3d_mass_profile, "Peak_count_info.xls"), "w")
    # fp_out.write(report)
    # fp_out.close()


def draw_2d_mass_profile(xpi_all_obj, config_xpi_viz, list_labels, xpi_peak_obj, ms2_window, dir_2d_mass_profile):
    # Parse xpi_all_obj and get dictionary containing all the parsed information
    dict_pep_intensities = {}
    for tr_id in xpi_all_obj.data:
        ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)

        # If peptide was not detected in PRM data
        if peptide_sequence not in xpi_peak_obj.data:
            continue

        for file_name in xpi_all_obj.data[tr_id]:
            metrics = "{0} {1}".format(config_xpi_viz.data[file_name]['MR'],
                                       config_xpi_viz.data[file_name]['Unit'])
            # Initialize dict_pep_intensities
            if peptide_sequence not in dict_pep_intensities:
                dict_pep_intensities[peptide_sequence] = {}
            if metrics not in dict_pep_intensities[peptide_sequence]:
                dict_pep_intensities[peptide_sequence][metrics] = {}
            if ion_series not in dict_pep_intensities[peptide_sequence][metrics]:
                dict_pep_intensities[peptide_sequence][metrics][ion_series] = {}
            # ratio_name = config_xpi_viz.data[file_name]['RatioName']
            # Generate a dictionary dict_pep_intensities from XPI_output_all object
            # for labeling_group in xpi_all_obj.data[tr_id][file_name]:
            for labeling_group in list_labels:
                if labeling_group in xpi_all_obj.data[tr_id][file_name]:
                    list_ints = xpi_all_obj.data[tr_id][file_name][labeling_group]['intensity']
                    list_mz_diff = xpi_all_obj.data[tr_id][file_name][labeling_group]['mz_diff']
                    list_rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['rt']
                    list_total_rt = xpi_all_obj.data[tr_id][file_name][labeling_group]['total_rt']
                else:
                    list_ints = []
                    list_mz_diff = []
                    list_rt = []
                    list_total_rt = []
                # retrieve peak selection data
                if labeling_group not in xpi_peak_obj.data[peptide_sequence][file_name]:
                    rt_start = 0.0
                    rt_end = 0.0
                    max_delta_mass = ms2_window
                elif ion_series in xpi_peak_obj.data[peptide_sequence][file_name][labeling_group]:
                    xpi_peak_data = xpi_peak_obj.data[peptide_sequence][file_name][labeling_group][ion_series]
                    peak_range = xpi_peak_data['peak_range']
                    list_rt_selected = peak_range.split(delimiter_residue)
                    rt_start = float(list_rt_selected[0])
                    rt_end = float(list_rt_selected[2])
                    max_delta_mass = xpi_peak_data['max_delta_mass']
                else:
                    rt_start = 0.0
                    rt_end = 0.0
                    max_delta_mass = ms2_window

                # If there is no data, skip it
                # if len(list_ints) == 0:
                #    continue
                if labeling_group not in dict_pep_intensities[peptide_sequence][metrics][ion_series]:
                    dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group] = \
                        {'intensity': [], 'mz_diff': [], 'rt': []}
                # Assign intensities, mz_diff and rt
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group]['intensity'] = list_ints
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group][
                    'mz_diff'] = list_mz_diff
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group]['rt'] = list_rt
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group][
                    'total_rt'] = list_total_rt
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group]['rt_start'] = rt_start
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group]['rt_end'] = rt_end
                dict_pep_intensities[peptide_sequence][metrics][ion_series][labeling_group][
                    'max_del'] = max_delta_mass

    # Calculate the total number of images
    pg_total = 0
    for peptide_sequence in dict_pep_intensities:
        for metrics in dict_pep_intensities[peptide_sequence]:
            for ion_series in dict_pep_intensities[peptide_sequence][metrics]:
                if len(list(dict_pep_intensities[peptide_sequence][metrics][ion_series].keys())) > 1:
                    pg_total += 1
    # Draw 2D plots
    pg_num = 0
    pg = Progress(pg_total, "Drawing 2D-plots...")
    for peptide_sequence in dict_pep_intensities:
        for metrics in dict_pep_intensities[peptide_sequence]:
            for ion_series in dict_pep_intensities[peptide_sequence][metrics]:
                # dict_ion2data: labeling_group -> {'intensity': [], 'mz_diff': [], 'rt': []}
                if len(list(dict_pep_intensities[peptide_sequence][metrics][ion_series].keys())) > 1:
                    pg_num += 1
                    pg.print_progress(pg_num)
                    dict_peptide2data = dict_pep_intensities[peptide_sequence][metrics][ion_series]
                    plot_title = "{0}_{2}_{1}".format(peptide_sequence, metrics, ion_series)
                    # dict_peptide2data: labeling_group -> {"intensity": [], "mz_diff": [], "rt": [], "total_rt": []}
                    # For general purpose
                    draw_2d_mass_profile_multi(dict_peptide2data, dir_2d_mass_profile, plot_title, config_xpi_viz.dpi,
                                               ms2_window, list_labels)
                    """
                    # For special purpose (only two labeling groups)
                    draw_2d_mass_profile_for_two(dict_peptide2data, dir_2d_mass_profile, plot_title, config_xpi_viz.dpi)
                    """


def do_common_process_scatter_plot(img_directory, xpi_4check, picking_method, config_xpi_viz):
    # calculate pearson correlation coefficient (pcc) and draw standard curves
    standard_curve_obj = Scatterplot(dict_observed_ratio=xpi_4check.data[picking_method],
                                     dict_run2info=config_xpi_viz.data,
                                     dpi=config_xpi_viz.dpi,
                                     data_directory=img_directory,
                                     is_log=config_xpi_viz.flag_log)
    return standard_curve_obj


def draw_time_series_plot_for_peptides(mode_arg, dir_output, xpi_4check, picking_method, config_xpi_viz, dict_tr2sum_ri,
                                       max_length_fragment, transition_obj):
    standard_curve_obj = do_common_process_scatter_plot(dir_output, xpi_4check, picking_method, config_xpi_viz)
    # Draw scatter plot with median trend line for multiple transitions of one peptide
    dict_seq2profile = standard_curve_obj.merge_ions_by_peptide(dict_tr2sum_ri)
    standard_curve_obj.get_time_points()
    pg_num = 1
    pg = Progress(len(list(dict_seq2profile.keys())), "Drawing {0} plots of peptides...".format(mode_arg))
    for sequence in dict_seq2profile:
        pg.print_progress(pg_num)
        pg_num += 1
        for ratio_name in dict_seq2profile[sequence]:
            x_local = dict_seq2profile[sequence][ratio_name]['intended']
            y = dict_seq2profile[sequence][ratio_name]['observed']
            protein = dict_seq2profile[sequence][ratio_name]['Protein']
            # m0_intensity = dict_seq2profile[sequence][ratio_name]['m0_intensity']
            x_axis_label = dict_seq2profile[sequence][ratio_name]['Unit']
            y_axis_label = ratio_name
            if config_xpi_viz.percent_ratio:
                y = [x_tmp * 100 for x_tmp in y]
                y_axis_label = "{0} (%)".format(ratio_name)
            if standard_curve_obj.is_log:
                if re.match("Log", x_axis_label):
                    x_axis_label = "{0}".format(x_axis_label)
                else:
                    x_axis_label = "Log10({0})".format(x_axis_label)
                y_axis_label = "Log10({0})".format(ratio_name.split(delimiter_residue)[1])
            ion_series = dict_seq2profile[sequence][ratio_name]['ion_series']
            ion_series_wo_repname = dict_seq2profile[sequence][ratio_name]['ion_series2']
            if mode_arg == "scatter":
                # To draw scatter plot for all the peptide ions
                draw_scatter_plot_med_line(x_local, y, ion_series_wo_repname, sequence, dir_output,
                                           standard_curve_obj.list_time_points, x_axis_label, y_axis_label,
                                           config_xpi_viz.dpi, max_length_fragment=max_length_fragment,
                                           is_trend_line=config_xpi_viz.flag_trend_line, 
                                           y_upper_lim=config_xpi_viz.filters[protein][colname_ylim])
            elif mode_arg == "filling":
                draw_filling_plot_for_peptides(x_local, y, ion_series_wo_repname, sequence, x_axis_label, y_axis_label,
                                               get_whole_pepties_list(transition_obj, config_xpi_viz), dir_output,
                                               config_xpi_viz, y_upper_lim=config_xpi_viz.filters[protein][colname_ylim])
            else:
                # To draw standard curve for all the peptide ions
                draw_scatter_plot_reg_line(x_local, y, "{0}_{1}".format(sequence, ratio_name), dir_output,
                                           x_axis_label, y_axis_label, config_xpi_viz.dpi)


def get_whole_pepties_list(transition_obj, config_xpi_viz_arg):
    list_pepseq_return = []
    list_rnames = get_rep_names(config_xpi_viz_arg)
    for transition_id in transition_obj.data:
        ion_series_local, peptide_sequence_local = get_ion_series_from_tr_id(transition_id)
        """
        # if there are replicate name in XPIViz_Config
        if len(list_rnames) > 0:
            for rname in list_rnames:
                pep_rep_name = make_unique_replicate_transition_id(peptide_sequence_local, rname)
                if pep_rep_name not in list_pepseq_return:
                    list_pepseq_return.append(pep_rep_name)
        else:
            if peptide_sequence_local not in list_pepseq_return:
                list_pepseq_return.append(peptide_sequence_local)
        """
        if peptide_sequence_local not in list_pepseq_return:
            list_pepseq_return.append(peptide_sequence_local)
    return sorted(list_pepseq_return)


def get_whole_transition_list(transition_obj, config_xpi_viz_arg):
    list_transition_return = []
    list_rnames = get_rep_names(config_xpi_viz_arg)
    for transition_id in transition_obj.data:
        # if there are replicate name in XPIViz_Config
        if len(list_rnames) > 0:
            for rname in list_rnames:
                tr_rep_name = make_unique_replicate_transition_id(transition_id, rname)
                if transition_id not in list_transition_return:
                    list_transition_return.append(tr_rep_name)
        else:
            if transition_id not in list_transition_return:
                list_transition_return.append(transition_id)
    return sorted(list_transition_return)


def make_unique_replicate_transition_id(tr_id_arg, rep_name_arg):
    if rep_name_arg != '':
        return "{0}{1}{2}".format(tr_id_arg, delimeter_replicates, rep_name_arg)
    else:
        return tr_id_arg


def get_rep_names(xpi_viz_config_arg):
    list_return = []
    for prm_run_local in xpi_viz_config_arg.data:
        rep_name_local = xpi_viz_config_arg.data[prm_run_local]['Replicate']
        if rep_name_local == '':
            continue
        if rep_name_local not in list_return:
            list_return.append(rep_name_local)
    return list_return


def draw_scatter_filtering_plot_for_proteins(mode_plot, dir_output, xpi_4check, picking_method, config_xpi_viz,
                                             dict_tr2sum_ri, transition_obj, flag_show=True):
    standard_curve_obj = do_common_process_scatter_plot(dir_output, xpi_4check, picking_method, config_xpi_viz)
    # Get XPI data for al the proteins
    if mode_plot == "enrichment_filter" or mode_plot == "enrichment_filter_log":
        dict_seq2profile = standard_curve_obj.merge_ions_by_protein(dict_tr2sum_ri, 'all')
    else:
        dict_seq2profile = standard_curve_obj.merge_ions_by_protein(dict_tr2sum_ri)
    # Progress
    pg_num = 1
    pg = Progress(len(list(dict_seq2profile.keys())), "Drawing {0} plots...".format(mode_plot))
    # Get time points
    standard_curve_obj.get_time_points()
    # Draw plots for each proteins
    for protein in sorted(standard_curve_obj.prot_list):
        if protein not in dict_seq2profile:
            draw_scatter_plot_med_line_empty(protein, dir_output)
            continue
        sum_ratio_threshold = config_xpi_viz.filters[protein]["sum_ratio"]
        sum_inten_threshold = config_xpi_viz.filters[protein]["sum_intensity"]
        sum_ratio_max_threshold = config_xpi_viz.filters[protein]["sum_ratio_max"]
        pg.print_progress(pg_num)
        pg_num += 1
        for ratio_name in dict_seq2profile[protein]:
            x_local = dict_seq2profile[protein][ratio_name]['intended']
            m0_intensity = dict_seq2profile[protein][ratio_name]['m0_intensity']
            x_axis_label = dict_seq2profile[protein][ratio_name]['Unit']
            y_axis_label = ratio_name
            # Get y value (ratio and m3 intensity) and apply percent ratio option
            y_m3 = dict_seq2profile[protein][ratio_name]['m3_intensity']
            if config_xpi_viz.percent_ratio:
                y = dict_seq2profile[protein][ratio_name]['observed']
                y = [x_tmp * 100 for x_tmp in y]
                y_axis_label = "{0} (%)".format(ratio_name)
                ## Changed this to corresponding the numbers on csv files of "Filtering" directory
                ## 12/11/18 by Lang Ho Lee
                ratio_sum_threshold = float(sum_ratio_threshold)
                sum_ratio_max_threshold = float(sum_ratio_max_threshold)
                # ratio_sum_threshold = float(sum_ratio_threshold) * 100
                # sum_ratio_max_threshold = float(sum_ratio_max_threshold) * 100
            else:
                y = dict_seq2profile[protein][ratio_name]['observed']
                ratio_sum_threshold = float(sum_ratio_threshold)
                sum_ratio_max_threshold = float(sum_ratio_max_threshold)
            if standard_curve_obj.is_log:
                if re.match("Log", x_axis_label):
                    x_axis_label = "{0}".format(x_axis_label)
                else:
                    x_axis_label = "Log10({0})".format(x_axis_label)
                if config_xpi_viz.percent_ratio:
                    percent_symbol = " (%)"
                else:
                    percent_symbol = ""
                y_axis_label = "Log10({0}){1}".format(ratio_name.split(delimiter_residue)[1], percent_symbol)
            ion_series = dict_seq2profile[protein][ratio_name]['ion_series']
            ion_series_wo_repname = dict_seq2profile[protein][ratio_name]['ion_series2']
            list_pepseq = get_whole_pepties_list(transition_obj, config_xpi_viz)
            if mode_plot == "time_series":
                # Calculate CV
                # print("CV of {1}: {0}".format(numpy.std(y)/numpy.mean(y), protein))
                if config_xpi_viz.is_colorful:
                    print("This function, 'colorful' is not provided anymore")
                    # To make protein scatter plot by selected peptides
                    # draw_protein_plot_by_peptides(x_local, y, ion_series, protein, dir_output, x_axis_label, y_axis_label, config_xpi_viz.dpi)
                else:
                    # Draw protein plot with black circles
                    draw_scatter_plot_med_line_4protein(x_local, y, ion_series_wo_repname, protein, dir_output,
                                                        x_axis_label, y_axis_label, config_xpi_viz.dpi,
                                                        is_trend_line=config_xpi_viz.flag_trend_line,
                                                        average_peptide_flag=True,
                                                        regression_flag=regression_flag,
                                                        y_upper_lim=config_xpi_viz.filters[protein][colname_ylim]) 
            elif mode_plot == 'filling':
                draw_filling_plot_for_proteins(x_local, y, ion_series_wo_repname, protein, x_axis_label, y_axis_label,
                                               list_pepseq, dir_output, mode_plot, config_xpi_viz)
            elif mode_plot == 'error_bar':
                draw_filling_plot_for_proteins(x_local, y, ion_series_wo_repname, protein, x_axis_label, y_axis_label,
                                               list_pepseq, dir_output, mode_plot, config_xpi_viz)
            elif mode_plot == "enrichment_filter":
                flag_log = False
                draw_ratio_intensity_plot(m0_intensity, y, y_m3, ion_series, protein, y_axis_label, ratio_sum_threshold,
                                          flag_log, sum_inten_threshold,
                                          get_whole_transition_list(transition_obj, config_xpi_viz),
                                          dir_output, sum_ratio_max_threshold, config_xpi_viz, flag_valid=flag_show)
            elif mode_plot == "enrichment_filter_log":
                flag_log = True
                draw_ratio_intensity_plot(m0_intensity, y, y_m3, ion_series, protein, y_axis_label, ratio_sum_threshold,
                                          flag_log, sum_inten_threshold,
                                          get_whole_transition_list(transition_obj, config_xpi_viz),
                                          dir_output, sum_ratio_max_threshold, config_xpi_viz, flag_valid=flag_show)
            else:
                UserMessageXPIViz("Wrong option for 'draw_time_series_plot_for_proteins' module")


def get_pepseq_from_ion_series(ion_series_arg):
    return ion_series_arg.split(delimiter_dash)[0]


def draw_ratio_intensity_plot(m0_intensity, y, y_m3, ion_series, protein, y_axis_label, ratio_sum_threshold, flag_log,
                              intensity_sum_threshold, list_whole_tr_id, dir_output, sum_ratio_max_threshold,
                              config_xpi_viz, flag_valid=False):
    dict_data = {}
    for m0_local, y_local, y_m3_local, ion_local in zip(m0_intensity, y, y_m3, ion_series):
        if y_local == numpy.inf:
            continue
        if ion_local not in dict_data:
            dict_data[ion_local] = {'m0_intensity': [], 'y': [], 'y_m3': []}
        dict_data[ion_local]['m0_intensity'].append(m0_local)
        dict_data[ion_local]['y'].append(y_local)
        dict_data[ion_local]['y_m3'].append(y_m3_local)
    draw_intensity_ratio_plot(dict_data, dir_output, protein, y_axis_label, config_xpi_viz.dpi,
                              config_xpi_viz.standard_label, flag_log, ratio_sum_threshold,
                              intensity_sum_threshold, sum_ratio_max_threshold, list_whole_tr_id,
                              flag_show_valid_frags_only=flag_valid)


def draw_filling_plot_for_proteins(x_arg, y, ion_series, protein, x_axis_label, y_axis_label, list_pepseq, dir_output,
                                   mode_arg, config_xpi_viz):
    dict_data = {}
    for x_local, y_local, ion_local in zip(x_arg, y, ion_series):
        if y_local == numpy.inf:
            continue
        if y_local == -1 * numpy.inf:
            continue
        try:
            # For replicate data
            peptide_local = get_pepseq_from_ion_series(ion_local)
            rep_name = extract_rep_name_tr_id(ion_local)
            pep_seq_local = make_unique_replicate_transition_id(peptide_local, rep_name)
        except IndexError:
            pep_seq_local = get_pepseq_from_ion_series(ion_local)
        if pep_seq_local not in dict_data:
            dict_data[pep_seq_local] = {}
        if x_local not in dict_data[pep_seq_local]:
            dict_data[pep_seq_local][x_local] = []
        dict_data[pep_seq_local][x_local].append(y_local)
    if mode_arg == "filling":
        # To make protein contour plot by selected peptides
        draw_filling_plot_by_peptides(dict_data, protein, dir_output, x_axis_label, y_axis_label,
                                      config_xpi_viz.dpi, list_pepseq, 0, y_upper_lim=config_xpi_viz.filters[protein][colname_ylim])
    elif mode_arg == "error_bar":
        draw_median_line_with_error_bar(dict_data, protein, dir_output, x_axis_label, y_axis_label,
                                        config_xpi_viz.dpi, list_pepseq, 0, y_upper_lim=config_xpi_viz.filters[protein][colname_ylim])
    else:
        UserMessageXPIViz.print_message("Wrong option. Choose 'filling' or 'error_bar': {0}".format(mode_arg))


def draw_filling_plot_for_peptides(x_arg, y, ion_series_arg, protein, x_axis_label, y_axis_label, list_pepseq,
                                   dir_output, config_xpi_viz, y_upper_lim):
    dict_data = {}
    for x_local, y_local, frag_ion in zip(x_arg, y, ion_series_arg):
        try:
            rep_name_local = extract_rep_name_tr_id(frag_ion)
            pep_seq_local = make_unique_replicate_transition_id(protein, rep_name_local)
        except IndexError:
            pep_seq_local = protein
        if y_local == numpy.inf:
            continue
        if x_local == numpy.inf:
            continue
        if pep_seq_local not in dict_data:
            dict_data[pep_seq_local] = {}
        if x_local not in dict_data[pep_seq_local]:
            dict_data[pep_seq_local][x_local] = []
        dict_data[pep_seq_local][x_local].append(y_local)
    # To make protein contour plot by selected peptides
    draw_filling_plot_by_peptides(dict_data, protein, dir_output, x_axis_label, y_axis_label,
                                  config_xpi_viz.dpi, list_pepseq, y_upper_lim=y_upper_lim)


def draw_standard_curve(dir_std_curves, xpi_4check, picking_method, config_xpi_viz, dict_tr2sum_ri):
    standard_curve_obj = do_common_process_scatter_plot(dir_std_curves, xpi_4check, picking_method, config_xpi_viz)
    standard_curve_obj.process_ratio_data(dict_tr2sum_ri)
    standard_curve_obj.print_pcc(os.path.join(dir_std_curves, "Standard_Curves_R2.csv"))


def calculate_rss(prot_name_arg, data_directory, dict_tr2sum_ri, config_xpi_viz):
    out_file_name = os.path.join(data_directory, file_out)
    xpi_output_obj = XPIOutput(out_file_name)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.set_size_inches(5, 3)
    # Get x (time point) and y (ratio) from XPI_Output.txt
    dict_x2y = {}
    for tr_id_local in xpi_output_obj.data:
        if tr_id_local not in dict_tr2sum_ri:
            continue
        elif Scatterplot.filter_fragment_by_sum_ratio(dict_tr2sum_ri[tr_id_local]):
            for prm_run_local in xpi_output_obj.data[tr_id_local]:
                # unit_local = config_XPIViz.data[prm_run_local]['Unit']
                ratio_name_local = config_xpi_viz.data[prm_run_local]['RatioName']
                time_point_local = float(config_xpi_viz.data[prm_run_local]['MR'])
                row = xpi_output_obj.data[tr_id_local][prm_run_local]
                ratio = row[ratio_name_local]
                prot_name = row["Protein_Name"]
                if ratio == "NA":
                    continue
                # Only for the paper
                if prot_name != prot_name_arg:
                    continue
                if time_point_local == 70:
                    continue
                if time_point_local not in dict_x2y:
                    dict_x2y[time_point_local] = []
                dict_x2y[time_point_local].append(float(ratio))
                plt.scatter(time_point_local, float(ratio), marker='o', s=circle_size)
    list_x = sorted(dict_x2y)
    med_y = []
    for time_point_local in list_x:
        med_y.append(numpy.median(dict_x2y[time_point_local]))
    # Make polynomial model
    co = numpy.polyfit(list_x, med_y, 30)
    p = numpy.poly1d(co)
    plt.plot(list_x, p(list_x), 'r-')
    # Calculate RSS
    rss = 0.0
    zipped_list = zip(list_x, med_y)
    for time_point_local, median_y_local in zipped_list:
        for y_local in dict_x2y[time_point_local]:
            rss += numpy.power(p(time_point_local) - y_local, 2)
    ax.set_title("RSS={0}".format(rss))
    ax.set_ylim(0, 0.01)
    ax.set_xlim(0, 25)
    img_file_path = os.path.join(data_directory, "RSS_Fit_{0}.png".format(prot_name_arg))
    plt.savefig(img_file_path, dpi=100)
    fig.clf()
    plt.close(fig)


def print_xpi_output(transition_obj, data_directory, list_labeling, picking_method, config_quant_obj, xpi_all_obj):
    transition_obj.get_mz_fragments()
    xpi_peak_info_obj = XPIPeakInfo(os.path.join(data_directory, file_peak_info))

    out_file_name = os.path.join(data_directory, file_out)
    fp_out = open(out_file_name, "w")
    # make header for different labeling method
    header_labeling = ""
    for label_group_name in list_labeling:
        for header in list_label_header:
            if header == "Intensity":
                header = "{0}_{1}".format(picking_method, header)
            header_labeling = append_string("{0}_{1}".format(label_group_name, header),
                                            header_labeling, delimiter_tab)
    # make header for ratio calculation between labeling method
    list_target_ratios = config_quant_obj.data[colname_config_ratio].split(delimiter_config)
    prefix = "{0}{1}{2}".format(delimiter_tab, colname_config_ratio, delimiter_residue)
    # If TTR is true, change ratio column header
    if config_quant_obj.data[colname_ttr]:
        header_ratio = ''
        for target_ratio in list_target_ratios:
            list_ratio = target_ratio.split(delimiter_ratio)
            ttr_header = "{0}{1}{4}({1}{3}{2})".format(prefix, list_ratio[0], list_ratio[1],
                                                       delimiter_config, delimiter_ratio)
            header_ratio = append_string(ttr_header, header_ratio, '')
    # If TTR is false, get ratio name as it is
    else:
        header_ratio = "{0}{1}".format(prefix, prefix.join(list_target_ratios))
    # Concatenate column headers
    header_xpi = "{2}\t{0}{1}\n".format(header_labeling, header_ratio, xpi_output_header)
    fp_out.write(header_xpi)

    # Print transition and labeled PRM data
    for transition_id in sorted(xpi_all_obj.data):
        ion_series, peptide_sequence = get_ion_series_from_tr_id(transition_id)
        # If transition_id is not in the transition list, skip it (For replicates handling and manual adjustment)
        if transition_id not in transition_obj.data:
            continue
        # Get transition information
        transition_output = "{0}\t{1}\t{2}\t{3}\t{4}_{5}\t{6}\t{7}".format(
            transition_id, transition_obj.data[transition_id]['Protein_Name'],
            transition_obj.data[transition_id]['Filter_String'],
            transition_obj.data[transition_id]['Precursor_Charge'],
            transition_obj.data[transition_id]['Low_RT'],
            transition_obj.data[transition_id]['High_RT'],
            transition_obj.data[transition_id]['Fragment_Sequence'],
            transition_obj.data[transition_id]['Ion_Series'])
        for MSRun in sorted(xpi_all_obj.data[transition_id]):
            scan_output = ""
            dict_label2intensity = {}
            # Check if the peptide was found in the MS_Run
            if MSRun not in xpi_peak_info_obj.data[peptide_sequence]:
                continue
            # Get PRM information of different labels
            for label_name in list_labeling:
                # Get peak info that passed thresholds
                dict_out_all = xpi_all_obj.data[transition_id][MSRun][label_name]
                dict_tmp = get_filtered_intensities(dict_out_all, xpi_peak_info_obj.data, peptide_sequence,
                                                    MSRun, label_name, ion_series)
                # Get narrow down peak info
                new_intensity_list = dict_tmp['intensity']
                new_rt_list = dict_tmp['rt']
                new_mz_diff_list = dict_tmp['mz_diff']
                new_mz_list = dict_tmp['mz']
                # Get representative values of fragment m/z, intensity, rt and delta_mass
                avg_mz = pick_intensity(new_mz_list, 'avg')
                if avg_mz == 0:
                    avg_mz = "NA"
                if len(new_rt_list) == 0:
                    rt_selected = "NA"
                else:
                    rt_selected = "{0:.2f}_{1:.2f}".format(pick_intensity(new_rt_list, 'min'),
                                                           pick_intensity(new_rt_list, 'max'))
                intensity_quantified = pick_intensity(new_intensity_list, picking_method)
                if len(new_mz_diff_list) == 0:
                    max_delta_mz = "NA"
                else:
                    max_delta_mz = pick_intensity(new_mz_diff_list, 'max')
                # Add results to output string
                scan = "{0}\t{1}\t{2}\t{3}\t{4}".format(avg_mz, intensity_quantified, len(new_intensity_list), rt_selected, max_delta_mz)
                scan_output = append_string(short_string=scan, template=scan_output,
                                            glue=delimiter_tab)
                # For ratio calculation
                dict_label2intensity[label_name] = float(intensity_quantified)
            # Get ratios described at XPIQuant config file
            ratio_output = ""
            for ratio_label in list_target_ratios:
                # Get ratio numerator and denominator from XPIQuant config file
                list_label_ratio = ratio_label.split(delimiter_ratio)
                if len(list_label_ratio) != 2:
                    UserMessageXPIViz("Wrong ratio format: {0}".format(ratio_label))
                numerator = get_value_by_col_name(list_label_ratio[0], dict_label2intensity)
                denominator = get_value_by_col_name(list_label_ratio[1], dict_label2intensity)
                # Calculate ratio
                ratio = get_ratio(numerator, denominator, config_quant_obj.data[colname_ttr])
                if isinstance(ratio, float):
                    ratio_output = append_string("{0:.4f}".format(ratio), ratio_output, delimiter_tab)
                else:
                    ratio_output = append_string(ratio, ratio_output, delimiter_tab)
            fp_out.write("{0}\t{1}\t{2}\t{3}\n".format(MSRun, transition_output, scan_output, ratio_output))
    fp_out.close()


def update_xpi_4check(data_directory, list_labeling, xpi_all_obj, xpi_peak_obj, config_quant_obj):
    ms2_window = float(config_quant_obj.data[colname_ms2_window])
    file_check_path = os.path.join(data_directory, file_check)
    header_quant = ''
    for label_group_name in list_labeling:
        for method in list_intensity_picking_method:
            # Extend header for existing intensity picking methods
            header_quant = append_string(short_string="{0}:{1}".format(label_group_name, method),
                                         template=header_quant, glue='\t')
    dict_narrowed_rt = xpi_peak_obj.data
    fp_out = open(file_check_path, "w")
    fp_out.write("PRM_Run_Name\tTransition_ID\tProtein_Name\t" + header_quant + "\n")
    for tr_id in sorted(xpi_all_obj.data):
        ion_series, peptide_sequence = get_ion_series_from_tr_id(tr_id)
        for file_name in sorted(xpi_all_obj.data[tr_id]):
            prot_name = ''
            out_string = ''
            # Check if the peptide was not found in the given XPIPeak output
            if peptide_sequence not in dict_narrowed_rt:
                continue
            # Check if the peptide was not found in the given file
            if not file_name in dict_narrowed_rt[peptide_sequence]:
                continue
            # Retrieve data from xpi_all_obj
            for label_group_name in list_labeling:
                dict_out_all = xpi_all_obj.data[tr_id][file_name][label_group_name]
                dict_tmp = get_filtered_intensities(dict_out_all, dict_narrowed_rt, peptide_sequence, file_name,
                                                    label_group_name, ion_series, ms2_window=ms2_window)
                prot_name = xpi_all_obj.tr2prot[tr_id]
                new_intensity_list = dict_tmp['intensity']
                # new_rt_list = dict_tmp['rt']
                # new_mz_diff_list = dict_tmp['mz_diff']
                # Quantification with various options
                out_string += "\t{0}".format(
                    "\t".join(map(str, get_statistics_intensity(new_intensity_list))))
            fp_out.write("{0}\t{1}\t{2}{3}\n".format(file_name, tr_id, prot_name, out_string))
    fp_out.close()
    ###########################################################################
    # E-N-D: Applying peak picking results to XPI outputs (file_check) #
    ###########################################################################


def print_xpi_output_selected(data_directory):    
    fp_selected_ions = open(os.path.join(data_directory, 'XPI_selected.csv'))
    fp_xpi_outputs = open(os.path.join(data_directory, 'XPI_output.txt'))
    tr_ids = []
    rows = csv.DictReader(fp_selected_ions, delimiter=',')
    for row in rows:
        flag_sum_intensity = row["flag_sum_intensity"]
        flag_sum_ratio = row["flag_sum_ratio"]
        flag_sum_ratio_max = row["flag_sum_ratio_max"]
        if flag_sum_intensity.lower() == "true":
            continue
        if flag_sum_ratio.lower() == "true":
            continue
        if flag_sum_ratio_max.lower() == "true":
            continue
        tr_ids.append(row[colname_transit_id])    
    fp_out = open(os.path.join(data_directory, "XPI_output_selected.txt"),"w")
    for data_line in fp_xpi_outputs:
        if re.match(colname_prm_run,data_line):
            fp_out.write(data_line)
        else:
            list_data = data_line.strip().split('\t')
            if list_data[1] in tr_ids:
                fp_out.write(data_line)
    fp_out.close()
########################
# END XPIViz libraries #
########################
